@@ -23,6 +23,7 @@
 #include <ptlib/sockets.h>
 #include <ptclib/telnet.h>
 #include <h225.h>
+#include <vector>
 #include "gk_const.h"
 #include "stl_supp.h"
 #include "SoftPBX.h"
@@ -33,10 +34,10 @@
 #include "gk.h"
 #include "GkStatus.h"
 
-
 void ReloadHandler(); // avoid to include...
 
 static const char *authsec="GkStatus::Auth";
+static const char *filteringsec="GkStatus::Filtering";
 
 // a very lightweight implementation of telnet socket
 class TelnetSocket : public ServerSocket {
@@ -225,6 +226,50 @@ private:
 		PString cmd
 		);
 
+    // Adds regular expression filter
+    void AddFilter(
+	// filter vector
+	std::vector<PString>& regexFilters,
+	// Regex to be matched against messages
+	PString& regex
+	);
+
+    /** Remove regular expression filter located
+	at the given index from the specified vector
+    */
+    void RemoveFilter(
+	// filter vector
+	std::vector<PString>& regexFilters,
+	// Index of filter to be removed
+	const PINDEX index
+	);
+
+    // Checks whether the given string is to be exclude
+    bool IsExcludeMessage(
+	// String to be check against the exclude regular expressions
+	const PString &msg
+	) const;
+
+    // Print a list of all filters in the specified vector
+    void PrintFilters(
+	// filter vector
+	std::vector<PString>& regexFilters
+	);
+
+    // Checks whether the given string is to be include
+    bool IsIncludeMessage(
+	// String to be check against the include regular expressions
+	const PString &msg
+	) const;
+    
+    // Match the given string against filters held by the specified vector
+    bool MatchFilter(
+	// filter vector
+	const std::vector<PString>& regexFilters,
+	// String to be matched
+	const PString &msg
+	) const;
+
 	/// the most recent command
 	PString m_lastCmd;
 	/// command being currently entered (and not yet completed)
@@ -241,6 +286,13 @@ private:
 	int m_instanceNo;
 	/// output trace level for this client
 	int m_traceLevel;
+
+    	// vectors of regular expressions to be matched against Status messages
+    	std::vector<PString> m_excludeFilterRegex;
+        std::vector<PString> m_includeFilterRegex;		
+	
+        // This flag indicates whether filtering is active or not
+        bool m_isFilteringActive;
 };
 
 
@@ -261,10 +313,11 @@ BOOL TelnetSocket::Accept(PSocket & socket)
 	SendWill(PTelnetSocket::StatusOption);
 	SendDont(PTelnetSocket::EchoOption);
 #ifndef LARGE_FDSET
-	Address addr;
-	WORD pt;
-	GetPeerAddress(addr, pt);
-	SetName(AsString(addr, pt));
+	Address raddr, laddr;
+	WORD rport = 0, lport = 0;
+	GetPeerAddress(raddr, rport);
+	GetLocalAddress(laddr, lport);
+	SetName(AsString(raddr, rport) + "=>" + AsString(laddr, lport));
 	SetReadTimeout(0);
 #else
 	// name already be set
@@ -281,6 +334,8 @@ int TelnetSocket::ReadChar()
 		if (!TCPSocket::Read(&currentByte, 1)) {
 			if (GetErrorCode(PSocket::LastReadError) != PSocket::Timeout) {
 				PTRACE(3, "TELNET\t" << GetName() << " closed the connection ("
+					<< GetErrorCode(PSocket::LastReadError) << '/' 
+					<< GetErrorNumber(PSocket::LastReadError) << ": "
 					<< GetErrorText(PSocket::LastReadError) << ')'
 					);
 				Close();
@@ -462,6 +517,7 @@ void GkStatus::AuthenticateClient(
 		newClient->WriteString("\r\nAccess forbidden!\r\n");
 		newClient->Flush();
 		delete newClient;
+		newClient = NULL;
 	}
 }
 
@@ -670,6 +726,13 @@ void GkStatus::OnStart()
 	m_commands["rotatelog"] = e_RotateLog;
 	m_commands["setlog"] = e_SetLogFilename;
 #endif
+	m_commands["addincludefilter"] = e_AddIncludeFilter;
+	m_commands["removeincludefilter"] = e_RemoveIncludeFilter;
+	m_commands["addexcludefilter"] = e_AddExcludeFilter;
+	m_commands["removeexcludefilter"] = e_RemoveExcludeFilter;
+	m_commands["filter"] = e_Filter;
+	m_commands["printexcludefilters"] = e_PrintExcludeFilters;
+	m_commands["printincludefilters"] = e_PrintIncludeFilters;
 }
 
 void GkStatus::ReadSocket(
@@ -694,6 +757,8 @@ void GkStatus::CleanUp()
 				m_removed.erase(i);
 				--m_rmsize;
 				delete client;
+				client = NULL;
+				iter = m_removed.begin();	// reset iterator after changing list
 			}
 		}
 	}
@@ -710,9 +775,25 @@ StatusClient::StatusClient(
 	m_gkStatus(GkStatus::Instance()),
 	m_numExecutingCommands(0),
 	m_instanceNo(instanceNo),
-	m_traceLevel(MAX_STATUS_TRACE_LEVEL)
+	m_traceLevel(MAX_STATUS_TRACE_LEVEL),
+	m_isFilteringActive(false)
 {
-	SetWriteTimeout(10);
+        PStringToString filters = GkConfig()->GetAllKeyValues(filteringsec);
+        SetWriteTimeout(10);
+	
+	PString key;
+	PString data;
+	for (PINDEX i = 0; i < filters.GetSize(); i++) {
+	    key = filters.GetKeyAt(i);
+	    data = filters.GetDataAt(i);
+	    PStringArray regexArray(data.Tokenise("\n", false));
+	    for (PINDEX k = 0; k < regexArray.GetSize(); ++k) {
+		if(filters.GetKeyAt(i) == "IncludeFilter")
+		    AddFilter(m_includeFilterRegex, regexArray[k]);
+		else if(filters.GetKeyAt(i) == "ExcludeFilter")
+		    AddFilter(m_excludeFilterRegex, regexArray[k]);
+	    }
+	}
 }
 
 bool StatusClient::ReadCommand(
@@ -780,8 +861,21 @@ bool StatusClient::WriteString(
 	if (level > m_traceLevel)
 		return true;
 	if (CanFlush())
-		Flush();
-	return WriteData(msg, msg.GetLength());
+	    Flush();
+
+	if (m_isFilteringActive == false)
+	    return WriteData(msg, msg.GetLength());
+
+	// If we use filters check if the message is to be ignored
+	if (IsExcludeMessage(msg))
+	    return true;
+	
+	// If we use filters check if the message is to be shown
+	if (IsIncludeMessage(msg))
+	    return WriteData(msg, msg.GetLength());
+
+	// Otherwise, do not show the message at all
+	return true;
 }
 
 /*
@@ -856,6 +950,9 @@ void StatusClient::DoDebug(
 	const PStringArray& args
 	)
 {
+        bool tmp = m_isFilteringActive;
+	m_isFilteringActive = false;
+
 	if (args.GetSize() <= 1) {
 		WriteString("Debug options:\r\n"
 			"  trc [+|-|n]       Show/modify trace level for the log\r\n"
@@ -910,6 +1007,7 @@ void StatusClient::DoDebug(
 			WriteString("Unknown debug command!\r\n");
 		}
 	}
+	m_isFilteringActive = tmp;
 }
 
 bool StatusClient::CheckAuthRule(
@@ -965,6 +1063,9 @@ bool StatusClient::AuthenticateUser()
 	const time_t now = time(NULL);
 	const int delay = GkConfig()->GetInteger(authsec, "DelayReject", 0);
 	const int loginTimeout = GkConfig()->GetInteger(authsec, "LoginTimeout", 120);
+	bool tmp = m_isFilteringActive;
+	
+	m_isFilteringActive = false;
 
 	for (int retries = 0; retries < 3; ++retries) {
 		PString login, password;
@@ -987,6 +1088,7 @@ bool StatusClient::AuthenticateUser()
 			PTRACE(5, "STATUS\tCould not find password in the config for user " << login);
 		else if (!password && password == storedPassword) {
 			m_user = login;
+			m_isFilteringActive = tmp;
 			return true;
 		} else
 			PTRACE(5, "STATUS\tPassword mismatch for user " << login);
@@ -996,6 +1098,7 @@ bool StatusClient::AuthenticateUser()
 		if ((time(NULL) - now) > loginTimeout)
 			break;
 	}
+	m_isFilteringActive = tmp;
 	return false;
 }
 
@@ -1024,30 +1127,38 @@ void StatusClient::ExecCommand(
 		// disconnect call on this IP number
 		if (args.GetSize() == 2)
 			SoftPBX::DisconnectIp(args[1]);
-		else
+		else {
 			WriteString("Syntax Error: DisconnectIp IP_ADDRESS\r\n");
+			PTRACE(2, "Syntax Error: DisconnectIp IP_ADDRESS");
+		}
 		break;
 	case GkStatus::e_DisconnectAlias:
 		// disconnect call on this alias
 		if (args.GetSize() == 2)
 			SoftPBX::DisconnectAlias(args[1]);
-		else
+		else {
 			WriteString("Syntax Error: DisconnectAlias ALIAS\r\n");
+			PTRACE(2, "Syntax Error: DisconnectAlias ALIAS");
+		}
 		break;
 	case GkStatus::e_DisconnectCall:
 		// disconnect call with this call number
 		if (args.GetSize() >= 2)
 			for (PINDEX p=1; p < args.GetSize(); ++p)
 				SoftPBX::DisconnectCall(args[p].AsInteger());
-		else
+		else {
 			WriteString("Syntax Error: DisconnectCall CALL_NUMBER [CALL_NUMBER...]\r\n");
+			PTRACE(2, "Syntax Error: DisconnectCall CALL_NUMBER [CALL_NUMBER...]");
+		}
 		break;
 	case GkStatus::e_DisconnectEndpoint:
 		// disconnect call on this alias
 		if (args.GetSize() == 2)
 			SoftPBX::DisconnectEndpoint(args[1]);
-		else
+		else {
 			WriteString("Syntax Error: DisconnectEndpoint ENDPOINT_IDENTIFIER\r\n");
+			PTRACE(2, "Syntax Error: DisconnectEndpoint ENDPOINT_IDENTIFIER");
+		}
 		break;
 	case GkStatus::e_DisconnectSession:
 		// disconnect a user from status port
@@ -1056,8 +1167,10 @@ void StatusClient::ExecCommand(
 				WriteString("Session " + args[1] + " disconnected\r\n");
 			else
 				WriteString("Session " + args[1] + " not found\r\n");
-		else
+		else {
 			WriteString("Syntax Error: DisconnectSession SESSION_ID\r\n");
+			PTRACE(2, "Syntax Error: DisconnectSession SESSION_ID");
+		}
 		break;
 	case GkStatus::e_ClearCalls:
 		SoftPBX::DisconnectAll();
@@ -1088,14 +1201,18 @@ void StatusClient::ExecCommand(
 	case GkStatus::e_Find:
 		if (args.GetSize() == 2)
 			SoftPBX::PrintEndpoint(args[1], this, FALSE);
-		else
+		else {
 			WriteString("Syntax Error: Find ALIAS\r\n");
+			PTRACE(2, "Syntax Error: Find ALIAS");
+		}
 		break;
 	case GkStatus::e_FindVerbose:
 		if (args.GetSize() == 2)
 			SoftPBX::PrintEndpoint(args[1], this, TRUE);
-		else
+		else {
 			WriteString("Syntax Error: FindVerbose ALIAS\r\n");
+			PTRACE(2, "Syntax Error: FindVerbose ALIAS");
+		}
 		break;
 	case GkStatus::e_Yell:
 		m_gkStatus->SignalStatus(PString("  "+ WhoAmI() + ": " + cmd + "\r\n"));
@@ -1127,27 +1244,35 @@ void StatusClient::ExecCommand(
 		// unregister this alias
 		if (args.GetSize() == 2)
 			SoftPBX::UnregisterAlias(args[1]);
-		else
+		else {
 			WriteString("Syntax Error: UnregisterAlias ALIAS\r\n");
+			PTRACE(2, "Syntax Error: UnregisterAlias ALIAS");
+		}
 		break;
 	case GkStatus::e_UnregisterIp:
 		// unregister this IP
 		if (args.GetSize() == 2)
 			SoftPBX::UnregisterIp(args[1]);
-		else
+		else {
 			WriteString("Syntax Error: UnregisterIp IP_ADDRESS\r\n");
-			break;
+			PTRACE(2, "Syntax Error: UnregisterIp IP_ADDRESS");
+		}
+		break;
 	case GkStatus::e_TransferCall:
 		if (args.GetSize() == 3)
 			SoftPBX::TransferCall(args[1], args[2]);
-		else
+		else {
 			WriteString("Syntax Error: TransferCall SOURCE DESTINATION\r\n");
+			PTRACE(2, "Syntax Error: TransferCall SOURCE DESTINATION");
+		}
 		break;
 	case GkStatus::e_MakeCall:
 		if (args.GetSize() == 3)
 			SoftPBX::MakeCall(args[1], args[2]);
-		else
+		else {
 			WriteString("Syntax Error: MakeCall SOURCE DESTINATION\r\n");
+			PTRACE(2, "Syntax Error: MakeCall SOURCE DESTINATION");
+		}
 		break;
 	case GkStatus::e_Reload:
 		{
@@ -1161,6 +1286,7 @@ void StatusClient::ExecCommand(
 	case GkStatus::e_Shutdown:
 		if (!Toolkit::AsBool(GkConfig()->GetString(authsec, "Shutdown", "1"))) {
 			WriteString("Shutdown not allowed!\r\n");
+			PTRACE(2, "Shutdown not allowed!");
 			break;
 		}
 		SoftPBX::PrintStatistics(this, true);
@@ -1168,21 +1294,36 @@ void StatusClient::ExecCommand(
 		break;
 	case GkStatus::e_RouteToAlias:
 		if (args.GetSize() == 4) {
-			RasServer::Instance()->GetVirtualQueue()->RouteToAlias(args[1], "", args[2], args[3].AsUnsigned());
-		} else
-			WriteString("Syntax Error: RouteToAlias TARGET_ALIAS CALLING_ENDPOINT_ID CRV\r\n");
+			RasServer::Instance()->GetVirtualQueue()->RouteToAlias(args[1], "", args[2], args[3].AsUnsigned(), "");
+		} else if (args.GetSize() == 5) {
+			args[4].Replace("-", " ", true);
+			RasServer::Instance()->GetVirtualQueue()->RouteToAlias(args[1], "", args[2], args[3].AsUnsigned(), args[4]);
+		} else {
+			WriteString("Syntax Error: RouteToAlias TARGET_ALIAS CALLING_ENDPOINT_ID CRV [CALLID]\r\n");
+			PTRACE(2, "Syntax Error: RouteToAlias TARGET_ALIAS CALLING_ENDPOINT_ID CRV [CALLID]");
+		}
 		break;
 	case GkStatus::e_RouteToGateway:
 		if (args.GetSize() == 5) {
-			RasServer::Instance()->GetVirtualQueue()->RouteToAlias(args[1], args[2], args[3], args[4].AsUnsigned());
-		} else
-			WriteString("Syntax Error: RouteToGateway TARGET_ALIAS TARGET_IP CALLING_ENDPOINT_ID CRV\r\n");
+			RasServer::Instance()->GetVirtualQueue()->RouteToAlias(args[1], args[2], args[3], args[4].AsUnsigned(), "");
+		} else if (args.GetSize() == 6) {
+			args[5].Replace("-", " ", true);
+			RasServer::Instance()->GetVirtualQueue()->RouteToAlias(args[1], args[2], args[3], args[4].AsUnsigned(), args[5]);
+		} else {
+			WriteString("Syntax Error: RouteToGateway TARGET_ALIAS TARGET_IP CALLING_ENDPOINT_ID CRV [CALLID]\r\n");
+			PTRACE(2, "Syntax Error: RouteToGateway TARGET_ALIAS TARGET_IP CALLING_ENDPOINT_ID CRV [CALLID]");
+		}
 		break;
 	case GkStatus::e_RouteReject:
 		if (args.GetSize() == 3) {
-			RasServer::Instance()->GetVirtualQueue()->RouteReject(args[1], args[2].AsUnsigned());
-		} else
-			WriteString("Syntax Error: RouteReject CALLING_ENDPOINT_ID CRV\r\n");
+			RasServer::Instance()->GetVirtualQueue()->RouteReject(args[1], args[2].AsUnsigned(), "");
+		} else if (args.GetSize() == 4) {
+			args[3].Replace("-", " ", true);
+			RasServer::Instance()->GetVirtualQueue()->RouteReject(args[1], args[2].AsUnsigned(), args[3]);
+		} else {
+			WriteString("Syntax Error: RouteReject CALLING_ENDPOINT_ID CRV [CALLID]\r\n");
+			PTRACE(2, "Syntax Error: RouteReject CALLING_ENDPOINT_ID CRV [CALLID]");
+		}
 		break;
 	case GkStatus::e_Trace:
 		if (args.GetSize() == 2) {
@@ -1197,6 +1338,7 @@ void StatusClient::ExecCommand(
 					m_traceLevel = level;
 				else {
 					WriteString("Syntax Error: trace 0|1|2|\"min\"|\"max\"\r\n");
+					PTRACE(2, "Syntax Error: trace 0|1|2|\"min\"|\"max\"");
 					break;
 				}
 			}
@@ -1216,23 +1358,179 @@ void StatusClient::ExecCommand(
 		if (args.GetSize() == 2) {
 			if (Gatekeeper::SetLogFilename(args[1])) {
 				WriteString("Logging to the file '" + args[1] + "'\r\n");
-			} else						
+			} else {					
 				WriteString("Failed to open the log file'" + args[1] + "'\r\n");
-		} else
-			WriteString("Syntax Error: setlogfilename <logfilepath>\r\n");
+				PTRACE(2, "Failed to open the log file'" + args[1] + "'");
+			}
+		} else {
+			WriteString("Syntax Error: setlog <logfilepath>\r\n");
+			PTRACE(2, "Syntax Error: setlog <logfilepath>");
+		}
 		break;				
 #endif
 
+	case GkStatus::e_AddIncludeFilter:
+	    if (args.GetSize() == 2) {
+		AddFilter(m_includeFilterRegex, args[1]);
+	    } else {
+			WriteString("Syntax Error: addincludefilter REGEX\r\n");
+			PTRACE(2, "Syntax Error: addincludefilter REGEX");
+		}
+	    break;
+	case GkStatus::e_RemoveIncludeFilter:
+	    if (args.GetSize() == 2) {
+		RemoveFilter(m_includeFilterRegex, atoi(args[1]));
+	    } else {
+			WriteString("Syntax Error: removeincludefilter FILTER_INDEX\r\n");
+			PTRACE(2, "Syntax Error: removeincludefilter FILTER_INDEX");
+		}
+	    break;
+	case GkStatus::e_AddExcludeFilter:
+	    if (args.GetSize() == 2) {
+		AddFilter(m_excludeFilterRegex, args[1]);
+	    } else {
+			WriteString("Syntax Error: addexcludefilter REGEX\r\n");
+			PTRACE(2, "Syntax Error: addexcludefilter REGEX");
+		}
+	    break;
+	case GkStatus::e_RemoveExcludeFilter:
+	    if (args.GetSize() == 2) {
+		RemoveFilter(m_excludeFilterRegex, atoi(args[1]));
+	    } else {
+			WriteString("Syntax Error: removeincludefilter FILTER_INDEX\r\n");
+			PTRACE(2, "Syntax Error: removeincludefilter FILTER_INDEX");
+		}
+	    break;
+	case GkStatus::e_Filter:
+	    if (args.GetSize() == 2) {
+		if (!(args[1] *= "0") && !(args[1] *= "1")) {
+		    WriteString("Syntax Error: filter 0|1\r\n");
+			PTRACE(2, "Syntax Error: filter 0|1");
+		    break;
+		}
+		m_isFilteringActive = Toolkit::AsBool(args[1]);
+	    }
+	    if (m_isFilteringActive) {
+		PString msg("Filtering is active\r\n");
+		WriteData(msg, msg.GetLength());
+	    }
+	    else {
+		PString msg("Filtering is not active\r\n");
+		WriteData(msg, msg.GetLength());
+	    }
+	    break;
+	case GkStatus::e_PrintExcludeFilters:
+	    PrintFilters(m_excludeFilterRegex);
+	    break;
+	case GkStatus::e_PrintIncludeFilters:
+	    PrintFilters(m_includeFilterRegex);
+	    break;
 	default:
 		// commmand not recognized
 		WriteString("Error: Unknown command " + cmd + "\r\n");
-		PTRACE(5, "STATUS\tUnknown command '" << cmd << "' from client " << Name());
+		PTRACE(2, "STATUS\tUnknown command '" << cmd << "' from client " << Name());
 		break;
 	}
 	PWaitAndSignal lock(m_cmutex);
 	--m_numExecutingCommands;
 }
 
+void StatusClient::AddFilter(
+    // vector of filters
+    std::vector<PString>& regexFilters,
+    // Regular expression
+    PString& regex
+    )
+{
+    regexFilters.push_back(regex);
+}
+
+void StatusClient::RemoveFilter(
+    // vector of filters
+    std::vector<PString>& regexFilters,
+    // Index of filter to be removed
+    const PINDEX index
+    )
+{
+    if (index < 0 || index >= (PINDEX) regexFilters.size()) {
+		PString msg("Index mismatch.\r\n");
+		WriteData(msg, msg.GetLength());
+		return;
+    }
+
+    std::vector<PString>::iterator it = regexFilters.begin();
+
+    PINDEX i = 0;
+    while(i != index) {
+		++it;
+		++i;
+    }
+
+    regexFilters.erase(it);
+}
+
+bool StatusClient::IsExcludeMessage(
+    // String to be chacked against exclude regular expressions 
+    const PString &msg
+    ) const
+{
+    return MatchFilter(m_excludeFilterRegex, msg);
+}
+
+bool StatusClient::MatchFilter(
+    // filter vector
+    const std::vector<PString>& regexFilters,
+    // String to be matched against filters
+    const PString &msg
+    ) const
+{
+    std::vector<PString>::const_iterator it = regexFilters.begin();
+    std::vector<PString>::const_iterator itEnd = regexFilters.end();
+
+    for(; it != itEnd; ++it) {
+	if (Toolkit::MatchRegex(msg, *it))
+	    return true;
+    }
+
+    return false;
+}
+
+bool StatusClient::IsIncludeMessage(
+    // String to be chacked against include regular expressions 
+    const PString &msg
+    ) const
+{
+    return MatchFilter(m_includeFilterRegex, msg);
+}
+
+void StatusClient::PrintFilters(
+    // filter vector
+    std::vector<PString>& regexFilters
+    )
+{
+    std::vector<PString>::const_iterator it = regexFilters.begin();
+    std::vector<PString>::const_iterator itEnd = regexFilters.end();
+    unsigned int index = 0;
+    PString count;
+
+    if (regexFilters.size() == 0) {
+		PString msg("No Filters are defined\r\n");
+		WriteData(msg, msg.GetLength());
+		return;
+    }
+
+    PString msg("Filter List:\r\n");
+    WriteData(msg, msg.GetLength());
+    for(; it != itEnd; ++it, ++index) {
+		count = index;
+		PString item(count);
+		item += ") " + *it + "\r\n";
+		WriteData(item, item.GetLength());
+    }
+    
+    msg = ";\r\n";
+    WriteData(msg, msg.GetLength());
+}
 
 // class StatusListener
 StatusListener::StatusListener(
@@ -1241,11 +1539,14 @@ StatusListener::StatusListener(
 	)
 {
 	const unsigned queueSize = GkConfig()->GetInteger("ListenQueueLength", GK_DEF_LISTEN_QUEUE_LENGTH);
-	if (!Listen(addr, queueSize, port, PSocket::CanReuseAddress))
+	if (!Listen(addr, queueSize, port, PSocket::CanReuseAddress)) {
 		PTRACE(1, "STATUS\tCould not open listening socket at " << addr << ':' << port
-			<< ", error(" << GetErrorCode(PSocket::LastGeneralError) << ", " 
-			<< GetErrorText(PSocket::LastGeneralError) << ')'
+			<< " - error " << GetErrorCode(PSocket::LastGeneralError) << '/'
+			<< GetErrorNumber(PSocket::LastGeneralError) << ": " 
+			<< GetErrorText(PSocket::LastGeneralError)
 			);
+		Close();
+	}
 	SetName(AsString(addr, GetPort()));
 }
 
