@@ -15,7 +15,7 @@
 //////////////////////////////////////////////////////////////////
 
 #ifndef PROXYCHANNEL_H
-#define PROXYCHANNEL_H "@(#) $Id: ProxyChannel.h,v 1.40 2005/03/21 16:19:59 zvision Exp $"
+#define PROXYCHANNEL_H "@(#) $Id: ProxyChannel.h,v 1.47 2006/09/16 20:42:50 shorne Exp $"
 
 #include <vector>
 #include <list>
@@ -65,7 +65,11 @@ public:
 		Error
 	};
 
-	ProxySocket(IPSocket *, const char *);
+	ProxySocket(
+		IPSocket *self,
+		const char *type,
+		WORD buffSize = 1536
+	);
 	~ProxySocket() = 0; // abstract class
 
 	// new virtual function
@@ -81,6 +85,11 @@ public:
 	ProxyHandler *GetHandler() const { return handler; }
 	void SetHandler(ProxyHandler *h) { handler = h; }
 
+private:
+	ProxySocket();
+	ProxySocket(const ProxySocket&);
+	ProxySocket& operator=(const ProxySocket&);
+
 protected:
 	BYTE *wbuffer;
 	WORD wbufsize, buflen;
@@ -107,6 +116,13 @@ public:
 	virtual bool ForwardData();
 	virtual bool TransmitData(const PBYTEArray &);
 
+	void RemoveRemoteSocket();
+	
+private:
+	TCPProxySocket();
+	TCPProxySocket(const TCPProxySocket&);
+	TCPProxySocket& operator=(const TCPProxySocket&);
+
 protected:
 	bool ReadTPKT();
 
@@ -120,7 +136,6 @@ private:
 	BYTE *bufptr;
 };
 
-
 class CallSignalSocket : public TCPProxySocket {
 public:
 	CallSignalSocket();
@@ -148,12 +163,19 @@ public:
 	bool IsNATSocket() const { return m_isnatsocket; }
 	void OnH245ChannelClosed() { m_h245socket = 0; }
 	void SetPeerAddress(const Address &, WORD);
+	Address GetLocalAddr() { return localAddr; }
+	Address GetPeerAddr() { return peerAddr; }
+	Address GetMasqAddr() { return masqAddr; }
 	void BuildFacilityPDU(Q931 &, int, const PObject * = 0);
 	void RemoveCall();
 
 	// override from class ServerSocket
 	virtual void Dispatch();
-
+	void DispatchNextRoute();
+	Result RetrySetup();
+	void TryNextRoute();
+	void RemoveH245Handler();
+	
 protected:
 	CallSignalSocket(CallSignalSocket *);
 	
@@ -179,6 +201,8 @@ protected:
 	template<class UUIE> bool HandleH245Address(UUIE & uu)
 	{
 		if (uu.HasOptionalField(UUIE::e_h245Address)) {
+			if (m_call)
+				m_call->SetH245ResponseReceived();
 			if (SetH245Address(uu.m_h245Address))
 				return (m_h245handler != 0);
 			uu.RemoveOptionalField(UUIE::e_h245Address);
@@ -186,10 +210,16 @@ protected:
 		}
 		return false;
 	}
+	
 	template<class UUIE> bool HandleFastStart(UUIE & uu, bool fromCaller)
 	{
-		return (m_h245handler && uu.HasOptionalField(UUIE::e_fastStart)) ?
-			OnFastStart(uu.m_fastStart, fromCaller) : false;
+		if (!uu.HasOptionalField(UUIE::e_fastStart))
+			return false;
+			
+		if (!fromCaller && m_call)
+			m_call->SetFastStartResponseReceived();
+			
+		return m_h245handler != NULL ? OnFastStart(uu.m_fastStart, fromCaller) : false;
 	}
 
 private:
@@ -201,6 +231,7 @@ private:
 	// if return false, the h245Address field will be removed
 	bool SetH245Address(H225_TransportAddress &);
 	bool InternalConnectTo();
+	bool ForwardCallConnectTo();
 
 	/** @return
 	    A string that can be used to identify a calling number.
@@ -228,12 +259,14 @@ private:
 		const SetupMsg& setup
 		) const;
 
+	void SetCallTypePlan(Q931 *q931);
+
 protected:
 	callptr m_call;
 	// localAddr is NOT the local address the socket bind to,
 	// but the local address that remote socket bind to
 	// they may be different in multi-homed environment
-	Address localAddr, peerAddr;
+	Address localAddr, peerAddr, masqAddr;
 	WORD peerPort;
 
 private:
@@ -249,6 +282,9 @@ private:
 	bool m_callerSocket;
 	/// H.225.0 protocol version in use by the remote party
 	unsigned m_h225Version;
+	/// raw Setup data as received from the caller
+	PBYTEArray m_rawSetup;
+	PMutex infomutex;    // Information PDU processing Mutex
 };
 
 class CallSignalListener : public TCPListenSocket {
@@ -272,7 +308,10 @@ public:
 	void Insert(UDPProxySocket *, UDPProxySocket *);
 	void MoveTo(ProxyHandler *, TCPProxySocket *);
 	bool IsEmpty() const { return m_socksize == 0; }
-
+	void LoadConfig();
+	bool Detach(TCPProxySocket *);
+	void Remove(TCPProxySocket *);
+	
 private:
 	// override from class RegularJob
 	virtual void OnStart();
@@ -285,7 +324,7 @@ private:
 	void AddPairSockets(IPSocket *, IPSocket *);
 	void FlushSockets();
 	void Remove(iterator);
-	void Remove(ProxySocket *socket);
+	void DetachSocket(IPSocket *socket);
 
 	ProxyHandler();
 	ProxyHandler(const ProxyHandler&);
@@ -293,6 +332,8 @@ private:
 	
 private:
 	std::list<PTime *> m_removedTime;
+	/// time to wait before deleting a closed socket
+	PTimeInterval m_socketCleanupTimeout;
 };
 
 class HandlerList {
