@@ -38,8 +38,16 @@
 #include "gkacct.h"
 #include "gktimer.h"
 #include "RasSrv.h"
+#include "pwlib_compat.h"
 
+#ifdef hasH460
+  #include <h4601.h>
+#endif
+
+const char *LRQFeaturesSection = "RasSrv::LRQFeatures";
+const char *RRQFeatureSection = "RasSrv::RRQFeatures";
 using namespace std;
+using Routing::Route;
 
 #ifndef NEED_BROADCASTLISTENER
 #if (defined P_LINUX) || (defined P_FREEBSD) || (defined P_HPUX9) || (defined P_SOLARIS)
@@ -104,6 +112,14 @@ private:
 		ARQAuthData& authData
 		) const;
 
+	/** @return
+	    A string that can be used to identify the billing number.
+	*/
+	PString GetCallLinkage(
+		/// additional data, like call record and requesting endpoint
+		ARQAuthData& authData
+		) const;
+
 	endptr RequestingEP, CalledEP;
 	PString destinationString;
 };
@@ -163,14 +179,17 @@ bool GatekeeperMessage::Read(RasListener *socket)
 	const int buffersize = 4096;
 	BYTE buffer[buffersize];
 	if (!socket->Read(buffer, buffersize)) {
-		PTRACE(1, "RAS\tRead error: " << socket->GetErrorText(PSocket::LastReadError));
+		PTRACE(1, "RAS\tRead error " << socket->GetErrorCode(PSocket::LastReadError)
+			<< '/' << socket->GetErrorNumber(PSocket::LastReadError) << ": "
+			<< socket->GetErrorText(PSocket::LastReadError)
+			);
 		return false;
 	}
 	socket->GetLastReceiveAddress(m_peerAddr, m_peerPort);
 	PTRACE(2, "RAS\tRead from " << m_peerAddr << ':' << m_peerPort);
 	m_rasPDU = PPER_Stream(buffer, socket->GetLastReadCount());
 	bool result = m_recvRAS.Decode(m_rasPDU);
-	PTRACE_IF(1, !result, "RAS\tCouldn't decode message!");
+	PTRACE_IF(1, !result, "RAS\tCould not decode message from " << m_peerAddr << ':' << m_peerPort);
 	return result;
 }
 
@@ -183,12 +202,22 @@ bool GatekeeperMessage::Reply() const
 // class RasListener
 RasListener::RasListener(const Address & addr, WORD pt) : m_ip(addr)
 {
-	Listen(addr, 0, pt, PSocket::CanReuseAddress);
+	if (!Listen(addr, 0, pt, PSocket::CanReuseAddress)) {
+		PTRACE(1, "RAS\tCould not open listening socket at " << addr << ':' << pt
+			<< " - error " << GetErrorCode(PSocket::LastGeneralError) << '/'
+			<< GetErrorNumber(PSocket::LastGeneralError) << ": " 
+			<< GetErrorText(PSocket::LastGeneralError)
+			);
+		Close();
+	}
 	SetWriteTimeout(1000); // TODO: read from config
 	SetName(AsString(addr, pt) + "(U)");
 	m_signalPort = 0;
 	// note: this won't be affected by reloading
 	m_virtualInterface = GkConfig()->HasKey("NetworkInterfaces");
+    // Check if we have external IP setting 
+	if (!m_virtualInterface)  
+		m_virtualInterface = GkConfig()->HasKey("ExternalIP");
 }
 
 RasListener::~RasListener()
@@ -234,11 +263,20 @@ bool RasListener::SendRas(const H225_RasMessage & rasobj, const Address & addr,
 	if (result)
 		PTRACE(5, "RAS\tSent Successful");
 	else
-		PTRACE(4, "RAS\tWrite error: " << GetErrorText(PSocket::LastWriteError));
+		PTRACE(1, "RAS\tWrite error " << GetErrorCode(PSocket::LastWriteError) << '/'
+			<< GetErrorNumber(PSocket::LastWriteError) << ": "
+			<< GetErrorText(PSocket::LastWriteError)
+			);
 #endif
 	return result;
 }
 
+PIPSocket::Address RasListener::GetPhysicalAddr(const Address & addr) const
+{
+	// Return the physical address. This is used when setting sockets
+	return m_ip;
+}
+
 PIPSocket::Address RasListener::GetLocalAddr(const Address & addr) const
 {
 	return m_virtualInterface ? Toolkit::Instance()->GetRouteTable()->GetLocalAddress(addr) : m_ip;
@@ -495,6 +533,7 @@ bool GkInterface::ValidateSocket(IPSocket *socket, WORD & port)
 		} else {
 			PTRACE(1, "Can't listen to " << socket->GetName());
 			delete socket;
+			socket = NULL;
 		}
 	}
 	return false;
@@ -544,6 +583,7 @@ void RasHandler::ProcessRAS(RasMsg *ras)
 {
 	ras->Exec();
 	delete ras;
+	ras = NULL;
 }
 
 // class RasRequester
@@ -604,6 +644,7 @@ void RasRequester::Process(RasMsg *ras)
 		m_timeout = rip.m_delay;
 		m_sentTime = PTime();
 		delete ras;
+		ras = NULL;
 	} else {
 		AddReply(ras);
 		m_timeout = 0;
@@ -641,14 +682,16 @@ void RasRequester::AddReply(RasMsg *ras)
 
 
 // class RasServer
-RasServer::RasServer() : Singleton<RasServer>("RasServer"), requestSeqNum(0),
+RasServer::RasServer() : Singleton<RasServer>("RasSrv"), requestSeqNum(0),
 	acctList(NULL)
 {
-	SetName("RasServer");
+	SetName("RasSrv");
 
 	authList = 0;
 	sigHandler = 0;
 	broadcastListener = 0;
+	GKRoutedSignaling = false;
+	GKRoutedH245 = false;
 	altGKs = new H225_ArrayOf_AlternateGK;
 }
 
@@ -693,7 +736,7 @@ void RasServer::SetRoutedMode(bool routedSignaling, bool routedH245)
 	} else {
 		// warning: dangerous
 		delete sigHandler;
-		sigHandler = 0;
+		sigHandler = NULL;
 	}
 	GKRoutedH245 = GKRoutedSignaling ? routedH245 : false;
 
@@ -709,7 +752,7 @@ bool RasServer::AcceptUnregisteredCalls(const PIPSocket::Address & addr) const
 {
 	if (Toolkit::AsBool(GkConfig()->GetString(RoutedSec, "AcceptUnregisteredCalls", "0")))
 		return true;
-	return Toolkit::AsBool(GkConfig()->GetString(RoutedSec, "AcceptNeighborsCalls", "0")) ? neighbors->CheckIP(addr) : false;
+	return Toolkit::AsBool(GkConfig()->GetString(RoutedSec, "AcceptNeighborsCalls", "1")) ? neighbors->CheckIP(addr) : false;
 }
 
 bool RasServer::RegisterHandler(RasHandler *handler)
@@ -747,8 +790,8 @@ void RasServer::LoadConfig()
 	}
 
 	SocketsReader::CleanUp();
-	ifiterator biter = interfaces.begin(), eiter = interfaces.end();
-	while (biter != eiter) {
+	ifiterator biter = interfaces.begin();
+	while (biter != interfaces.end()) {
 		ifiterator iter = biter++;
 		int i = -1;
 		while (++i < hsize)
@@ -759,6 +802,7 @@ void RasServer::LoadConfig()
 			GkInterface * r = *iter;
 			interfaces.erase(iter);
 			delete r;
+			biter = interfaces.begin();
 		}
 	}
 	if (broadcastListener && !bUseBroadcastListener) {
@@ -770,20 +814,22 @@ void RasServer::LoadConfig()
 
 	for (int i = 0; i < hsize; ++i) {
 		Address addr(GKHome[i]);
-		biter = interfaces.begin(), eiter = interfaces.end();
-		ifiterator iter = find_if(biter, eiter, bind2nd(mem_fun(&GkInterface::IsBoundTo), &addr));
-		if (iter == eiter) {
+		biter = interfaces.begin();
+		ifiterator iter = find_if(biter, interfaces.end(), bind2nd(mem_fun(&GkInterface::IsBoundTo), &addr));
+		if (iter == interfaces.end()) {
 			GkInterface *gkif = CreateInterface(addr);
 			if (gkif->CreateListeners(this))
 				interfaces.push_back(gkif);
 			else
 				delete gkif;
+				gkif = NULL;
 		} else {
 			GkInterface *gkif = *iter;
 			// re-create if changed
 			if (!gkif->CreateListeners(this)) {
 				interfaces.erase(iter);
 				delete gkif;
+				gkif = NULL;
 			}
 		}
 	}
@@ -796,12 +842,12 @@ void RasServer::LoadConfig()
 	if (bUseBroadcastListener && !broadcastListener) {
 		broadcastListener = new BroadcastListener(interfaces.front()->GetRasPort());
 		if (broadcastListener->IsOpen()) {
-			PTRACE(1, "RAS\tAdd broadcast listener");
+			PTRACE(1, "RAS\tBroadcast listener listening at " << broadcastListener->GetName());
 			AddSocket(broadcastListener);
 		} else {
-			PTRACE(1, "RAS\tUnable to open broadcast listener");
+			PTRACE(1, "RAS\tCannot start broadcast listener at " << broadcastListener->GetName());
 			delete broadcastListener;
-			broadcastListener = 0;
+			broadcastListener = NULL;
 		}
 	}
 #endif
@@ -825,8 +871,10 @@ void RasServer::AddListener(TCPListenSocket *socket)
 {
 	if (socket->IsOpen())
 		listeners->AddListener(socket);
-	else
+	else {
 		delete socket;
+		socket = NULL;
+	}
 }
 
 bool RasServer::CloseListener(TCPListenSocket *socket)
@@ -854,7 +902,12 @@ const GkInterface *RasServer::SelectInterface(const Address & addr) const
 
 PIPSocket::Address RasServer::GetLocalAddress(const Address & addr) const
 {
-	return SelectInterface(addr)->GetRasListener()->GetLocalAddr(addr);
+	return SelectInterface(addr)->GetRasListener()->GetPhysicalAddr(addr);
+}
+
+PIPSocket::Address RasServer::GetMasqAddress(const Address & addr) const
+{
+	return SelectInterface(addr)->GetRasListener()->GetLocalAddr(addr);	
 }
 
 H225_TransportAddress RasServer::GetRasAddress(const Address & addr) const
@@ -1087,6 +1140,9 @@ void RasServer::Run()
 	if (m_socksize > 0) {
 		CreateJob(this, &RasServer::HouseKeeping, "HouseKeeping");
 		RegularJob::Run();
+	} else {
+		cerr << "FATAL: No valid interfaces to listen! Shutdown!" << endl;
+		PTRACE(0, "FATAL: No valid interfaces to listen! Shutdown!");
 	}
 	acctList->LogAcctEvent(GkAcctLogger::AcctOff,nullcall);
 }
@@ -1108,7 +1164,7 @@ void RasServer::OnStop()
 	listeners->Stop();
 
 	delete sigHandler;
-	sigHandler = 0;
+	sigHandler = NULL;
 
 	delete Routing::Analyzer::Instance();
 
@@ -1232,6 +1288,7 @@ void RasServer::ReadSocket(IPSocket *socket)
 					PTRACE(2, "RAS\tDuplicate " << msg->GetTagName() << ", deleted");
 //					(*i)->SetNext(ras);
 					delete ras;
+					ras = NULL;
 				} else {
 					requests.push_back(ras);
 					Job *job = new Jobs(ras);
@@ -1372,6 +1429,37 @@ bool RegistrationRequestPDU::Process()
 		}
 	}
 
+///////////////////////////////////////////////////////////////////////////////////////////
+// H460 support Code
+	BOOL supportNAT = false;
+	int RegPrior =0;
+	bool preemptsupport = false;
+	BOOL preempt = false;
+
+#ifdef hasH460
+
+// Registration Priority and Pre-emption
+// This allows the unregistration of duplicate aliases with lower priority 
+	OpalOID rPriFS = OpalOID(OID6);    
+
+	if (request.HasOptionalField(H225_RegistrationRequest::e_featureSet)) {
+		H460_FeatureSet fs = H460_FeatureSet(request.m_featureSet);
+
+		if (fs.HasFeature(rPriFS)) {
+			H460_FeatureOID * feat = (H460_FeatureOID *)fs.GetFeature(rPriFS);
+			if (feat->Contains(priorityOID)) {
+				unsigned prior = feat->Value(priorityOID);
+				RegPrior = (int)prior;  
+			}
+			if (feat->Contains(preemptOID)) {
+				preemptsupport = true;
+                preempt = feat->Value(preemptOID);
+			}
+		}
+	}
+#endif
+///////////////////////////////////////////////////////////////////////////////////////////////
+
 	// lightweight registration update
 	if (request.HasOptionalField(H225_RegistrationRequest::e_keepAlive) && request.m_keepAlive) {
 		endptr ep = request.HasOptionalField(H225_RegistrationRequest::e_endpointIdentifier) ?
@@ -1381,10 +1469,10 @@ bool RegistrationRequestPDU::Process()
 		bool bReject = !ep;
 		// check if the RRQ was sent from the registered endpoint
 		if (ep && bShellSendReply) { // not forwarded RRQ
-			if (ep->IsNATed())
+			if (ep->IsNATed()) {
 				// for nated endpoint, only check rx_addr
-			       	bReject = (ep->GetNATIP() != rx_addr);
-			else {
+			    bReject = (ep->GetNATIP() != rx_addr);
+			} else {
 				PIPSocket::Address oaddr, raddr;
 				WORD oport, rport;
 				if (request.m_callSignalAddress.GetSize() >= 1) {
@@ -1394,19 +1482,29 @@ bool RegistrationRequestPDU::Process()
 						if (oaddr == raddr && oport == rport)
 							break;
 					}
-				} else if (request.m_rasAddress.GetSize() >= 1)
+				} else if (request.m_rasAddress.GetSize() >= 1) {
 					GetIPAndPortFromTransportAddr(ep->GetRasAddress(), oaddr, oport),
 					GetIPAndPortFromTransportAddr(request.m_rasAddress[0], raddr, rport);
-				else
+				} else {
 					GetIPAndPortFromTransportAddr(ep->GetCallSignalAddress(), oaddr, oport),
 					raddr = oaddr, rport = oport;
-				bReject = (oaddr != raddr) || (oport != rport) || (raddr != rx_addr);
+				}
+				 bReject = (oaddr != raddr) || (oport != rport) || (raddr != rx_addr);
 			}
-		}
-
+		} 
 		if (bReject) {
-			PTRACE_IF(1, ep && bShellSendReply, "RAS\tWarning: Possibly endpointId collide or security attack!!");
-			// endpoint was NOT registered
+			if (ep && bShellSendReply) {
+			 PTRACE(1, "RAS\tWarning: Possibly endpointId collide,security attack or IP change");
+			   if (Toolkit::AsBool(Kit->Config()->GetString(RRQFeatureSection, "SupportDynamicIP", "0"))) {
+			     PTRACE(1, "RAS\tDynamic IP? Removing existing Endpoint record and force reregistration.");
+   						while (callptr call = CallTbl->FindCallRec(ep)) {
+							call->Disconnect();
+							CallTbl->RemoveCall(call);
+						}
+						EndpointTbl->RemoveByEndptr(ep);                              
+			   }
+			}
+			// endpoint was NOT registered and force Full Registration
 			return BuildRRJ(H225_RegistrationRejectReason::e_fullRegistrationRequired);
 		} else {
 			// forward lightweights, too
@@ -1447,6 +1545,12 @@ bool RegistrationRequestPDU::Process()
 	// or the request is sent from the original endpoint that has
 	// this EndpointIdentifier. Otherwise the request will be rejected.
 	if (request.HasOptionalField(H225_RegistrationRequest::e_endpointIdentifier)) {
+
+	// Alternate Gatekeepers based on rules
+//	   if (ResolveAlternateGatekeeper(request.m_endpointIdentifier,rx_addr))
+//		  return BuildRRJ(H225_RegistrationRejectReason::e_invalidRASAddress);
+
+
 		endptr ep = EndpointTbl->FindByEndpointId(request.m_endpointIdentifier);
 		if (ep && ep->GetCallSignalAddress() != SignalAddr)
 			// no reason named invalidEndpointIdentifier? :(
@@ -1477,18 +1581,42 @@ bool RegistrationRequestPDU::Process()
 
 			const endptr ep = EndpointTbl->FindByAliases(Alias);
 			if (ep) {
-				bNewEP = (ep->GetCallSignalAddress() != SignalAddr);
+				bNewEP = ((ep->GetCallSignalAddress() != SignalAddr) || (preempt) || ((RegPrior > 0) && (RegPrior > ep->Priority())));
 				if (bNewEP) {
-					if (Toolkit::AsBool(Kit->Config()->GetString("RasSrv::RRQFeatures", "OverwriteEPOnSameAddress", "0"))) {
+					if ((RegPrior > 0) || (preempt) ||
+					  (Toolkit::AsBool(Kit->Config()->GetString("RasSrv::RRQFeatures", "OverwriteEPOnSameAddress", "0")))) {
 						// If the operators policy allows this case:
-						// 1) unregister the active ep - sends URQ and
+						// 1a) terminate all calls on active ep and
+						// 1b) unregister the active ep - sends URQ and
 						// 2) remove the ep from the EndpointTable, then
 						// 3) allow the new ep to register - see below
-						ep->Unregister();
+						while (callptr call = CallTbl->FindCallRec(ep)) {
+							call->Disconnect();
+							CallTbl->RemoveCall(call);
+						}
+						if (RegPrior > ep->Priority())
+						   ep->Unregisterpreempt(1);   // Unregistered by high priority
+						else if (preempt)
+						   ep->Unregisterpreempt(2);   // Unregistered by preempt notification 
+						else
+						   ep->Unregister();
+
 						EndpointTbl->RemoveByEndptr(ep);
 					} else {
 						BuildRRJ(H225_RegistrationRejectReason::e_duplicateAlias);
 						H225_RegistrationReject & rrj = m_msg->m_replyRAS;
+#ifdef hasH460
+						// notify that EP can pre-empt previous registration
+						if ((preemptsupport) && (RegPrior == ep->Priority())) {  
+							rrj.IncludeOptionalField(H225_RegistrationReject::e_genericData);
+							H460_FeatureOID pre = H460_FeatureOID(rPriFS);
+                            pre.Add(preNotOID,H460_FeatureContent(TRUE));
+		                    H225_ArrayOf_GenericData & data = rrj.m_genericData;
+				               PINDEX lastPos = data.GetSize();
+				               data.SetSize(lastPos+1);
+				               data[lastPos] = pre;
+						}
+#endif
 						H225_ArrayOf_AliasAddress & duplicateAlias = rrj.m_rejectReason;
 						duplicateAlias = Alias;
 						return true;
@@ -1543,10 +1671,22 @@ bool RegistrationRequestPDU::Process()
 		return BuildRRJ(H225_RegistrationRejectReason::e_undefinedReason);
 	}
 
+#ifdef h323v6
+	if (request.HasOptionalField(H225_RegistrationRequest::e_assignedGatekeeper)) 
+             ep->SetAssignedGatekeeper(request.m_assignedGatekeeper);
+#endif
+
 	if (nated)
 		ep->SetNATAddress(rx_addr);
-	else
+	else {
 		ep->SetNAT(false);
+		if (supportNAT)
+		  ep->SetSupportNAT(true);
+	}
+
+	ep->SetPriority(RegPrior);
+	ep->SetPreemption(preemptsupport);
+
 	if (bShellSendReply) {
 		//
 		// OK, now send RCF
@@ -1802,6 +1942,34 @@ PString AdmissionRequestPDU::GetCalledStationId(
 	return id;
 }
 
+PString AdmissionRequestPDU::GetCallLinkage( 
+	/// additional data
+	ARQAuthData& authData
+	) const
+{
+		if (!authData.m_callLinkage)
+		return authData.m_callLinkage;
+
+	const H225_AdmissionRequest& arq = request;
+	const bool hasCall = authData.m_call.operator->() != NULL;
+	PString id;
+				
+	if (!arq.m_answerCall) {
+		if (arq.HasOptionalField(H225_AdmissionRequest::e_callLinkage)) {
+			const H225_CallLinkage & cl = arq.m_callLinkage;
+			if (cl.HasOptionalField(H225_CallLinkage::e_globalCallId))
+			   id = cl.m_globalCallId.AsString();
+		}
+	} else if (hasCall)
+		id = authData.m_call->GetCallLinkage();
+
+	if (!id)
+		return id;
+	else
+		return PString();  // No Call Linkage detected.
+
+}
+
 bool AdmissionRequestPDU::Process()
 {
 	// OnARQ
@@ -1876,6 +2044,7 @@ bool AdmissionRequestPDU::Process()
 
 	authData.m_callingStationId = GetCallingStationId(authData);
 	authData.m_calledStationId = GetCalledStationId(authData);
+	authData.m_callLinkage = GetCallLinkage(authData);
 	
 	if (!RasSrv->ValidatePDU(*this, authData)) {
 		if (authData.m_rejectReason < 0)
@@ -1912,10 +2081,10 @@ bool AdmissionRequestPDU::Process()
 	// Bandwidth
 	// and GkManager admission
 	//
-	int BWRequest = 1280;
+	int BWRequest = request.m_bandWidth.GetValue();
 	// hack for Netmeeting 3.0x
-	if (request.m_bandWidth.GetValue() >= 100)
-		BWRequest = request.m_bandWidth.GetValue();
+	if ((BWRequest > 0) && (BWRequest < 100))
+		BWRequest = 1280;
 	// check if it is the first arrived ARQ
 	if (pExistingCallRec) {
 		// request more bandwidth?
@@ -1941,23 +2110,42 @@ bool AdmissionRequestPDU::Process()
 	// routing decision
 	bool toParent = false;
 	H225_TransportAddress CalledAddress;
+	Routing::AdmissionRequest arq(request, this);
 	if (!answer) {
-		Routing::AdmissionRequest arq(request, this, CalledEP);
-		arq.SetProxyMode(authData.m_proxyMode);
-		if ((authData.m_routeToIP != NULL 
-				&& arq.SetDestination(*authData.m_routeToIP, true))
-			|| arq.Process() != NULL) {
-			CalledAddress = *arq.GetDestination();
-			toParent = arq.GetFlags() & Routing::AdmissionRequest::e_toParent;
-			aliasesChanged = aliasesChanged || (arq.GetFlags() & Routing::AdmissionRequest::e_aliasesChanged);
-	
-			// record neighbor used for rewriting purposes
-			if (arq.GetNeighborUsed() != 0) {
-				out_rewrite_source = RasSrv->GetNeighbors()->GetNeighborIdBySigAdr(arq.GetNeighborUsed());
-			}
+		if (!authData.m_destinationRoutes.empty()) {
+			list<Route>::const_iterator i = authData.m_destinationRoutes.begin();
+			while (i != authData.m_destinationRoutes.end())
+				arq.AddRoute(*i++);
 		} else
+			arq.Process();
+			
+		if (arq.GetRoutes().empty())
 			return BuildReply(arq.GetRejectReason());
-		authData.m_proxyMode = arq.GetProxyMode();
+			
+		list<Route>::iterator r = arq.GetRoutes().begin();
+		while (r != arq.GetRoutes().end()) {
+			// PTRACE(1, "JW route = " << r->AsString() );
+			if (authData.m_proxyMode != CallRec::ProxyDetect)
+				r->m_proxyMode = authData.m_proxyMode;
+			++r;
+		}
+		
+		Route route;
+		arq.GetFirstRoute(route);
+		
+		if (route.m_destEndpoint && !route.m_destEndpoint->HasAvailableCapacity())
+			return BuildReply(H225_AdmissionRejectReason::e_resourceUnavailable);
+
+		CalledEP = route.m_destEndpoint;
+		CalledAddress = route.m_destAddr;
+		toParent = route.m_flags & Route::e_toParent;
+		aliasesChanged = aliasesChanged || (arq.GetFlags() & Routing::AdmissionRequest::e_aliasesChanged);
+	
+		// record neighbor used for rewriting purposes
+		if (route.m_flags & Route::e_toNeighbor)
+			out_rewrite_source = route.m_routeId;
+		
+		authData.m_proxyMode = route.m_proxyMode;
 	}
 
 #ifdef ARJREASON_ROUTECALLTOSCN
@@ -2046,13 +2234,18 @@ bool AdmissionRequestPDU::Process()
 
 		if (CalledEP)
 			pCallRec->SetCalled(CalledEP);
-		else
+		else {
+			if (answer)
+				pCallRec->SetCalled(RequestingEP);
 			pCallRec->SetDestSignalAddr(CalledAddress);
+		}
 		if (!answer)
 			pCallRec->SetCalling(RequestingEP);
 		if (toParent)
 			pCallRec->SetToParent(true);
 
+		pCallRec->SetNewRoutes(arq.GetRoutes());
+		
 		if (authData.m_callDurationLimit > 0)
 			pCallRec->SetDurationLimit(authData.m_callDurationLimit);
 		if (!authData.m_callingStationId)
@@ -2085,8 +2278,11 @@ bool AdmissionRequestPDU::Process()
 		acf.m_destCallSignalAddress = CalledAddress;
 	}
 
-	acf.IncludeOptionalField(H225_AdmissionConfirm::e_irrFrequency);
-	acf.m_irrFrequency.SetValue(120);
+	long irrFrq = GkConfig()->GetInteger("CallTable", "IRRFrequency", 120);
+	if (irrFrq > 0) {
+		acf.IncludeOptionalField ( H225_AdmissionConfirm::e_irrFrequency );
+		acf.m_irrFrequency.SetValue( irrFrq );
+	}
 
 	if( !answer && aliasesChanged 
 		&& request.HasOptionalField(H225_AdmissionRequest::e_canMapAlias)
@@ -2115,33 +2311,51 @@ bool AdmissionRequestPDU::BuildReply(int reason)
 
 	PString log;
 	if (reason == e_routeRequest) {
-		log = PString(PString::Printf, "RouteRequest|%s|%s|%u|%s|%s;",
-			(const unsigned char *) source,
-			(const unsigned char *) RequestingEP->GetEndpointIdentifier().GetValue(),
-			(unsigned) request.m_callReferenceValue,
-			(const unsigned char *) destinationString,
-			(const unsigned char *) srcInfo
-		      );
+		log = PString(PString::Printf, "RouteRequest|%s|%s|%u|%s|%s",
+				(const unsigned char *) source,
+				(const unsigned char *) RequestingEP->GetEndpointIdentifier().GetValue(),
+				(unsigned) request.m_callReferenceValue,
+				(const unsigned char *) destinationString,
+				(const unsigned char *) srcInfo
+		      	);
+		if (Toolkit::AsBool(GkConfig()->GetString("Gatekeeper::Main", "SignalCallId", 0))) {
+			PString callid = AsString(request.m_callIdentifier.m_guid);
+			callid.Replace(" ", "-", true);
+			log += PString("|") + callid;
+		}
+		log += PString(";");
 	} else if (reason < 0) {
-		log = PString(PString::Printf, "ACF|%s|%s|%u|%s|%s|%s;",
-			(const unsigned char *) source,
-			(const unsigned char *) RequestingEP->GetEndpointIdentifier().GetValue(),
-			(unsigned) request.m_callReferenceValue,
-			(const unsigned char *) destinationString,
-			(const unsigned char *) srcInfo,
-			answerCall
-		      );
+		log = PString(PString::Printf, "ACF|%s|%s|%u|%s|%s|%s",
+				(const unsigned char *) source,
+				(const unsigned char *) RequestingEP->GetEndpointIdentifier().GetValue(),
+				(unsigned) request.m_callReferenceValue,
+				(const unsigned char *) destinationString,
+				(const unsigned char *) srcInfo,
+				answerCall
+		      	);
+		if (Toolkit::AsBool(GkConfig()->GetString("Gatekeeper::Main", "SignalCallId", 0))) {
+			PString callid = AsString(request.m_callIdentifier.m_guid);
+			callid.Replace(" ", "-", true);
+			log += PString("|") + callid;
+		}
+		log += PString(";");
 	} else {
 		H225_AdmissionReject & arj = BuildReject(reason);
 		if (reason == H225_AdmissionRejectReason::e_resourceUnavailable)
 			RasSrv->SetAltGKInfo(arj);
-		log = PString(PString::Printf, "ARJ|%s|%s|%s|%s|%s;",
-			(const unsigned char *) source,
-			(const unsigned char *) destinationString,
-			(const unsigned char *) srcInfo,
-			answerCall,
-			(const unsigned char *) arj.m_rejectReason.GetTagName()
-		      );
+		log = PString(PString::Printf, "ARJ|%s|%s|%s|%s|%s",
+				(const unsigned char *) source,
+				(const unsigned char *) destinationString,
+				(const unsigned char *) srcInfo,
+				answerCall,
+				(const unsigned char *) arj.m_rejectReason.GetTagName()
+		      	);
+		if (Toolkit::AsBool(GkConfig()->GetString("Gatekeeper::Main", "SignalCallId", 0))) {
+			PString callid = AsString(request.m_callIdentifier.m_guid);
+			callid.Replace(" ", "-", true);
+			log += PString("|") + callid;
+		}
+		log += PString(";");
 	}
 	return PrintStatus(log);
 }
@@ -2149,8 +2363,10 @@ bool AdmissionRequestPDU::BuildReply(int reason)
 template<> bool RasPDU<H225_BandwidthRequest>::Process()
 {
 	// OnBRQ
-	// hack for Netmeeting 3.0
-	int bandwidth = (request.m_bandWidth.GetValue() < 100) ? 1280 : int(request.m_bandWidth);
+	int bandwidth = request.m_bandWidth.GetValue();
+	// hack for Netmeeting 3.0x
+	if ((bandwidth > 0) && (bandwidth < 100))
+		bandwidth = 1280;
 
 	callptr pCall = request.HasOptionalField(H225_BandwidthRequest::e_callIdentifier) ?
 		CallTbl->FindCallRec(request.m_callIdentifier) :
@@ -2211,21 +2427,33 @@ template<> bool RasPDU<H225_DisengageRequest>::Process()
 	PString log;
 	if (bReject) {
 		H225_DisengageReject & drj = BuildReject(rsn);
-		log = PString(PString::Printf, "DRJ|%s|%s|%u|%s;",
-			inet_ntoa(m_msg->m_peerAddr),
-			(const unsigned char *) request.m_endpointIdentifier.GetValue(),
-			(unsigned) request.m_callReferenceValue,
-			(const unsigned char *) drj.m_rejectReason.GetTagName()
-		      );
+		log = PString(PString::Printf, "DRJ|%s|%s|%u|%s",
+				inet_ntoa(m_msg->m_peerAddr),
+				(const unsigned char *) request.m_endpointIdentifier.GetValue(),
+				(unsigned) request.m_callReferenceValue,
+				(const unsigned char *) drj.m_rejectReason.GetTagName()
+		      	);
+		if (Toolkit::AsBool(GkConfig()->GetString("Gatekeeper::Main", "SignalCallId", 0))) {
+			PString callid = AsString(request.m_callIdentifier.m_guid);
+			callid.Replace(" ", "-", true);
+			log += PString("|") + callid;
+		}
+		log += PString(";");
 	} else {
 		BuildConfirm();
 		// always signal DCF
-		log = PString(PString::Printf, "DCF|%s|%s|%u|%s;",
-			inet_ntoa(m_msg->m_peerAddr),
-			(const unsigned char *) request.m_endpointIdentifier.GetValue(),
-			(unsigned) request.m_callReferenceValue,
-			(const unsigned char *) request.m_disengageReason.GetTagName()
-		      );
+		log = PString(PString::Printf, "DCF|%s|%s|%u|%s",
+				inet_ntoa(m_msg->m_peerAddr),
+				(const unsigned char *) request.m_endpointIdentifier.GetValue(),
+				(unsigned) request.m_callReferenceValue,
+				(const unsigned char *) request.m_disengageReason.GetTagName()
+		      	);
+		if (Toolkit::AsBool(GkConfig()->GetString("Gatekeeper::Main", "SignalCallId", 0))) {
+			PString callid = AsString(request.m_callIdentifier.m_guid);
+			callid.Replace(" ", "-", true);
+			log += PString("|") + callid;
+		}
+		log += PString(";");
 
 		if (!RasSrv->IsGKRouted() || RasSrv->RemoveCallOnDRQ())
 			CallTbl->RemoveCall(request, ep);
@@ -2261,18 +2489,32 @@ template<> bool RasPDU<H225_LocationRequest>::Process()
 	if (request.HasOptionalField(H225_LocationRequest::e_endpointIdentifier))
 		if (endptr ep = EndpointTbl->FindByEndpointId(request.m_endpointIdentifier))
 			fromRegEndpoint = replyAddrMatch ? true : ep->IsNATed();
-	bool bReject = !((fromRegEndpoint || RasSrv->GetNeighbors()->CheckLRQ(this)) && RasSrv->ValidatePDU(*this, reason));
+
+    // Neighbors do not need Validation
+	bool bReject = !(fromRegEndpoint || RasSrv->GetNeighbors()->CheckLRQ(this));
+
+	// If not Neighbor and support non neighbor LRQ's then validate the PDU
+	if (bReject && (Toolkit::AsBool(GkConfig()->GetString(LRQFeaturesSection, "AcceptNonNeighborLRQ", "0"))))
+		                   bReject = !RasSrv->ValidatePDU(*this, reason);
+
 
 	PString sourceInfoString(fromRegEndpoint ? request.m_endpointIdentifier.GetValue() : request.HasOptionalField(H225_LocationRequest::e_gatekeeperIdentifier) ? request.m_gatekeeperIdentifier.GetValue() : m_msg->m_peerAddr.AsString());
 
 	if (!bReject) {
 		endptr WantedEndPoint;
-		Routing::LocationRequest lrq(request, this, WantedEndPoint);
-		if (H225_TransportAddress *dest = lrq.Process()) {
-			H225_LocationConfirm & lcf = BuildConfirm();
-			GetRasAddress(lcf.m_rasAddress);
-			if (RasSrv->IsGKRouted()) {
-				GetCallSignalAddress(lcf.m_callSignalAddress);
+		Route route;
+		Routing::LocationRequest lrq(request, this);
+		lrq.Process();
+		if (lrq.GetFirstRoute(route)) {
+			WantedEndPoint = route.m_destEndpoint;
+			if (WantedEndPoint && !WantedEndPoint->HasAvailableCapacity()) {
+				bReject = true;
+				reason = H225_LocationRejectReason::e_resourceUnavailable;
+			} else {
+				H225_LocationConfirm & lcf = BuildConfirm();
+				GetRasAddress(lcf.m_rasAddress);
+				if (RasSrv->IsGKRouted()) {
+					GetCallSignalAddress(lcf.m_callSignalAddress);
 /* The access token should be standarized somehow and use a correct object 
    identifier. As it does not (currently), we disable it to remove interop problems.
 				PINDEX s = 0;
@@ -2283,15 +2525,27 @@ template<> bool RasPDU<H225_LocationRequest>::Process()
 				lcf.m_cryptoTokens.SetSize(s + 1);
 				lcf.m_cryptoTokens[s] = Neighbors::BuildAccessToken(*dest, ipaddr);
 */
-			} else
-				lcf.m_callSignalAddress = *dest;
+				} else {
+					lcf.m_callSignalAddress = route.m_destAddr;
+				}
 
-			log = PString(PString::Printf, "LCF|%s|%s|%s|%s;",
-				inet_ntoa(m_msg->m_peerAddr),
-				(const unsigned char *) (WantedEndPoint ? WantedEndPoint->GetEndpointIdentifier().GetValue() : AsDotString(*dest)),
-				(const unsigned char *) AsString(request.m_destinationInfo),
-				(const unsigned char *) sourceInfoString
-			      );
+				// canMapAlias: include destinationInfo if it has been changed
+				if (lrq.GetFlags() & Routing::LocationRequest::e_aliasesChanged
+					&& request.HasOptionalField(H225_LocationRequest::e_canMapAlias)
+					&& request.m_canMapAlias) {
+					if (!lcf.HasOptionalField(H225_LocationConfirm::e_destinationInfo)) {
+						lcf.IncludeOptionalField(H225_LocationConfirm::e_destinationInfo);
+					}
+					lcf.m_destinationInfo = request.m_destinationInfo;
+				}
+
+				log = PString(PString::Printf, "LCF|%s|%s|%s|%s;",
+					inet_ntoa(m_msg->m_peerAddr),
+					(const unsigned char *) (WantedEndPoint ? WantedEndPoint->GetEndpointIdentifier().GetValue() : AsDotString(route.m_destAddr)),
+					(const unsigned char *) AsString(request.m_destinationInfo),
+					(const unsigned char *) sourceInfoString
+					);
+			}
 		} else {
 			if (m_msg->m_replyRAS.GetTag() == H225_RasMessage::e_requestInProgress) {
 				// LRQ is forwarded
@@ -2333,6 +2587,13 @@ template<> bool RasPDU<H225_InfoRequestResponse>::Process()
 	// OnIRR
 	if (endptr ep = EndpointTbl->FindByEndpointId(request.m_endpointIdentifier)) {
 		ep->Update(m_msg->m_recvRAS);
+		callptr call;
+		if (request.HasOptionalField(H225_InfoRequestResponse::e_perCallInfo))
+			call = CallTbl->FindCallRec(request.m_perCallInfo[0].m_callIdentifier);
+		else
+			call = CallTbl->FindBySignalAdr(request.m_callSignalAddress[0]);
+		if (call)
+			call->Update(request);
 		if (request.HasOptionalField(H225_InfoRequestResponse::e_needResponse) && request.m_needResponse) {
 			BuildConfirm();
 			PrintStatus(PString(PString::Printf, "IACK|%s;", inet_ntoa(m_msg->m_peerAddr)));
