@@ -37,6 +37,10 @@
 #include "gkacct.h"
 #include "RasTbl.h"
 
+#ifdef hasH460
+  #include <h4601.h>
+#endif
+
 using std::copy;
 using std::partition;
 using std::back_inserter;
@@ -49,13 +53,16 @@ using std::find_if;
 using std::distance;
 using std::sort;
 using std::string;
+using Routing::Route;
 
 const char *CallTableSection = "CallTable";
 const char *RRQFeaturesSection = "RasSrv::RRQFeatures";
+const char *proxysection = "Proxy";
 
 namespace {
-const long DEFAULT_SIGNAL_TIMEOUT = 15000;
+const long DEFAULT_SIGNAL_TIMEOUT = 30000;
 const long DEFAULT_ALERTING_TIMEOUT = 180000;
+const int DEFAULT_IRQ_POLL_COUNT = 1;
 }
 
 EndpointRec::EndpointRec(
@@ -66,10 +73,11 @@ EndpointRec::EndpointRec(
 	)
 	: m_RasMsg(ras), m_timeToLive(1),
 	m_activeCall(0), m_connectedCall(0), m_totalCall(0),
-	m_pollCount(2), m_usedCount(0),
-	m_nat(false), m_natsocket(0), m_permanent(permanent), 
+	m_pollCount(GkConfig()->GetInteger(RRQFeaturesSection, "IRQPollCount", DEFAULT_IRQ_POLL_COUNT)),
+	m_usedCount(0), m_nat(false), m_natsocket(NULL), m_permanent(permanent), 
 	m_hasCallCreditCapabilities(false), m_callCreditSession(-1),
-	m_capacity(-1)
+	m_capacity(-1), m_calledTypeOfNumber(-1), m_callingTypeOfNumber(-1), m_proxy(0),
+	m_registrationPriority(0), m_registrationPreemption(false)
 {
 	switch (m_RasMsg.GetTag())
 	{
@@ -159,9 +167,12 @@ void EndpointRec::SetEndpointRec(H225_LocationConfirm & lcf)
 
 EndpointRec::~EndpointRec()
 {
+	PWaitAndSignal lock(m_usedLock);
 	PTRACE(3, "Gk\tDelete endpoint: " << m_endpointIdentifier.GetValue() << " " << m_usedCount);
-	if (m_natsocket)
+	if (m_natsocket) {
 		m_natsocket->SetDeletable();
+		m_natsocket->Close();
+	}
 }
 
 bool EndpointRec::LoadConfig()
@@ -173,6 +184,8 @@ bool EndpointRec::LoadConfig()
 
 void EndpointRec::LoadEndpointConfig()
 {
+	const char* RoutedSec = "RoutedMode";
+	Toolkit* toolkit = Toolkit::Instance();
 	PConfig* const cfg = GkConfig();
 	const PStringList sections = cfg->GetSections();
 	
@@ -180,7 +193,21 @@ void EndpointRec::LoadEndpointConfig()
 		const PString key = "EP::" + AsString(m_terminalAliases[i], FALSE);
 		if (sections.GetStringsIndex(key) != P_MAX_INDEX) {
 			m_capacity = cfg->GetInteger(key, "Capacity", -1);
-			PTRACE(5, "RAS\tEndpoint " << key << " capacity: " << m_capacity);
+			int type = cfg->GetInteger(key, "CalledTypeOfNumber", -1);
+			if (type == -1)
+				m_calledTypeOfNumber = toolkit->Config()->GetInteger(RoutedSec, "CalledTypeOfNumber", -1);
+			type = cfg->GetInteger(key, "CallingTypeOfNumber", -1);
+			if (type == -1)
+				m_callingTypeOfNumber = toolkit->Config()->GetInteger(RoutedSec, "CallingTypeOfNumber", -1);
+			m_proxy = cfg->GetInteger(key, "Proxy", 0);
+			PString log;
+			if (m_calledTypeOfNumber > -1)
+				log += " Called Type Of Number: " +  PString(m_calledTypeOfNumber);
+			if (m_callingTypeOfNumber > -1)
+				log += " Calling Type Of Number: " + PString(m_callingTypeOfNumber);
+			if (m_proxy > 0)
+				log += " proxy: " + PString(m_proxy);
+			PTRACE(5, "RAS\tEndpoint " << key << " capacity: " << m_capacity << log);
 			break;
 		}
 	}
@@ -251,6 +278,10 @@ void EndpointRec::SetTimeToLive(int seconds)
 void EndpointRec::SetSocket(CallSignalSocket *socket)
 {
 	PWaitAndSignal lock(m_usedLock);
+
+	if (!socket->IsConnected())
+		return;
+
 	if (m_natsocket != socket) {
 		PTRACE(3, "Q931\tNAT socket detected at " << socket->Name() << " for endpoint " << GetEndpointIdentifier().GetValue());
 		if (m_natsocket) {
@@ -258,17 +289,12 @@ void EndpointRec::SetSocket(CallSignalSocket *socket)
 				<< " is overwritten by " << socket->Name()
 				);
 			m_natsocket->SetDeletable();
+			m_natsocket->Close();
 		}
 		m_natsocket = socket;
 	}
 }
 
-void EndpointRec::SetRasAddress(const H225_TransportAddress &a)
-{
-	PWaitAndSignal lock(m_usedLock);
-	m_rasAddress = a;
-}
-
 void EndpointRec::SetEndpointIdentifier(const H225_EndpointIdentifier &i)
 {
 	PWaitAndSignal lock(m_usedLock);
@@ -366,44 +392,28 @@ void EndpointRec::Update(const H225_RasMessage & ras_msg)
 	}
 	PWaitAndSignal lock(m_usedLock);
 	m_updatedTime = PTime();
-	m_pollCount = 2;
+	m_pollCount = GkConfig()->GetInteger(RRQFeaturesSection, "IRQPollCount", DEFAULT_IRQ_POLL_COUNT);
+}
+
+EndpointRec *EndpointRec::Unregisterpreempt(int type)
+{
+	PTRACE(1, "EP\tUnregistering " << AsDotString(GetRasAddress()) << " Reason " << type);
+	SendURQ(H225_UnregRequestReason::e_maintenance,type);
+	return this;
 }
 
 EndpointRec *EndpointRec::Unregister()
 {
-	SendURQ(H225_UnregRequestReason::e_maintenance);
+	SendURQ(H225_UnregRequestReason::e_maintenance,0);
 	return this;
 }
 
 EndpointRec *EndpointRec::Expired()
 {
-	SendURQ(H225_UnregRequestReason::e_ttlExpired);
+	SendURQ(H225_UnregRequestReason::e_ttlExpired,0);
 	return this;
 }
 
-/*
-void EndpointRec::BuildACF(H225_AdmissionConfirm & obj_acf) const
-{
-	obj_acf.IncludeOptionalField(H225_AdmissionConfirm::e_destinationInfo);
-	obj_acf.m_destinationInfo = GetAliases();
-	obj_acf.IncludeOptionalField(H225_AdmissionConfirm::e_destinationType);
-	obj_acf.m_destinationType = GetEndpointType();
-}
-
-void EndpointRec::BuildLCF(H225_LocationConfirm & obj_lcf) const
-{
-	obj_lcf.m_callSignalAddress = GetCallSignalAddress();
-	obj_lcf.m_rasAddress = GetRasAddress();
-	extern const char *LRQFeaturesSection;
-	if (Toolkit::AsBool(GkConfig()->GetString(LRQFeaturesSection, "IncludeDestinationInfoInLCF", "1"))) {
-		obj_lcf.IncludeOptionalField(H225_LocationConfirm::e_destinationInfo);
-		obj_lcf.m_destinationInfo = GetAliases();
-		obj_lcf.IncludeOptionalField(H225_LocationConfirm::e_destinationType);
-		obj_lcf.m_destinationType = GetEndpointType();
-	}
-}
-*/
-
 PString EndpointRec::PrintOn(bool verbose) const
 {
 	PString msg(PString::Printf, "%s|%s|%s|%s\r\n",
@@ -413,16 +423,16 @@ PString EndpointRec::PrintOn(bool verbose) const
 		    (const unsigned char *) GetEndpointIdentifier().GetValue() );
 	if (verbose) {
 		msg += GetUpdatedTime().AsString();
+		PWaitAndSignal lock(m_usedLock);
 		if (IsPermanent())
 			msg += " (permanent)";
 		PString natstring(IsNATed() ? m_natip.AsString() : PString());
-		PWaitAndSignal lock(m_usedLock);
 		msg += PString(PString::Printf, " C(%d/%d/%d) %s <%d>\r\n", m_activeCall, m_connectedCall, m_totalCall, (const unsigned char *)natstring, m_usedCount);
 	}
 	return msg;
 }
 
-bool EndpointRec::SendURQ(H225_UnregRequestReason::Choices reason)
+bool EndpointRec::SendURQ(H225_UnregRequestReason::Choices reason,  int preemption)
 {
 	if (GetRasAddress().GetTag() != H225_TransportAddress::e_ipAddress)
 		return false;  // no valid ras address
@@ -441,6 +451,22 @@ bool EndpointRec::SendURQ(H225_UnregRequestReason::Choices reason)
 	urq.IncludeOptionalField(H225_UnregistrationRequest::e_reason);
 	urq.m_reason.SetTag(reason);
 
+#ifdef hasH460
+	if (preemption > 0) {
+		urq.IncludeOptionalField(H225_UnregistrationRequest::e_genericData);
+		H460_FeatureOID pre = H460_FeatureOID(OpalOID(OID6));
+		if (preemption == 1)  // Higher Priority 
+           pre.Add(priNotOID,H460_FeatureContent(TRUE));          
+		else if (preemption == 2)  // Pre-empted
+           pre.Add(preNotOID,H460_FeatureContent(TRUE));
+
+		H225_ArrayOf_GenericData & data = urq.m_genericData;
+			PINDEX lastPos = data.GetSize();
+			data.SetSize(lastPos+1);
+			data[lastPos] = pre;
+	}
+#endif
+
 	PString msg(PString::Printf, "URQ|%s|%s|%s;\r\n", 
 			(const unsigned char *) AsDotString(GetRasAddress()),
 			(const unsigned char *) GetEndpointIdentifier().GetValue(),
@@ -456,9 +482,10 @@ bool EndpointRec::SendURQ(H225_UnregRequestReason::Choices reason)
 
 bool EndpointRec::SendIRQ()
 {
-	if (m_pollCount-- == 0 || GetRasAddress().GetTag() != H225_TransportAddress::e_ipAddress)
+	if (m_pollCount <= 0 || GetRasAddress().GetTag() != H225_TransportAddress::e_ipAddress)
 		return false;
-
+	--m_pollCount;
+	
 	RasServer *RasSrv = RasServer::Instance();
 	H225_RasMessage ras_msg;
 	ras_msg.SetTag(H225_RasMessage::e_infoRequest);
@@ -572,7 +599,7 @@ void GatewayRec::SetPriority(
 void GatewayRec::SetEndpointType(const H225_EndpointType &t)
 {
 	if (!t.HasOptionalField(H225_EndpointType::e_gateway)) {
-		PTRACE(1, "RRJ: terminal type changed|" << (const unsigned char *)m_endpointIdentifier.GetValue());
+		PTRACE(1, "RRJ: terminal type changed|" << GetEndpointIdentifier().GetValue());
 		return;
 	}
 	EndpointRec::SetEndpointType(t);
@@ -662,7 +689,7 @@ int GatewayRec::PrefixMatch(
 			const PString alias = AsString(aliases[i], FALSE);
 			// we also allow h_323_ID aliases consisting only from digits
 			if (tag == H225_AliasAddress::e_h323_ID)
-				if( strspn(alias,"1234567890*#") != strlen(alias) )
+				if( strspn(alias,"1234567890*#+,") != strlen(alias) )
 					continue;
 					
 			const_prefix_iterator Iter = Prefixes.begin();
@@ -684,11 +711,11 @@ int GatewayRec::PrefixMatch(
 	}
 	
 	if (maxlen < 0) {
-		PTRACE(2, "RASTBL\tGateway " << (const unsigned char *)m_endpointIdentifier.GetValue() 
+		PTRACE(2, "RASTBL\tGateway " << GetEndpointIdentifier().GetValue() 
 			<< " skipped by prefix " << pfxiter->c_str()
 			);
 	} else if (maxlen > 0) {
-		PTRACE(2, "RASTBL\tGateway " << (const unsigned char *)m_endpointIdentifier.GetValue()
+		PTRACE(2, "RASTBL\tGateway " << GetEndpointIdentifier().GetValue()
 			<< " matched by prefix " << pfxiter->c_str()
 			);
 		return maxlen;
@@ -701,7 +728,7 @@ int GatewayRec::PrefixMatch(
 				matchedalias = i;
 				break;
 			}
-		PTRACE(2, "RASTBL\tGateway " << (const unsigned char *)m_endpointIdentifier.GetValue()
+		PTRACE(2, "RASTBL\tGateway " << GetEndpointIdentifier().GetValue()
 			<< " matched as a default gateway"
 			);
 		return 0;
@@ -921,6 +948,8 @@ template<class F> endptr RegistrationTable::InternalFind(const F & FindObject,
 
 endptr RegistrationTable::FindByEndpointId(const H225_EndpointIdentifier & epId) const
 {
+	PWaitAndSignal m(findmutex);
+
 	return InternalFind(compose1(bind2nd(equal_to<H225_EndpointIdentifier>(), epId),
 			mem_fun(&EndpointRec::GetEndpointIdentifier)));
 }
@@ -973,8 +1002,27 @@ endptr RegistrationTable::FindByAliases(const H225_ArrayOf_AliasAddress & alias)
 
 endptr RegistrationTable::FindEndpoint(const H225_ArrayOf_AliasAddress & alias, bool r, bool s)
 {
-	endptr ep = InternalFindEP(alias, &EndpointList, r);
-	return (ep) ? ep : s ? InternalFindEP(alias, &OuterZoneList, r) : endptr(0);
+	list<H225_TransportAddress> emptyIgnoreList;
+	endptr ep = InternalFindEP(alias, &EndpointList, r, emptyIgnoreList);
+	return (ep) ? ep : s ? InternalFindEP(alias, &OuterZoneList, r, emptyIgnoreList) : endptr(0);
+}
+
+endptr RegistrationTable::FindEndpoint(const H225_ArrayOf_AliasAddress & alias, bool r, bool s, const list<H225_TransportAddress> & ignoreList)
+{
+	endptr ep = InternalFindEP(alias, &EndpointList, r, ignoreList);
+	return (ep) ? ep : s ? InternalFindEP(alias, &OuterZoneList, r, ignoreList) : endptr(0);
+}
+
+void RegistrationTable::FindEndpoint(
+	const H225_ArrayOf_AliasAddress &aliases,
+	bool roundRobin,
+	bool searchOuterZone,
+	list<Route> &routes
+	)
+{
+	InternalFindEP(aliases, &EndpointList, roundRobin, routes);
+	if (searchOuterZone)
+		InternalFindEP(aliases, &OuterZoneList, roundRobin, routes);
 }
 
 namespace {
@@ -989,8 +1037,9 @@ private:
 } /* namespace */
 
 endptr RegistrationTable::InternalFindEP(const H225_ArrayOf_AliasAddress & alias,
-	std::list<EndpointRec *> *List, bool roundrobin)
+	std::list<EndpointRec *> *List, bool roundrobin, const list<H225_TransportAddress> & ignoreList)
 {
+	// TODO: temporarily remove all aliases in ignoreList from List for this search
 	endptr ep = InternalFind(bind2nd(mem_fun(&EndpointRec::CompareAlias), &alias), List);
 	if (ep) {
 		PTRACE(4, "Alias match for EP " << AsDotString(ep->GetCallSignalAddress()));
@@ -1002,14 +1051,16 @@ endptr RegistrationTable::InternalFindEP(const H225_ArrayOf_AliasAddress & alias
 	listLock.StartRead();
 	const_iterator Iter = List->begin(), IterLast = List->end();
 	while (Iter != IterLast) {
-		if ((*Iter)->IsGateway()) {
+		if ((*Iter)->IsGateway() &&
+			find(ignoreList.begin(), ignoreList.end(), (*Iter)->GetCallSignalAddress()) == ignoreList.end() ) {	// not on ignoreList
 			int len = dynamic_cast<GatewayRec *>(*Iter)->PrefixMatch(alias);
 			if (maxlen < len) {
 				GWlist.clear();
 				maxlen = len;
 			}
-			if (maxlen == len)
+			if (maxlen == len) {
 				GWlist.push_back(GWPtr(dynamic_cast<GatewayRec*>(*Iter)));
+			}
 		}
 		++Iter;
 	}
@@ -1037,6 +1088,74 @@ endptr RegistrationTable::InternalFindEP(const H225_ArrayOf_AliasAddress & alias
 	return endptr(0);
 }
 
+void RegistrationTable::InternalFindEP(
+	const H225_ArrayOf_AliasAddress &aliases,
+	list<EndpointRec*> *endpoints,
+	bool roundRobin,
+	list<Route> &routes
+	)
+{
+	// TODO: temporarily remove all aliases in ignoreList from List for this search
+	endptr ep = InternalFind(bind2nd(mem_fun(&EndpointRec::CompareAlias), &aliases), endpoints);
+	if (ep) {
+		PTRACE(4, "Alias match for EP " << AsDotString(ep->GetCallSignalAddress()));
+		routes.push_back(Route(ep));
+        return;
+	}
+
+	int maxlen = 0;
+	std::list<GWPtr> GWlist;
+	listLock.StartRead();
+	const_iterator Iter = endpoints->begin(), IterLast = endpoints->end();
+	while (Iter != IterLast) {
+		if ((*Iter)->IsGateway()) {	// not on ignoreList
+			int len = dynamic_cast<GatewayRec *>(*Iter)->PrefixMatch(aliases);
+			if (maxlen < len) {
+				GWlist.clear();
+				maxlen = len;
+			}
+			if (maxlen == len) {
+				GWlist.push_back(GWPtr(dynamic_cast<GatewayRec*>(*Iter)));
+			}
+		}
+		++Iter;
+	}
+	listLock.EndRead();
+
+	if (!GWlist.empty()) {
+		GWlist.sort();
+		
+		std::list<GWPtr>::const_iterator i = GWlist.begin();
+		GatewayRec *e = GWlist.front().gwptr;
+		++i;
+		while (!e->HasAvailableCapacity() && i != GWlist.end()) {
+			PTRACE(5, "Capacity exceeded in GW " << AsDotString(e->GetCallSignalAddress()));
+			e = (i++)->gwptr;
+		}
+		if (GWlist.size() > 1 && roundRobin) {
+			PTRACE(3, "Prefix apply round robin");
+			WriteLock lock(listLock);
+			endpoints->remove(e);
+			endpoints->push_back(e);
+		}
+		routes.push_back(Route(endptr(e)));
+		while (i != GWlist.end())
+			routes.push_back(Route(endptr((*i++).gwptr)));
+#if PTRACING
+		if (PTrace::CanTrace(4)) {
+			ostream &strm = PTrace::Begin(4, __FILE__, __LINE__);
+			strm << "RASTBL\tPrefix match for gateways: ";
+			list<Route>::const_iterator r = routes.begin();
+			while (r != routes.end()) {
+				strm << endl << AsDotString(r->m_destAddr);
+				++r;
+			}
+			PTrace::End(strm);
+		}
+#endif
+	}
+}
+
 void RegistrationTable::GenerateEndpointId(H225_EndpointIdentifier & NewEndpointId)
 {
 	NewEndpointId = PString(PString::Unsigned, ++recCnt) + endpointIdSuffix;
@@ -1084,6 +1203,7 @@ void RegistrationTable::InternalPrint(USocket *client, BOOL verbose, std::list<E
 	for (k = 0; k < s; k++)
 		msg += "RCF|" + eptr[k]->PrintOn(verbose);
 	delete [] eptr;
+	eptr = NULL;
 
 	msg += PString(PString::Printf, "Number of Endpoints: %u\r\n;\r\n", s);
 	client->TransmitData(msg);
@@ -1154,12 +1274,13 @@ void RegistrationTable::LoadConfig()
 				EndpointList.erase(epIter);
 				--regSize;
 				PTRACE(2, "Permanent endpoint " << ep->GetEndpointIdentifier().GetValue() << " removed");
+				// TODO JW: check if we need to reset iterator here
 			}
 		}
 	}
 	
 	for (PINDEX i = 0; i < cfgs.GetSize(); ++i) {
-		EndpointRec *ep;
+		EndpointRec *ep = NULL;
 		H225_RasMessage rrq_ras;
 		rrq_ras.SetTag(H225_RasMessage::e_registrationRequest);
 		H225_RegistrationRequest &rrq = rrq_ras;
@@ -1259,6 +1380,7 @@ void RegistrationTable::CheckEndpoints()
 			ep->Expired();
 			RemovedList.push_back(ep);
 			EndpointList.erase(i);
+			// TODO JW: check if we need to reset iterator here
 			--regSize;
 			PTRACE(2, "Endpoint " << ep->GetEndpointIdentifier().GetValue() << " expired.");
 		}
@@ -1298,9 +1420,11 @@ CallRec::CallRec(
 	m_connectTime(0), m_disconnectTime(0), m_disconnectCause(0), m_releaseSource(-1),
 	m_acctSessionId(Toolkit::Instance()->GenerateAcctSessionId()),
 	m_routeToAlias(NULL), m_callingSocket(NULL), m_calledSocket(NULL),
-	m_usedCount(0), m_nattype(none), 
+	m_usedCount(0), m_nattype(none), m_unregNAT(false),
 	m_h245Routed(RasServer::Instance()->IsH245Routed()),
-	m_toParent(false), m_forwarded(false), m_proxyMode(proxyMode)
+	m_toParent(false), m_forwarded(false), m_proxyMode(proxyMode),
+	m_callInProgress(false), m_h245ResponseReceived(false), m_fastStartResponseReceived(false),
+	m_singleFailoverCDR(true), m_mediaOriginatingIp(INADDR_ANY)
 {
 	const H225_AdmissionRequest& arq = arqPdu;
 
@@ -1313,6 +1437,11 @@ CallRec::CallRec(
 	CallTable* const ctable = CallTable::Instance();
 	m_timeout = ctable->GetSignalTimeout() / 1000;
 	m_durationLimit = ctable->GetDefaultDurationLimit();
+	m_singleFailoverCDR = ctable->SingleFailoverCDR();
+
+	m_irrFrequency = GkConfig()->GetInteger(CallTableSection, "IRRFrequency", 120);
+	m_irrCheck = Toolkit::AsBool(GkConfig()->GetString(CallTableSection, "IRRCheck", "0"));
+	m_irrCallerTimer = m_irrCalleeTimer = time(NULL);
 }
 
 CallRec::CallRec(
@@ -1334,8 +1463,10 @@ CallRec::CallRec(
 	m_disconnectTime(0), m_disconnectCause(0), m_releaseSource(-1),
 	m_acctSessionId(Toolkit::Instance()->GenerateAcctSessionId()),
 	m_routeToAlias(NULL), m_callingSocket(NULL), m_calledSocket(NULL),
-	m_usedCount(0), m_nattype(none), m_h245Routed(routeH245),
-	m_toParent(false), m_forwarded(false), m_proxyMode(proxyMode)
+	m_usedCount(0), m_nattype(none), m_unregNAT(false), m_h245Routed(routeH245),
+	m_toParent(false), m_forwarded(false), m_proxyMode(proxyMode),
+	m_callInProgress(false), m_h245ResponseReceived(false), m_fastStartResponseReceived(false),
+	m_singleFailoverCDR(true), m_mediaOriginatingIp(INADDR_ANY)
 {
 	if (setup.HasOptionalField(H225_Setup_UUIE::e_sourceAddress)) {
 		m_sourceAddress = setup.m_sourceAddress;
@@ -1351,12 +1482,59 @@ CallRec::CallRec(
 	CallTable* const ctable = CallTable::Instance();
 	m_timeout = ctable->GetSignalTimeout() / 1000;
 	m_durationLimit = ctable->GetDefaultDurationLimit();
+	m_singleFailoverCDR = ctable->SingleFailoverCDR();
+
+	m_irrFrequency = GkConfig()->GetInteger(CallTableSection, "IRRFrequency", 120);
+	m_irrCheck = Toolkit::AsBool(GkConfig()->GetString(CallTableSection, "IRRCheck", "0"));
+	m_irrCallerTimer = m_irrCalleeTimer = time(NULL);
+}
+
+CallRec::CallRec(
+	CallRec *oldCall
+	) : m_CallNumber(0), 
+	m_callIdentifier(oldCall->m_callIdentifier),
+	m_conferenceIdentifier(oldCall->m_conferenceIdentifier), 
+	m_crv(oldCall->m_crv), m_Calling(oldCall->m_Calling),
+	m_sourceAddress(oldCall->m_sourceAddress),
+	m_destinationAddress(oldCall->m_destinationAddress),
+	m_srcInfo(oldCall->m_srcInfo),
+	m_destInfo(oldCall->m_destInfo), m_bandwidth(oldCall->m_bandwidth),
+	m_callerAddr(oldCall->m_callerAddr), m_callerId(oldCall->m_callerId),
+	m_inbound_rewrite_id(oldCall->m_inbound_rewrite_id),
+	m_setupTime(0), m_alertingTime(0), m_connectTime(0), m_disconnectTime(0),
+	m_disconnectCause(0), m_releaseSource(-1),
+	m_acctSessionId(Toolkit::Instance()->GenerateAcctSessionId()),
+	m_srcSignalAddress(oldCall->m_srcSignalAddress),
+	m_callingStationId(oldCall->m_callingStationId), m_calledStationId(oldCall->m_calledStationId),
+	m_dialedNumber(oldCall->m_dialedNumber),
+	m_routeToAlias(NULL), m_callingSocket(NULL /*oldCall->m_callingSocket*/), m_calledSocket(NULL),
+	m_usedCount(0), m_nattype(oldCall->m_nattype & ~calledParty), m_unregNAT(oldCall->m_unregNAT),
+	m_h245Routed(oldCall->m_h245Routed),
+	m_toParent(false), m_forwarded(false), m_proxyMode(CallRec::ProxyDetect),
+	m_failedRoutes(oldCall->m_failedRoutes), m_newRoutes(oldCall->m_newRoutes),
+	m_callInProgress(false), m_h245ResponseReceived(false), m_fastStartResponseReceived(false),
+	m_singleFailoverCDR(oldCall->m_singleFailoverCDR), m_mediaOriginatingIp(INADDR_ANY)
+{
+	m_timer = m_acctUpdateTime = m_creationTime = time(NULL);
+	m_calleeId = m_calleeAddr = " ";
+
+	CallTable* const ctable = CallTable::Instance();
+	m_timeout = ctable->GetSignalTimeout() / 1000;
+	m_durationLimit = oldCall->m_durationLimit;
+
+	m_irrFrequency = oldCall->m_irrFrequency;
+	m_irrCheck = oldCall->m_irrCheck;
+	m_irrCallerTimer = m_irrCalleeTimer = time(NULL);
+	
+	if (m_singleFailoverCDR)
+		m_setupTime = oldCall->m_setupTime;
 }
 
 CallRec::~CallRec()
 {
 	PTRACE(3, "Gk\tDelete Call No. " << m_CallNumber);
 	delete m_routeToAlias;
+	m_routeToAlias = NULL;
 }
 
 void CallRec::SetProxyMode(
@@ -1376,6 +1554,11 @@ bool CallRec::GetSrcSignalAddr(
 	return GetIPAndPortFromTransportAddr(m_srcSignalAddress, addr, port);
 }
 
+H225_TransportAddress CallRec::GetDestSignalAddr() const
+{
+	return m_destSignalAddress;
+}
+
 bool CallRec::GetDestSignalAddr(
 	PIPSocket::Address& addr, 
 	WORD& port
@@ -1395,6 +1578,8 @@ int CallRec::GetNATType(
 		callingPartyNATIP = m_Calling->GetNATIP();
 	if (m_nattype & calledParty)
 		calledPartyNATIP = m_Called->GetNATIP();
+	if (m_unregNAT)
+		callingPartyNATIP = m_srcunregNATAddress;
 	return m_nattype;
 }
 
@@ -1406,6 +1591,12 @@ void CallRec::SetSrcSignalAddr(
 	m_callerAddr = AsDotString(addr);
 }
 
+void CallRec::SetSrcNATed(PIPSocket::Address & natip)
+{
+	m_unregNAT = true;
+	m_srcunregNATAddress = natip;
+}
+
 void CallRec::SetDestSignalAddr(
 	const H225_TransportAddress& addr
 	)
@@ -1647,20 +1838,6 @@ void CallRec::SendDRQ()
 	}
 }
 
-namespace { // end of anonymous namespace
-
-PString GetEPString(const endptr & ep, const CallSignalSocket *socket)
-{
-	if (ep) {
-		return PString(PString::Printf, "%s|%s",
-			(const char *)AsDotString(ep->GetCallSignalAddress()),
-			(const char *)ep->GetEndpointIdentifier().GetValue());
-	}
-	return socket ? socket->Name() + "| " : PString(" | ");
-}
-
-} // end of anonymous namespace
-
 PString CallRec::GenerateCDR(
 	const PString& timestampFormat
 	) const
@@ -1734,13 +1911,13 @@ PString CallRec::PrintOn(bool verbose) const
 void CallRec::SetSetupTime(time_t tm)
 {
 	PWaitAndSignal lock(m_usedLock);
-	if( m_setupTime == 0 )
+	if (m_setupTime == 0)
 		m_setupTime = tm;
-	if( m_connectTime == 0 )
+	if (m_connectTime == 0)
 		m_timer = tm;
 	// can be the case, because CallRec is usually created
 	// after Setup message has been received
-	if( m_creationTime > m_setupTime )
+	if (m_creationTime > m_setupTime)
 		m_creationTime = m_setupTime;
 }
 
@@ -1918,6 +2095,151 @@ void CallRec::SetRouteToAlias(
 	m_routeToAlias = new H225_AliasAddress(alias);
 }
 
+void CallRec::Update(const H225_InfoRequestResponse & irr)
+{
+	if (irr.HasOptionalField(H225_InfoRequestResponse::e_perCallInfo) &&
+		irr.m_perCallInfo[0].HasOptionalField(H225_InfoRequestResponse_perCallInfo_subtype::e_originator)) {
+		if (irr.m_perCallInfo[0].m_originator)
+			m_irrCallerTimer = time(NULL);
+		else
+			m_irrCalleeTimer = time(NULL);
+	} else {
+		if (irr.m_endpointIdentifier == m_Calling->GetEndpointIdentifier())
+			m_irrCallerTimer = time(NULL);
+		else
+			m_irrCalleeTimer = time(NULL);
+	}
+}
+
+void CallRec::SetNewRoutes(
+	const std::list<Routing::Route> &routes
+	)
+{
+	m_newRoutes = routes;
+}
+
+void CallRec::SetFailedRoutes(
+	const std::list<Routing::Route> &routes
+	)
+{
+	m_failedRoutes = routes;
+}
+
+bool CallRec::MoveToNextRoute()
+{
+	if (! Toolkit::AsBool(GkConfig()->GetString(RoutedSec, "ActivateFailover", "0")))
+		return false;
+
+	if (!m_newRoutes.empty()) {
+		m_failedRoutes.push_back(m_newRoutes.front());
+		m_newRoutes.pop_front();
+	}
+	return !m_newRoutes.empty();
+}
+
+bool CallRec::IsCallInProgress() const
+{
+	return m_callInProgress;
+}
+
+void CallRec::SetCallInProgress()
+{
+	m_callInProgress = true;
+}
+
+bool CallRec::IsH245ResponseReceived() const
+{
+	return m_h245ResponseReceived;
+}
+
+void CallRec::SetH245ResponseReceived()
+{
+	m_h245ResponseReceived = true;
+}
+	
+bool CallRec::IsFastStartResponseReceived() const
+{
+	return m_fastStartResponseReceived;
+}
+
+void CallRec::SetFastStartResponseReceived()
+{
+	m_fastStartResponseReceived = true;
+}
+
+bool CallRec::SingleFailoverCDR() const
+{
+	return m_singleFailoverCDR;
+}
+
+int CallRec::GetNoCallAttempts() const
+{
+	int attempts = m_failedRoutes.size();
+	if (m_newRoutes.size() > 0)
+		attempts += 1;
+	return attempts;
+}
+
+int CallRec::GetNoRemainingRoutes() const
+{
+	return m_newRoutes.size();
+}
+
+bool CallRec::DisableRetryChecks() const
+{
+	return Toolkit::AsBool(GkConfig()->GetString(RoutedSec, "DisableRetryChecks", "0"));
+}
+
+void CallRec::SetCodec(
+	const PString &codec
+	)
+{
+	PWaitAndSignal lock(m_usedLock);
+	m_codec = codec;
+}
+
+PString CallRec::GetCodec() const
+{
+	PWaitAndSignal lock(m_usedLock);
+	return m_codec;
+}
+
+void CallRec::SetMediaOriginatingIp(
+	const PIPSocket::Address &addr
+	)
+{
+	PWaitAndSignal lock(m_usedLock);
+	m_mediaOriginatingIp = addr;
+}
+
+bool CallRec::GetMediaOriginatingIp(PIPSocket::Address &addr) const
+{
+	PWaitAndSignal lock(m_usedLock);
+	if (m_mediaOriginatingIp.IsValid()) {
+		addr = m_mediaOriginatingIp;
+		return true;
+	} else
+		return false;
+}
+
+void CallRec::SetRADIUSClass(const PBYTEArray &bytes)
+{
+	PWaitAndSignal lock(m_usedLock);
+	m_radiusClass = bytes;
+}
+
+void CallRec::SetRADIUSClass(void * bytes, PINDEX len)
+{
+	PWaitAndSignal lock(m_usedLock);
+	m_radiusClass = PBYTEArray(static_cast<const BYTE*>(bytes), len);
+}
+
+PBYTEArray CallRec::GetRADIUSClass() const
+{
+	PWaitAndSignal lock(m_usedLock);
+	return m_radiusClass;
+}
+
 /*
 bool CallRec::IsTimeout(
 	const time_t now,
@@ -1992,6 +2314,7 @@ void CallTable::LoadConfig()
 		m_acctUpdateInterval = std::max(m_acctUpdateInterval, 10L);
 		
 	m_timestampFormat = GkConfig()->GetString(CallTableSection, "TimestampFormat", "RFC822");
+	m_singleFailoverCDR = Toolkit::AsBool(GkConfig()->GetString(CallTableSection, "SingleFailoverCDR", "1"));
 }
 
 void CallTable::Insert(CallRec * NewRec)
@@ -2063,13 +2386,14 @@ callptr CallTable::FindBySignalAdr(const H225_TransportAddress & SignalAdr) cons
 void CallTable::ClearTable()
 {
 	WriteLock lock(listLock);
-	iterator Iter = CallList.begin(), eIter = CallList.end();
-	while (Iter != eIter) {
+	iterator Iter = CallList.begin();
+	while (Iter != CallList.end()) {
 		iterator i = Iter++;
 		(*i)->SetDisconnectCause(Q931::TemporaryFailure);
 		(*i)->SetReleaseSource(CallRec::ReleasedByGatekeeper);
 		(*i)->Disconnect();
 		InternalRemove(i);
+		Iter = CallList.begin(); // reset invalidated iterator
 	}
 }
 
@@ -2157,6 +2481,16 @@ bool CallTable::InternalRemovePtr(CallRec *call)
 	return true; // useless, workaround for VC
 }
 
+void CallTable::RemoveFailedLeg(const callptr & call)
+{
+	if (call) {
+		CallRec *callrec = call.operator->();
+		PTRACE(6, "GK\tRemoving callptr: " << AsString(call->GetCallIdentifier().m_guid));
+		WriteLock lock(listLock);
+		InternalRemoveFailedLeg(find(CallList.begin(), CallList.end(), callrec));
+	}
+}
+
 void CallTable::InternalRemove(const H225_CallIdentifier & CallId)
 {
 	PTRACE(5, "GK\tRemoving CallId: " << AsString(CallId.m_guid));
@@ -2218,6 +2552,50 @@ void CallTable::InternalRemove(iterator Iter)
 	call->RemoveSocket();
 }
 
+void CallTable::InternalRemoveFailedLeg(iterator Iter)
+{
+	if (Iter == CallList.end()) {
+		return;
+	}
+
+	callptr call(*Iter);
+	call->SetDisconnectTime(time(NULL));
+
+	--m_activeCall;
+	
+	if (m_capacity >= 0)
+		m_capacity += call->GetBandwidth();
+
+	CallList.erase(Iter);
+	RemovedList.push_back(call.operator->());
+
+	WriteUnlock unlock(listLock);
+
+	if (call->SingleFailoverCDR() && !call->GetNewRoutes().empty())	{
+		PTRACE(2, "CDR\tIgnoring failed call leg");
+	} else {
+		if ((m_genNBCDR || call->GetCallingParty()) && (m_genUCCDR || call->IsConnected())) {
+			PString cdrString(call->GenerateCDR(m_timestampFormat) + "\r\n");
+			GkStatus::Instance()->SignalStatus(cdrString, STATUS_TRACE_LEVEL_CDR);
+			PTRACE(1, cdrString);
+#if PTRACING
+		} else {
+			if (!call->IsConnected())
+				PTRACE(2, "CDR\tignore not connected call");
+			else	
+				PTRACE(2, "CDR\tignore caller from neighbor");
+#endif
+		}
+
+		RasServer::Instance()->LogAcctEvent(GkAcctLogger::AcctStop, call);
+	}
+	
+	if (call->GetCalledParty())
+		call->GetCalledParty()->RemoveCall();
+	
+	call->SetSocket(NULL, NULL);
+}
+
 void CallTable::InternalStatistics(unsigned & n, unsigned & act, unsigned & nb, unsigned & np, PString & msg, BOOL verbose) const
 {
 	ReadLock lock(listLock);
