@@ -15,24 +15,36 @@
 //////////////////////////////////////////////////////////////////
 
 #ifndef RASTBL_H
-#define RASTBL_H "@(#) $Id: RasTbl.h,v 1.89 2005/03/18 14:54:47 zvision Exp $"
+#define RASTBL_H "@(#) $Id: RasTbl.h,v 1.107 2006/12/06 16:34:04 zvision Exp $"
 
 #include <list>
 #include <vector>
 #include <string>
 #include "rwlock.h"
 #include "singleton.h"
+#include "h225.h"
+#include "sigmsg.h"
+#include "h323util.h"
+#include "pwlib_compat.h"
+
 
 #if (_MSC_VER >= 1200)
 #pragma warning( disable : 4786 ) // warning about too long debug symbol off
 #pragma warning( disable : 4800 )
 #endif
 
+namespace Routing {
+	struct Route;
+}
+
 class GkDestAnalysisList;
 class USocket;
 class CallSignalSocket;
 class RasServer;
 class Q931;
+class SignalingMsg;
+template <class> class H225SignalingMsg;
+typedef H225SignalingMsg<H225_Setup_UUIE> SetupMsg;
 
 // Template of smart pointer
 // The class T must have Lock() & Unlock() methods
@@ -86,6 +98,9 @@ public:
 	PIPSocket::Address GetNATIP() const;
 	CallSignalSocket *GetSocket();
 
+	int GetCallTypeOfNumber(bool called = true) const { return called ? m_calledTypeOfNumber : m_callingTypeOfNumber; }
+	int GetProxyType() const { return m_proxy; }
+
 	/** checks if the given aliases are prefixes of the aliases which are stored
 	    for the endpoint in the registration table. #fullMatch# returns #TRUE# if
 	    a full match is found.
@@ -95,6 +110,7 @@ public:
 	                                bool &fullMatch) const;
 
 	virtual void SetRasAddress(const H225_TransportAddress &);
+	virtual void SetCallSignalAddress(const H225_TransportAddress &);
 	virtual void SetEndpointIdentifier(const H225_EndpointIdentifier &);
 	virtual void SetTimeToLive(int);
 	virtual void SetAliases(const H225_ArrayOf_AliasAddress &);
@@ -135,6 +151,7 @@ public:
 	*/
 	virtual bool LoadConfig();
 
+	virtual EndpointRec *Unregisterpreempt(int type);
 	virtual EndpointRec *Unregister();
 	virtual EndpointRec *Expired();
 
@@ -146,6 +163,11 @@ public:
 	void SetNAT(bool nat);
 	void SetNATAddress(const PIPSocket::Address &);
 	void SetSocket(CallSignalSocket *);
+	void SetSupportNAT(bool support);
+	void SetPriority(int priority) { m_registrationPriority = priority; };
+	void SetPreemption(bool support) { m_registrationPreemption = support; };
+	void SetAssignedGatekeeper(const H225_AlternateGK & gk) { m_assignedGatekeeper = gk; };
+
 
 	/** @return
 		true if this is a permanent endpoint loaded from the config file entry.
@@ -155,6 +177,8 @@ public:
 	bool IsUpdated(const PTime *) const;
 	bool IsFromParent() const;
 	bool IsNATed() const;
+	bool SupportNAT() const;
+	int  Priority() const { return m_registrationPriority; }
 	bool HasNATSocket() const;
 	PTime GetUpdatedTime() const;
 
@@ -214,7 +238,7 @@ protected:
 	void SetEndpointRec(H225_AdmissionConfirm &);
 	void SetEndpointRec(H225_LocationConfirm &);
 
-	bool SendURQ(H225_UnregRequestReason::Choices);
+	bool SendURQ(H225_UnregRequestReason::Choices, int preemption);
 
 private:
 	/// Load general endpoint settings from the config
@@ -243,7 +267,7 @@ protected:
 	mutable PMutex m_usedLock;
 
 	PTime m_updatedTime;
-	bool m_fromParent, m_nat;
+	bool m_fromParent, m_nat, m_natsupport;
 	PIPSocket::Address m_natip;
 	CallSignalSocket *m_natsocket;
 	/// permanent (preconfigured) endpoint flag
@@ -254,6 +278,16 @@ protected:
 	int m_callCreditSession;
 	/// endpoint call capacity, -1 means no limit
 	int m_capacity;
+	int m_calledTypeOfNumber, m_callingTypeOfNumber;
+	/// Proxy Type
+	int m_proxy;
+	/// Registration Priority Number
+	int m_registrationPriority;
+	/// Support Registration PreEmption
+	bool m_registrationPreemption;
+
+    /// Assigned Gatekeeper
+	H225_AlternateGK m_assignedGatekeeper;
 };
 
 typedef EndpointRec::Ptr endptr;
@@ -376,19 +410,13 @@ public:
 	endptr FindOZEPBySignalAdr(const H225_TransportAddress &) const;
 	endptr FindByAliases(const H225_ArrayOf_AliasAddress & alias) const;
 	endptr FindEndpoint(const H225_ArrayOf_AliasAddress & alias, bool RoundRobin, bool SearchOuterZone = true);
-
-	template<class MsgType> endptr getMsgDestination(const MsgType & msg, unsigned int & reason,
-	                                                 bool SearchOuterZone = true)
-	{
-	  endptr ep;
-	  bool ok = getGkDestAnalysisList().getMsgDestination(msg, EndpointList, listLock,
-	                                                      ep, reason);
-	  if (!ok && SearchOuterZone) {
-            ok = getGkDestAnalysisList().getMsgDestination(msg, OuterZoneList, listLock,
-	                                                   ep, reason);
-	  }
-	  return (ok) ? ep : endptr(0);
-	}
+	endptr FindEndpoint(const H225_ArrayOf_AliasAddress & alias, bool RoundRobin, bool SearchOuterZone, const std::list<H225_TransportAddress> & ignoreList);
+	void FindEndpoint(
+		const H225_ArrayOf_AliasAddress &aliases,
+		bool roundRobin,
+		bool searchOuterZone,
+		std::list<Routing::Route> &routes
+		);
 
 	void ClearTable();
 	void CheckEndpoints();
@@ -433,7 +461,8 @@ private:
 	        return endptr((Iter != ListToBeFound->end()) ? *Iter : 0);
 	}
 
-	endptr InternalFindEP(const H225_ArrayOf_AliasAddress & alias, std::list<EndpointRec *> *ListToBeFound, bool);
+	endptr InternalFindEP(const H225_ArrayOf_AliasAddress & alias, std::list<EndpointRec *> *ListToBeFound, bool roundrobin, const std::list<H225_TransportAddress> & ignoreList);
+	void InternalFindEP(const H225_ArrayOf_AliasAddress & alias, std::list<EndpointRec *> *ListToBeFound, bool roundrobin, std::list<Routing::Route> &routes);
 
 	void GenerateEndpointId(H225_EndpointIdentifier &);
 	void GenerateAlias(H225_ArrayOf_AliasAddress &, const H225_EndpointIdentifier &) const;
@@ -444,6 +473,7 @@ private:
 	std::list<EndpointRec *> RemovedList;
 	int regSize;
 	mutable PReadWriteMutex listLock;
+	PMutex findmutex;          // Endpoint Find Mutex
 	GkDestAnalysisList * m_destAnalysisList;
 
 	// counter to generate endpoint identifier
@@ -501,6 +531,10 @@ public:
 		/// override proxy mode global setting from the config
 		int proxyMode = ProxyDetect
 		);
+	
+	CallRec(
+		CallRec *oldCall
+		);
 		
 	virtual ~CallRec();
 
@@ -772,6 +806,10 @@ public:
 		const H225_TransportAddress & addr /// new signalling transport address
 		);
 
+	/** Set the unregistered calling party signalling channel as NATed.
+	*/
+	void SetSrcNATed(PIPSocket::Address & natip);
+
 	/** Set a new address for the called party signalling channel.
 	*/
 	void SetDestSignalAddr(
@@ -790,6 +828,8 @@ public:
 		WORD& port /// will receive the port number
 		) const;
 
+	H225_TransportAddress GetDestSignalAddr() const;
+
 	/** Get IP and port for the called party. It is a signal address
 		for registered endpoints and remote signalling socket address
 		for unregistered endpoints.
@@ -845,6 +885,17 @@ public:
 	*/
 	PString GetCalledStationId();
 
+	/// Set call linkage This the party to be charged for the call. 
+	void SetCallLinkage(
+		const PString& id /// Calling-Station-Id (to be charged)
+		);
+
+	/** @return
+        Call party to be charged for the call.
+	*/
+	PString GetCallLinkage()const { return m_callLinkage; };
+
+
 	/// Set calling party's number
 	void SetCalledStationId(
 		const PString& id /// Called-Station-Id
@@ -874,6 +925,43 @@ public:
 	// smart pointer for CallRec
 	typedef SmartPtr<CallRec> Ptr;
 
+	/// update IRR timers
+	void Update(const H225_InfoRequestResponse & irr);
+
+	void SetNewRoutes(
+		const std::list<Routing::Route> &routes
+		);
+	void SetFailedRoutes(
+		const std::list<Routing::Route> &routes
+		);
+	const std::list<Routing::Route> &GetNewRoutes() const { return m_newRoutes; }
+	const std::list<Routing::Route> &GetFailedRoutes() const { return m_failedRoutes; }
+	bool MoveToNextRoute();
+
+	bool IsCallInProgress() const;
+	void SetCallInProgress();
+
+	bool IsH245ResponseReceived() const;
+	void SetH245ResponseReceived();
+	
+	bool IsFastStartResponseReceived() const;
+	void SetFastStartResponseReceived();
+
+	bool SingleFailoverCDR() const;
+	int GetNoCallAttempts() const;
+	int GetNoRemainingRoutes() const;
+	bool DisableRetryChecks() const;
+
+	void SetCodec(const PString &codec);
+	PString GetCodec() const;
+
+	void SetMediaOriginatingIp(const PIPSocket::Address &addr);
+	bool GetMediaOriginatingIp(PIPSocket::Address &addr) const;
+
+	void SetRADIUSClass(const PBYTEArray &bytes);
+	void SetRADIUSClass(void * bytes, PINDEX len);
+	PBYTEArray GetRADIUSClass() const;
+
 private:
 	void SendDRQ();
 	void InternalSetEP(endptr &, const endptr &);
@@ -948,6 +1036,8 @@ private:
 	PString m_callingStationId;
 	/// called party's number
 	PString m_calledStationId;
+	/// party to be charged for the call
+	PString m_callLinkage;
 	/// dialed number (called party's number before rewrite)
 	PString m_dialedNumber;
 	/// fixed destination alias
@@ -956,9 +1046,13 @@ private:
 	CallSignalSocket *m_callingSocket, *m_calledSocket;
 
 	int m_usedCount;
-	mutable PMutex m_usedLock, m_sockLock;
+	mutable PTimedMutex m_usedLock, m_sockLock;
 	int m_nattype;
 
+	/// unregistered caller NAT'd
+	bool m_unregNAT;
+	PIPSocket::Address m_srcunregNATAddress;
+
 	bool m_h245Routed;
 	/// the call is routed to this gatekeeper's parent gatekeeper
 	bool m_toParent;
@@ -969,6 +1063,23 @@ private:
 	int m_proxyMode;
 	
 	H225_ArrayOf_CryptoH323Token m_accessTokens;
+
+	/// IRR checking
+	long m_irrFrequency;
+	bool m_irrCheck;
+	time_t m_irrCallerTimer;
+	time_t m_irrCalleeTimer;
+	
+	std::list<Routing::Route> m_failedRoutes;
+	std::list<Routing::Route> m_newRoutes;
+	bool m_callInProgress;
+	bool m_h245ResponseReceived;
+	bool m_fastStartResponseReceived;
+	bool m_singleFailoverCDR;
+	
+	PString m_codec;
+	PIPSocket::Address m_mediaOriginatingIp;
+	PBYTEArray m_radiusClass;
 };
 
 typedef CallRec::Ptr callptr;
@@ -1004,6 +1115,7 @@ public:
 
 	void RemoveCall(const H225_DisengageRequest & obj_drq, const endptr &);
 	void RemoveCall(const callptr &);
+	void RemoveFailedLeg(const callptr &);
 
 	void PrintCurrentCalls(USocket *client, BOOL verbose=FALSE) const;
 	PString PrintStatistics() const;
@@ -1033,6 +1145,9 @@ public:
 	*/
 	long GetDefaultDurationLimit() const { return m_defaultDurationLimit; }
 
+	/// @return	True to log accounting for each call leg
+	bool SingleFailoverCDR() const { return m_singleFailoverCDR; }
+
 private:
 	template<class F> callptr InternalFind(const F & FindObject) const
 	{
@@ -1045,6 +1160,7 @@ private:
 	void InternalRemove(const H225_CallIdentifier & CallId);
 	void InternalRemove(WORD CallRef);
 	void InternalRemove(iterator);
+	void InternalRemoveFailedLeg(iterator);
 
 	void InternalStatistics(unsigned & n, unsigned & act, unsigned & nb, unsigned & np, PString & msg, BOOL verbose) const;
 
@@ -1074,7 +1190,9 @@ private:
 	long m_acctUpdateInterval;
 	/// timestamp formatting string for CDRs
 	PString m_timestampFormat;
-	
+	/// flag to trigger per call leg accounting
+	bool m_singleFailoverCDR;
+
 	CallTable(const CallTable &);
 	CallTable& operator==(const CallTable &);
 };
@@ -1086,6 +1204,18 @@ inline H225_TransportAddress EndpointRec::GetRasAddress() const
 	return m_rasAddress;
 }
 
+inline void EndpointRec::SetRasAddress(const H225_TransportAddress & addr)
+{ 
+	PWaitAndSignal lock(m_usedLock);
+	m_rasAddress = addr;
+}
+
+inline void EndpointRec::SetCallSignalAddress(const H225_TransportAddress & addr) 
+{
+	PWaitAndSignal lock(m_usedLock);
+    m_callSignalAddress = addr;
+}
+
 inline H225_TransportAddress EndpointRec::GetCallSignalAddress() const
 { 
 	PWaitAndSignal lock(m_usedLock);
@@ -1120,11 +1250,21 @@ inline void EndpointRec::SetNAT(bool nat)
 	m_nat = nat;
 }
 
+inline void EndpointRec::SetSupportNAT(bool support)
+{
+	m_natsupport = support;
+}
+
 inline bool EndpointRec::IsNATed() const
 { 
 	return m_nat;
 }
 
+inline bool EndpointRec::SupportNAT() const
+{
+	return m_natsupport;
+}
+
 inline PIPSocket::Address EndpointRec::GetNATIP() const
 {
 	PWaitAndSignal lock(m_usedLock);
@@ -1296,7 +1436,14 @@ inline void CallRec::SetDisconnectCause( unsigned causeCode )
 
 inline bool CallRec::IsTimeout(const time_t now) const
 {
-	return m_timeout > 0 && now >= m_timer && (now - m_timer) >= m_timeout;
+	bool result = (m_timeout > 0) && (now >= m_timer) && ((now - m_timer) >= m_timeout);
+	if (m_irrCheck && (m_irrFrequency > 0)) {
+		if (m_Calling)
+			result |= (now >= m_irrCallerTimer) && ((now - m_irrCallerTimer) >= 2 * m_irrFrequency);
+		if (m_Called)
+			result |= (now >= m_irrCalleeTimer) && ((now - m_irrCalleeTimer) >= 2 * m_irrFrequency);
+	}
+	return result;
 }
 
 #endif // RASTBL_H
