@@ -20,6 +20,7 @@
 #endif
 
 #include <ptlib.h>
+#include <ptclib/enum.h>
 #include <h323pdu.h>
 #include "gk_const.h"
 #include "h323util.h"
@@ -31,15 +32,16 @@
 #include "GkStatus.h"
 #include "sigmsg.h"
 #include "Routing.h"
+#include "pwlib_compat.h"
 
 using std::string;
 using std::vector;
+using std::list;
 using std::stable_sort;
 using std::binary_function;
 
 namespace Routing {
 
-
 const char *SectionName[] = {
 	"RoutingPolicy::OnARQ",
 	"RoutingPolicy::OnLRQ",
@@ -48,84 +50,145 @@ const char *SectionName[] = {
 	"RoutingPolicy"
 };
 
+const long DEFAULT_ROUTE_REQUEST_TIMEOUT = 10;
+const char* const CTIsection = "CTI::Agents";
+
+Route::Route() : m_proxyMode(CallRec::ProxyDetect), m_flags(0)
+{
+	Toolkit::Instance()->SetRerouteCauses(m_rerouteCauses);
+}
+
+Route::Route(
+	const endptr &destEndpoint
+	) : m_destAddr(destEndpoint->GetCallSignalAddress()), m_destEndpoint(destEndpoint),
+	m_proxyMode(CallRec::ProxyDetect), m_flags(0)
+{
+	Toolkit::Instance()->SetRerouteCauses(m_rerouteCauses);
+}
+
+Route::Route(
+	const PString &policyName,
+	const H225_TransportAddress &destAddr
+	) : m_destAddr(destAddr), m_policy(policyName), m_proxyMode(CallRec::ProxyDetect), m_flags(0)
+{
+	Toolkit::Instance()->SetRerouteCauses(m_rerouteCauses);
+}
+
+Route::Route(
+	const PString &policyName,
+	const PIPSocket::Address &destIpAddr,
+	WORD destPort
+	) : m_destAddr(SocketToH225TransportAddr(destIpAddr, destPort)),
+	m_policy(policyName), m_proxyMode(CallRec::ProxyDetect), m_flags(0)
+{
+	Toolkit::Instance()->SetRerouteCauses(m_rerouteCauses);
+}
+
+PString Route::AsString() const
+{
+	return AsDotString(m_destAddr) + " (policy: " + m_policy + ", proxy: "
+		+ PString(m_proxyMode) + ", flags: " + PString(m_flags) + ")";
+}
+
+bool Route::IsFailoverActive(
+	unsigned cause
+	) const
+{
+	cause = cause & 0x7f;
+	return m_rerouteCauses[cause >> 3] & (1UL << (cause & 7));
+}
 
 // class RoutingRequest
+RoutingRequest::RoutingRequest()
+	: m_reason(-1), m_flags(0)
+{
+}
+
 RoutingRequest::RoutingRequest(
-	endptr &called
-	) : m_destination(0), m_neighbor_used(0), m_called(called),
-	m_reason(-1), m_flags(0), m_proxyMode(CallRec::ProxyDetect)
+	const std::list<Route> &failedRoutes
+	)
+	: m_reason(-1), m_flags(0), m_failedRoutes(failedRoutes)
 {
 }
 
 RoutingRequest::~RoutingRequest()
 {
-	delete m_destination;
 }
 
-bool RoutingRequest::SetDestination(const H225_TransportAddress & dest, bool find_called)
+bool RoutingRequest::AddRoute(
+	const Route &route
+	)
 {
-	if (m_destination) {
-		PTRACE(1, "ROUTING\tError: destination overwritten!");
-		return false;
-	}
 	PIPSocket::Address addr;
 	WORD port;
-	if (!(dest.IsValid() && GetIPAndPortFromTransportAddr(dest, addr, port) 
+	if (!(route.m_destAddr.IsValid() && GetIPAndPortFromTransportAddr(route.m_destAddr, addr, port) 
 			&& addr.IsValid() && port != 0)) {
-		PTRACE(1, "ROUTING\tInvalid destination address: " << dest);
+		PTRACE(1, "ROUTING\tInvalid destination address: " << route.m_destAddr);
 		return false;
 	}
-	m_destination = new H225_TransportAddress(dest);
-	if (find_called)
-		m_called = RegistrationTable::Instance()->FindBySignalAdr(dest);
+	list<Route>::const_iterator i = m_failedRoutes.begin();
+	while (i != m_failedRoutes.end()) {
+		if (i->m_destAddr == route.m_destAddr && i->m_policy == route.m_policy
+				&& i->m_routeId == route.m_routeId) {
+			PTRACE(5, "ROUTING\tSkipping failed route " << route.AsString());
+			return true;
+		}
+		++i;
+	}
+	m_routes.push_back(route);
 	return true;
 }
 
-bool RoutingRequest::SetCalledParty(const endptr & called)
+bool RoutingRequest::GetFirstRoute(
+	Route &route
+	)
 {
-	m_called = called;
-	if (called)
-		return SetDestination(called->GetCallSignalAddress());
-	else {
-		delete m_destination;
-		m_destination = NULL;
+	if (m_routes.empty())
 		return false;
-	}
+
+	route = *m_routes.begin();
+	return true;
 }
 
+void RoutingRequest::RemoveAllRoutes()
+{
+	m_routes.clear();
+}
 
 // class AdmissionRequest
 template<> H225_ArrayOf_AliasAddress *AdmissionRequest::GetAliases()
 {
-	return (m_request.HasOptionalField(H225_AdmissionRequest::e_destinationInfo) && m_request.m_destinationInfo.GetSize() > 0) ? &m_request.m_destinationInfo : 0;
+	return (m_request.HasOptionalField(H225_AdmissionRequest::e_destinationInfo) && m_request.m_destinationInfo.GetSize() > 0)
+		? &m_request.m_destinationInfo : NULL;
 }
 
+template<> void AdmissionRequest::SetAliases(H225_ArrayOf_AliasAddress & aliases)
+{
+	m_request.IncludeOptionalField(H225_AdmissionRequest::e_destinationInfo);
+	m_request.m_destinationInfo = aliases;
+}
 
 // class LocationRequest
 template<> H225_ArrayOf_AliasAddress *LocationRequest::GetAliases()
 {
-	return (m_request.m_destinationInfo.GetSize() > 0) ? &m_request.m_destinationInfo : 0;
+	return (m_request.m_destinationInfo.GetSize() > 0)
+		? &m_request.m_destinationInfo : NULL;
 }
 
-
 // class SetupRequest
 template<> H225_ArrayOf_AliasAddress *SetupRequest::GetAliases()
 {
-	return (m_request.HasOptionalField(H225_Setup_UUIE::e_destinationAddress)
-			&& m_request.m_destinationAddress.GetSize() > 0) 
+	return (m_request.HasOptionalField(H225_Setup_UUIE::e_destinationAddress) && m_request.m_destinationAddress.GetSize() > 0)
 		? &m_request.m_destinationAddress : NULL;
 }
 
-
 // class FacilityRequest
 template<> H225_ArrayOf_AliasAddress *FacilityRequest::GetAliases()
 {
-	return (m_request.HasOptionalField(H225_Facility_UUIE::e_alternativeAliasAddress) 
-			&& m_request.m_alternativeAliasAddress.GetSize() > 0)
+	return (m_request.HasOptionalField(H225_Facility_UUIE::e_alternativeAliasAddress) && m_request.m_alternativeAliasAddress.GetSize() > 0)
 		? &m_request.m_alternativeAliasAddress : NULL;
 }
 
-
 bool Policy::Handle(SetupRequest& request)
 {
 	if( IsActive() ) {
@@ -312,19 +375,40 @@ protected:
 bool ExplicitPolicy::OnRequest(AdmissionRequest & request)
 {
 	H225_AdmissionRequest & arq = request.GetRequest();
-	return arq.HasOptionalField(H225_AdmissionRequest::e_destCallSignalAddress) ? request.SetDestination(arq.m_destCallSignalAddress, true) : false;
+	if (arq.HasOptionalField(H225_AdmissionRequest::e_destCallSignalAddress)) {
+		Route route(m_name, arq.m_destCallSignalAddress);
+		route.m_destEndpoint = RegistrationTable::Instance()->FindBySignalAdr(
+			route.m_destAddr
+			);
+		return request.AddRoute(route);
+	}
+	return false;
 }
 
-bool ExplicitPolicy::OnRequest(SetupRequest & request)
+bool ExplicitPolicy::OnRequest(SetupRequest &request)
 {
-	H225_Setup_UUIE & setup = request.GetRequest();
-	return setup.HasOptionalField(H225_Setup_UUIE::e_destCallSignalAddress) ? request.SetDestination(setup.m_destCallSignalAddress, true) : false;
+	H225_Setup_UUIE &setup = request.GetRequest();
+	if (setup.HasOptionalField(H225_Setup_UUIE::e_destCallSignalAddress)) {
+		Route route(m_name, setup.m_destCallSignalAddress);
+		route.m_destEndpoint = RegistrationTable::Instance()->FindBySignalAdr(
+			route.m_destAddr
+			);
+		return request.AddRoute(route);
+	}
+	return false;
 }
 
 bool ExplicitPolicy::OnRequest(FacilityRequest & request)
 {
 	H225_Facility_UUIE & facility = request.GetRequest();
-	return facility.HasOptionalField(H225_Facility_UUIE::e_alternativeAddress) ? request.SetDestination(facility.m_alternativeAddress, true) : false;
+	if (facility.HasOptionalField(H225_Facility_UUIE::e_alternativeAddress)) {
+		Route route(m_name, facility.m_alternativeAddress);
+		route.m_destEndpoint = RegistrationTable::Instance()->FindBySignalAdr(
+			route.m_destAddr
+			);
+		return request.AddRoute(route);
+	}
+	return false;
 }
 
 
@@ -355,11 +439,6 @@ bool InternalPolicy::OnRequest(AdmissionRequest & request)
 	H225_ArrayOf_AliasAddress *aliases = request.GetAliases();
 	if (aliases == NULL || !FindByAliases(request, *aliases))
 		return false;
-	if (request.GetCalledParty())
-		if (!request.GetCalledParty()->HasAvailableCapacity()) {
-			request.SetCalledParty(endptr(NULL));
-			request.SetRejectReason(H225_AdmissionRejectReason::e_resourceUnavailable);
-		}
 
 	return true;
 }
@@ -369,24 +448,43 @@ bool InternalPolicy::OnRequest(SetupRequest & request)
 	H225_ArrayOf_AliasAddress *aliases = request.GetAliases();
 	if (aliases == NULL || !FindByAliases(request, *aliases))
 		return false;
-	if (request.GetCalledParty())
-		if (!request.GetCalledParty()->HasAvailableCapacity()) {
-			request.SetCalledParty(endptr(NULL));
-			request.SetRejectReason(H225_ReleaseCompleteReason::e_gatewayResources);
-		}
 
 	return true;
 }
 
-bool InternalPolicy::FindByAliases(RoutingRequest& request, H225_ArrayOf_AliasAddress & aliases)
+bool InternalPolicy::FindByAliases(
+	RoutingRequest &request, 
+	H225_ArrayOf_AliasAddress &aliases
+	)
 {
-	return request.SetCalledParty(RegistrationTable::Instance()->FindEndpoint(aliases, roundRobin));
+	list<Route> routes;
+	RegistrationTable::Instance()->FindEndpoint(
+		aliases, roundRobin, true, routes
+		);
+	list<Route>::iterator i = routes.begin();
+	while (i != routes.end()) {
+		i->m_policy = m_name;
+		request.AddRoute(*i++);
+	}
+	return !routes.empty();
 }
 
-bool InternalPolicy::FindByAliases(LocationRequest& request, H225_ArrayOf_AliasAddress & aliases)
+bool InternalPolicy::FindByAliases(
+	LocationRequest& request,
+	H225_ArrayOf_AliasAddress & aliases
+	)
 {
-	// do not apply round robin selection for gateways
-	return request.SetCalledParty(RegistrationTable::Instance()->FindEndpoint(aliases, false));
+	// do not apply round robin selection for Location ReQuests
+	list<Route> routes;
+	RegistrationTable::Instance()->FindEndpoint(
+		aliases, false, true, routes
+		);
+	list<Route>::iterator i = routes.begin();
+	while (i != routes.end()) {
+		i->m_policy = m_name;
+		request.AddRoute(*i++);
+	}
+	return !routes.empty();
 }
 
 
@@ -450,16 +548,24 @@ protected:
 	virtual bool FindByAliases(LocationRequest &, H225_ArrayOf_AliasAddress &);
 };
 
-bool DNSPolicy::FindByAliases(RoutingRequest & request, H225_ArrayOf_AliasAddress & aliases)
+bool DNSPolicy::FindByAliases(
+	RoutingRequest &request,
+	H225_ArrayOf_AliasAddress &aliases
+	)
 {
 	for (PINDEX i = 0; i < aliases.GetSize(); ++i) {
 		PString alias(AsString(aliases[i], FALSE));
 		PINDEX at = alias.Find('@');
+
 		PString domain = (at != P_MAX_INDEX) ? alias.Mid(at + 1) : alias;
 		H225_TransportAddress dest;
 		if (GetTransportAddress(domain, GK_DEF_ENDPOINT_SIGNAL_PORT, dest)) {
-			if (!request.SetDestination(dest, true))
+			PIPSocket::Address addr;
+			if (!(GetIPFromTransportAddr(dest, addr) && addr.IsValid()))
 				continue;
+			Route route(m_name, dest);
+			route.m_destEndpoint = RegistrationTable::Instance()->FindBySignalAdr(dest);
+			request.AddRoute(route);
 			request.SetFlag(RoutingRequest::e_aliasesChanged);
 			// remove the domain name part
 			H323SetAliasAddress(alias.Left(at), aliases[i]);
@@ -470,14 +576,11 @@ bool DNSPolicy::FindByAliases(RoutingRequest & request, H225_ArrayOf_AliasAddres
 }
 
 bool DNSPolicy::FindByAliases(LocationRequest & request, H225_ArrayOf_AliasAddress & aliases)
-{
+{ 
 	return DNSPolicy::FindByAliases((RoutingRequest&)request, aliases);
 }
 
 
-#define DEFAULT_ROUTE_REQUEST_TIMEOUT 10
-const char* CTIsection = "CTI::Agents";
-
 VirtualQueue::VirtualQueue()
 	:
 	m_active(false),
@@ -577,20 +680,28 @@ bool VirtualQueue::SendRouteRequest(
 	const PString& vqueue,
 	/// a sequence of aliases for the calling endpoint
 	/// (in the "alias:type[=alias:type]..." format)
-	const PString& sourceInfo
+	const PString& sourceInfo,
+	/// the callID as string
+	const PString& callID
 	)
 {
 	bool result = false;
 	bool duprequest = false;
 	const PString epid(caller->GetEndpointIdentifier().GetValue());
-	if (RouteRequest *r = InsertRequest(epid, crv, destinationInfo, callSigAdr, duprequest)) {
-		const PString msg(PString::Printf, "RouteRequest|%s|%s|%u|%s|%s;", 
+	if (RouteRequest *r = InsertRequest(epid, crv, callID, destinationInfo, callSigAdr, duprequest)) {
+		PString msg = PString(PString::Printf, "RouteRequest|%s|%s|%u|%s|%s", 
 				(const char *)AsDotString(caller->GetCallSignalAddress()),
 				(const char *)epid,
 				crv,
 				(const char *)vqueue,
 				(const char *)sourceInfo
 			   );
+		if (Toolkit::AsBool(GkConfig()->GetString("Gatekeeper::Main", "SignalCallId", 0))) {
+			PString cid = callID;
+			cid.Replace(" ", "-", true);
+			msg += PString("|") + cid;
+		}
+		msg += PString(";");
 		// signal RouteRequest to the status line only once
 		if( duprequest )
 			PTRACE(4, "VQueue\tDuplicate request: "<<msg);
@@ -613,6 +724,58 @@ bool VirtualQueue::SendRouteRequest(
 	return result;
 }
 
+bool VirtualQueue::SendRouteRequest(
+	/// source IP of the request (endpoint for ARQ, gatekeeper for LRQ)
+	const PString& source,
+	/// calling endpoint
+	const PString& epid,
+	/// requestSeqNum of the request
+	unsigned seq,
+	/// destination (virtual queue) aliases as specified
+	/// by the calling endpoint (modified by this function on successful return)
+	H225_ArrayOf_AliasAddress* destinationInfo,
+	/// actual virtual queue name (should be present in destinationInfo too)
+	const PString& vqueue,
+	/// a sequence of aliases for the calling endpoint
+	/// (in the "alias:type[=alias:type]..." format)
+	const PString& sourceInfo
+	)
+{
+	bool result = false;
+	PString * callSigAdr = new PString;
+	bool duprequest = false;
+	if (RouteRequest *r = InsertRequest(epid, seq, "", destinationInfo, callSigAdr, duprequest)) {
+		PString msg = PString(PString::Printf, "RouteRequest|%s|%s|%u|%s|%s",
+				(const char *)source,
+				(const char *)epid,
+				seq,
+				(const char *)vqueue,
+				(const char *)sourceInfo
+			);
+		msg += PString(";");
+		// signal RouteRequest to the status port only once
+		if( duprequest )
+			PTRACE(4, "VQueue\tDuplicate request: " << msg);
+		else {
+			PTRACE(2, msg);
+			GkStatus::Instance()->SignalStatus(msg + "\r\n", STATUS_TRACE_LEVEL_ROUTEREQ);
+		}
+
+		// wait for an answer from the status line (routetoalias,routereject)
+		result = r->m_sync.Wait(m_requestTimeout);
+		m_listMutex.Wait();
+		m_pendingRequests.remove(r);
+		m_listMutex.Signal();
+		if( !result )
+			PTRACE(5,"VQueue\tRoute request (EPID: "<<r->m_callingEpId
+				<<", CRV="<<r->m_crv<<") timed out"
+				);
+		delete r;
+		delete callSigAdr;
+	}
+	return result;
+}
+
 bool VirtualQueue::IsDestinationVirtualQueue(
 	const PString& destinationAlias /// alias to be matched
 	) const
@@ -640,7 +803,9 @@ bool VirtualQueue::RouteToAlias(
 	/// identifier of the endpoint associated with the route request
 	const PString& callingEpId, 
 	/// CRV of the call associated with the route request
-	unsigned crv
+	unsigned crv,
+	/// callID of the call associated with the route request
+	const PString& callID
 	)
 {
 	PWaitAndSignal lock(m_listMutex);
@@ -650,7 +815,12 @@ bool VirtualQueue::RouteToAlias(
 	RouteRequests::iterator i = m_pendingRequests.begin();
 	while (i != m_pendingRequests.end()) {
 		RouteRequest *r = *i;
-		if (r->m_callingEpId == callingEpId && r->m_crv == crv) {
+		bool match = ((r->m_callingEpId == callingEpId) && (r->m_crv == crv));
+		if (!r->m_callID.IsEmpty() && !callID.IsEmpty()) {
+			// backward compatibility: only check if set
+			match = match && (r->m_callID == callID);
+		}
+		if (match) {
 			// replace virtual queue aliases info with agent aliases
 			*r->m_agent = agent;
 			if (!destinationip.IsEmpty())	// RoteToGateway
@@ -685,31 +855,37 @@ bool VirtualQueue::RouteToAlias(
 	/// identifier of the endpoint associated with the route request
 	const PString& callingEpId, 
 	/// CRV for the call associated with the route request
-	unsigned crv
+	unsigned crv,
+	/// callID of the call associated with the route request
+	const PString& callID
 	)
 {
 	H225_ArrayOf_AliasAddress alias;
 	alias.SetSize(1);
 	H323SetAliasAddress(targetAlias, alias[0]);
-	return RouteToAlias(alias, destinationIp, callingEpId, crv);
+	return RouteToAlias(alias, destinationIp, callingEpId, crv, callID);
 }
 
 bool VirtualQueue::RouteReject(
 	/// identifier of the endpoint associated with the route request
 	const PString& callingEpId, 
 	/// CRV of the call associated with the route request
-	unsigned crv
+	unsigned crv,
+	/// callID of the call associated with the route request
+	const PString& callID
 	)
 {
 	H225_ArrayOf_AliasAddress nullAgent;
-	return RouteToAlias(nullAgent, "", callingEpId, crv);
+	return RouteToAlias(nullAgent, "", callingEpId, crv, callID);
 }
 
 VirtualQueue::RouteRequest* VirtualQueue::InsertRequest(
 	/// identifier for the endpoint associated with this request
-	const PString& callingEpId, 
+	const PString& callingEpId,
 	/// CRV for the call associated with this request
-	unsigned crv, 
+	unsigned crv,
+	/// callID for the call associated with this request
+	const PString& callID,
 	/// a pointer to an array to be filled with agent aliases
 	/// when the routing decision has been made
 	H225_ArrayOf_AliasAddress* agent,
@@ -729,7 +905,7 @@ VirtualQueue::RouteRequest* VirtualQueue::InsertRequest(
 	RouteRequests::iterator i = m_pendingRequests.begin();
 	while (i != m_pendingRequests.end()) {
 		RouteRequest *r = *i;
-		if (r->m_callingEpId == callingEpId && r->m_crv == crv)
+		if (r->m_callingEpId == callingEpId && r->m_crv == crv && r->m_callID == callID)
 			duprequests++;
 		++i;
 	}
@@ -743,7 +919,7 @@ VirtualQueue::RouteRequest* VirtualQueue::InsertRequest(
 	}
 
 	// insert the new pending route request
-	RouteRequest* r = new RouteRequest(callingEpId, crv, agent, callSigAdr);
+	RouteRequest* r = new RouteRequest(callingEpId, crv, callID, agent, callSigAdr);
 	m_pendingRequests.push_back(r);
 	return r;
 }
@@ -759,8 +935,8 @@ private:
 	virtual bool IsActive();
 
 	virtual bool OnRequest(AdmissionRequest &);
+	virtual bool OnRequest(LocationRequest &);
 	// TODO
-	//virtual bool OnRequest(LocationRequest &);
 	//virtual bool OnRequest(SetupRequest &);
 	//virtual bool OnRequest(FacilityRequest &);
 
@@ -790,7 +966,7 @@ bool VirtualQueuePolicy::OnRequest(AdmissionRequest & request)
 				);
 			endptr ep = RegistrationTable::Instance()->FindByEndpointId(arq.m_endpointIdentifier); // should not be null
 			PString * callSigAdr = new PString();
-			if (ep && m_vqueue->SendRouteRequest(ep, unsigned(arq.m_callReferenceValue), aliases, callSigAdr, agent, AsString(arq.m_srcInfo)))
+			if (ep && m_vqueue->SendRouteRequest(ep, unsigned(arq.m_callReferenceValue), aliases, callSigAdr, agent, AsString(arq.m_srcInfo), AsString(arq.m_callIdentifier.m_guid)))
 				request.SetFlag(RoutingRequest::e_aliasesChanged);
 			if (!callSigAdr->IsEmpty()) {
 				if (!arq.HasOptionalField(H225_AdmissionRequest::e_destCallSignalAddress)) {
@@ -816,6 +992,43 @@ bool VirtualQueuePolicy::OnRequest(AdmissionRequest & request)
 	return false;
 }
 
+bool VirtualQueuePolicy::OnRequest(LocationRequest & request)
+{
+	if (H225_ArrayOf_AliasAddress *aliases = request.GetAliases()) {
+		const PString agent(AsString((*aliases)[0], FALSE));
+		if (m_vqueue->IsDestinationVirtualQueue(agent)) {
+			H225_LocationRequest & lrq = request.GetRequest();
+			PTRACE(5,"Routing\tPolicy " << m_name << " destination matched "
+				"a virtual queue " << agent << " (LRQ "
+				<< lrq.m_requestSeqNum.GetValue() << ')'
+				);
+
+			// only use vqueue if sender is able to handle changed destination
+			if (lrq.HasOptionalField(H225_LocationRequest::e_canMapAlias)) {
+				if (!lrq.m_canMapAlias) {
+					PTRACE(5, "Sender can't map destination alias, skipping virtual queue");
+					return false;
+				}
+			}
+
+			PString source = AsDotString(lrq.m_replyAddress);
+			PString epid = lrq.m_endpointIdentifier.GetValue();
+			if (epid.IsEmpty())
+				epid = lrq.m_gatekeeperIdentifier.GetValue() + "_" + AsString(lrq.m_sourceInfo, false);
+			if (m_vqueue->SendRouteRequest(source, epid, unsigned(lrq.m_requestSeqNum), aliases, agent, AsString(lrq.m_sourceInfo)))
+				request.SetFlag(RoutingRequest::e_aliasesChanged);
+			// the trick: if empty, the request is rejected
+			// so we return true to terminate the routing
+			// decision process, otherwise the aliases is
+			// rewritten, we return false to let subsequent
+			// policies determine the request
+			if (m_next == NULL || aliases->GetSize() == 0)
+				return true;
+		}
+	}
+	return false;
+}
+
 class NumberAnalysisPolicy : public Policy {
 public:
 	struct PrefixEntry {
@@ -828,9 +1041,9 @@ public:
 
 protected:
 	virtual bool OnRequest(AdmissionRequest &);
+	virtual bool OnRequest(SetupRequest &);
 	// TODO
 	//virtual bool OnRequest(LocationRequest &);
-	//virtual bool OnRequest(SetupRequest &);
 	//virtual bool OnRequest(FacilityRequest &);
 
 private:
@@ -865,7 +1078,7 @@ NumberAnalysisPolicy::NumberAnalysisPolicy()
 	for (PINDEX i = 0; i < kv.GetSize(); i++) {
 		const PString &val = kv.GetDataAt(i);
 
-		m_prefixes[i].m_prefix = kv.GetKeyAt(i);
+		m_prefixes[i].m_prefix = string((const char*)(kv.GetKeyAt(i)));
 
 		const PINDEX sepIndex = val.Find(':');
 		if (sepIndex == P_MAX_INDEX) {
@@ -910,12 +1123,12 @@ bool NumberAnalysisPolicy::OnRequest(AdmissionRequest & request)
 			for (unsigned i = 0; i < m_prefixes.size(); i++)
 				if (MatchPrefix(s, m_prefixes[i].m_prefix.c_str()) != 0) {
 					if (s.GetLength() < m_prefixes[i].m_minLength) {
-						request.SetCalledParty(endptr(NULL));
+						request.RemoveAllRoutes();
 						request.SetRejectReason(H225_AdmissionRejectReason::e_incompleteAddress);
 						return true;
 					} else if (m_prefixes[i].m_maxLength >= 0
 							&& s.GetLength() > m_prefixes[i].m_maxLength) {
-						request.SetCalledParty(endptr(NULL));
+						request.RemoveAllRoutes();
 						request.SetRejectReason(H225_AdmissionRejectReason::e_undefinedReason);
 						return true;
 					}
@@ -926,6 +1139,84 @@ bool NumberAnalysisPolicy::OnRequest(AdmissionRequest & request)
 	return false;
 }
 
+bool NumberAnalysisPolicy::OnRequest(SetupRequest & request)
+{
+	H225_ArrayOf_AliasAddress *aliases = request.GetAliases();
+	if (aliases == NULL)
+		return false;
+
+	for (PINDEX i = 0; i < aliases->GetSize(); ++i) {
+		const H225_AliasAddress &alias = (*aliases)[i];
+		if (alias.GetTag() == H225_AliasAddress::e_dialedDigits
+				|| alias.GetTag() == H225_AliasAddress::e_partyNumber) {
+			const PString s = AsString(alias, FALSE);
+			for (unsigned i = 0; i < m_prefixes.size(); ++i)
+				if (MatchPrefix(s, m_prefixes[i].m_prefix.c_str()) != 0) {
+					if (s.GetLength() < m_prefixes[i].m_minLength) {
+						request.RemoveAllRoutes();
+						request.SetRejectReason(H225_ReleaseCompleteReason::e_badFormatAddress);
+						return true;
+					} else if (m_prefixes[i].m_maxLength >= 0
+							&& s.GetLength() > m_prefixes[i].m_maxLength) {
+						request.RemoveAllRoutes();
+						request.SetRejectReason(H225_ReleaseCompleteReason::e_badFormatAddress);
+						return true;
+					}
+					return false;
+				}
+		}
+	}
+	return false;
+}
+
+// a policy to look up the destination from ENUM Name Server
+class ENUMPolicy : public AliasesPolicy {
+public:
+	ENUMPolicy() { m_name = "ENUM"; }
+protected:
+	virtual bool FindByAliases(RoutingRequest &, H225_ArrayOf_AliasAddress &);
+	virtual bool FindByAliases(LocationRequest &, H225_ArrayOf_AliasAddress &);
+};
+
+bool ENUMPolicy::FindByAliases(RoutingRequest & request, H225_ArrayOf_AliasAddress & aliases)
+{
+#if P_DNS
+	for (PINDEX i = 0; i < aliases.GetSize(); ++i) {
+		PString alias(AsString(aliases[i], FALSE));
+
+		// make sure the number has only digits
+		alias.Replace("+","", true);
+		alias.Replace("*","", true);
+		alias.Replace("#","", true);
+		alias.Replace(",","", true);
+		while (alias.Left(1) *= "0")	// ENUM registries expect aliases without leading zeros
+			alias = alias.Mid(1);
+		PINDEX j;
+		for (j = 0; j < alias.GetLength(); ++j)
+			if (!isdigit(alias[j]))
+				break;
+
+		if (j >= alias.GetLength()) {
+			PString str;
+			if (PDNS::ENUMLookup(alias, "E2U+h323", str)) {
+				if (str.Left(5) *= "h323:")
+					str = str.Mid(5);
+				PTRACE(4, "\tENUM converted remote party " << alias << " to " << str);
+				request.SetFlag(RoutingRequest::e_aliasesChanged);
+				H323SetAliasAddress(str, aliases[i]);
+		  	}
+		}
+	}
+#endif
+
+	return false;
+}
+
+bool ENUMPolicy::FindByAliases(LocationRequest & request, H225_ArrayOf_AliasAddress & aliases)
+{
+	return ENUMPolicy::FindByAliases((RoutingRequest&)request, aliases);
+}
+
 
 namespace { // anonymous namespace
 	SimpleCreator<ExplicitPolicy> ExplicitPolicyCreator("explicit");
@@ -934,6 +1225,7 @@ namespace { // anonymous namespace
 	SimpleCreator<DNSPolicy> DNSPolicyCreator("dns");
 	SimpleCreator<VirtualQueuePolicy> VirtualQueuePolicyCreator("vqueue");
 	SimpleCreator<NumberAnalysisPolicy> NumberAnalysisPolicyCreator("numberanalysis");
+	SimpleCreator<ENUMPolicy> ENUMPolicyCreator("enum");
 }
 
 
