@@ -17,13 +17,13 @@
 //////////////////////////////////////////////////////////////////
 
 #ifndef ROUTING_H
-#define ROUTING_H "@(#) $Id: Routing.h,v 1.18 2005/04/24 16:39:43 zvision Exp $"
+#define ROUTING_H "@(#) $Id: Routing.h,v 1.23 2006/12/06 17:02:59 willamowius Exp $"
 
 #include <map>
 #include <list>
 #include "slist.h"
 #include "singleton.h"
-
+#include "RasTbl.h"
 
 // forward references to avoid includes
 class H225_AdmissionRequest;
@@ -39,58 +39,86 @@ typedef H225SignalingMsg<H225_Setup_UUIE> SetupMsg;
 typedef H225SignalingMsg<H225_Facility_UUIE> FacilityMsg;
 
 class RasMsg;
-class EndpointRec;
-template<class> class SmartPtr;
-typedef SmartPtr<EndpointRec> endptr;
 
 namespace Routing {
 
+/// An entry for a single call destination route
+struct Route {
+	// a policy can set flags to indicate extra status of a processed request
+	enum Flags {
+		e_toParent = 1,
+		e_toNeighbor = 2
+	};
+	
+	Route();
+	Route(
+		const endptr &destEndpoint
+		);
+	Route(
+		const PString &policyName,
+		const H225_TransportAddress &destAddr
+		);
+	Route(
+		const PString &policyName,
+		const PIPSocket::Address &destIpAddr,
+		WORD destPort
+		);
+		
+	PString AsString() const;
+
+	bool IsFailoverActive(
+		unsigned cause
+		) const;
+
+	H225_TransportAddress m_destAddr; /// destination address for signaling
+	endptr m_destEndpoint; /// destination endpoint record (if available)
+	H225_AliasAddress m_srcAddr; /// source alias that matched the route
+	PString m_policy; /// name of the policy that found the route
+	PString m_routeId; /// optional policy-specific route identifier
+	int m_proxyMode; /// per-route proxy mode flag
+	unsigned m_flags; /// additional route specific flags
+	unsigned char m_rerouteCauses[16]; /// bit flags to trigger rerouting on particular Q931 causes
+};
 
 class RoutingRequest {
 public:
-	// a policy can set flags to indicate extra status of a processed request
-	enum ExtraFlags {
+	enum Flags {
 		e_aliasesChanged = 1,
-		e_fromInternal	 = 2,
-		e_fromParent	 = 4,
-		e_fromNeighbor	 = 8,
-		e_toInternal	 = 16,
-		e_toParent	 = 32,
-		e_toNeighbor	 = 64
+		e_fromInternal = 2,
+		e_fromParent = 4,
+		e_fromNeighbor = 8
 	};
-
+	
 	// note this is not a polymorphic class
-	RoutingRequest(endptr &called);
+	RoutingRequest();
+	RoutingRequest(
+		const std::list<Route> &failedRoutes
+		);
 	~RoutingRequest();
 
-	H225_TransportAddress* GetDestination() const { return m_destination; }
-	bool SetDestination(const H225_TransportAddress &, bool = false);
-	bool SetCalledParty(const endptr &);
+	bool AddRoute(
+		const Route &route
+		);
+	bool GetFirstRoute(
+		Route &route
+		);
+	void RemoveAllRoutes();
+	std::list<Route> &GetRoutes() { return m_routes; }
+	
 	void SetRejectReason(unsigned reason) { m_reason = reason; }
 	void SetFlag(unsigned f) { m_flags |= f; }
 	unsigned GetRejectReason() const { return m_reason; }
 	unsigned GetFlags() const { return m_flags; }
-	endptr& GetCalledParty() { return m_called; }
-	void SetNeighborUsed(PIPSocket::Address neighbor) { m_neighbor_used = neighbor; }
-	PIPSocket::Address GetNeighborUsed() { return m_neighbor_used; }
-	int GetProxyMode() const { return m_proxyMode; }
-	void SetProxyMode(int mode) { m_proxyMode = mode; }
 
 private:
-	RoutingRequest();
 	RoutingRequest(const RoutingRequest&);
 	RoutingRequest& operator=(const RoutingRequest&);
 
-protected:
-	H225_TransportAddress *m_destination;
-	PIPSocket::Address m_neighbor_used;
-
 private:
-	endptr & m_called;
-	int m_reason;
-	unsigned m_flags;
-	/// override global proxy setting from the config (see #CallRec::ProxyMode enum#)
-	int m_proxyMode;
+	int m_reason; /// reject reason, if no routes are found
+	unsigned m_flags; /// request specific flags
+	std::list<Route> m_routes;
+	std::list<Route> m_failedRoutes;
 };
 
 template<class R, class W>
@@ -99,13 +127,16 @@ public:
 	typedef R ReqObj;
 	typedef W Wrapper;
 
-	Request(ReqObj & r, Wrapper *w, endptr & c) : RoutingRequest(c), m_request(r), m_wrapper(w) {}
+	Request(ReqObj & r, Wrapper *w) : m_request(r), m_wrapper(w) {}
+	Request(ReqObj & r, Wrapper *w, const std::list<Route> &failedRoutes)
+		: RoutingRequest(failedRoutes), m_request(r), m_wrapper(w) {}
 
-	H225_TransportAddress *Process();
+	bool Process();
 
 	ReqObj & GetRequest() { return m_request; }
 	Wrapper *GetWrapper() { return m_wrapper; }
 	H225_ArrayOf_AliasAddress *GetAliases();
+	void SetAliases(H225_ArrayOf_AliasAddress & aliases);
 	const ReqObj & GetRequest() const { return m_request; }
 	const Wrapper *GetWrapper() const { return m_wrapper; }
 	const H225_ArrayOf_AliasAddress *GetAliases() const
@@ -211,9 +242,9 @@ private:
 
 
 template<class R, class W>
-inline H225_TransportAddress *Request<R, W>::Process()
+inline bool Request<R, W>::Process()
 {
-	return Analyzer::Instance()->Parse(*this) ? m_destination : 0;
+	return Analyzer::Instance()->Parse(*this);
 }
 
 /** A class that supports ACD (Automatic Call Distribution). A call
@@ -265,6 +296,36 @@ public:
 		const PString& vqueue,
 		/// a sequence of aliases for the calling endpoint
 		/// (in the "alias:type[=alias:type]..." format)
+		const PString& sourceInfo,
+		/// the callID as string
+		const PString& callID
+		);
+
+	/** Send RouteRequest to the GK status line	and wait (LRQ version)
+		for a routing decision to be made by some external application
+		(ACD application).
+
+		@return
+		True if the external application routed the call (either by specifying
+		an alias or by rejecting the call), false if timed out waiting
+		for the routing decision.
+		If the request was rejected, destinationInfo is set to an epmty array
+		(0 elements).
+	*/
+	bool SendRouteRequest(
+		/// source IP of the request (endpoint for ARQ, gatekeeper for LRQ)
+		const PString& source,
+		/// calling endpoint
+		const PString& epid,
+		/// irequestSeqNum of the request
+		unsigned seq,
+		/// destination (virtual queue) aliases as specified
+		/// by the calling endpoint (modified by this function on successful return)
+		H225_ArrayOf_AliasAddress* destinationInfo,
+		/// an actual virtual queue name (should be present in destinationInfo too)
+		const PString& vqueue,
+		/// a sequence of aliases for the calling endpoint
+		/// (in the "alias:type[=alias:type]..." format)
 		const PString& sourceInfo
 		);
 
@@ -284,7 +345,9 @@ public:
 		/// identifier of the endpoint associated with the route request
 		const PString& callingEpId,
 		/// CRV of the call associated with the route request
-		unsigned crv
+		unsigned crv,
+		/// callID of the call associated with the route request
+		const PString& callID
 		);
 
 	/** Make a routing decision for a pending route request (inserted
@@ -302,7 +365,9 @@ public:
 		/// identifier of the endpoint associated with the route request
 		const PString& callingEpId,
 		/// CRV of the call associated with the route request
-		unsigned crv
+		unsigned crv,
+		/// callID of the call associated with the route request
+		const PString& callID
 		);
 
 	/** Reject a pending route request (inserted by SendRequest).
@@ -314,7 +379,9 @@ public:
 		/// identifier of the endpoint associated with the route request
 		const PString& callingEpId,
 		/// CRV of the call associated with the route request
-		unsigned crv
+		unsigned crv,
+		/// callID of the call associated with the route request
+		const PString& callID
 		);
 
 	/** @return
@@ -331,17 +398,20 @@ private:
 		RouteRequest(
 			const PString& callingEpId,
 			unsigned crv,
+			const PString& callID,
 			H225_ArrayOf_AliasAddress* agent,
 			PString* callsignaladdr
 			)
 			:
-			m_callingEpId((const char*)callingEpId), m_crv(crv),
+			m_callingEpId((const char*)callingEpId), m_crv(crv), m_callID(callID),
 			m_agent(agent), m_callsignaladdr(callsignaladdr) {}
 
 		/// identifier for the endpoint associated with this request
 		PString m_callingEpId;
 		/// CRV for the call associated with this request
 		unsigned m_crv;
+		/// callID for the call associated with this request
+		PString m_callID;
 		/// aliases for the virtual queue matched (on input)
 		/// aliases for the target agent - target route (on output)
 		H225_ArrayOf_AliasAddress* m_agent;
@@ -359,6 +429,8 @@ private:
 		const PString& callingEpId,
 		/// CRV for the call associated with this request
 		unsigned crv,
+		/// callID for the call associated with this request
+		const PString& callID,
 		/// a pointer to an array to be filled with agent aliases
 		/// when the routing decision has been made
 		H225_ArrayOf_AliasAddress* agent,
