@@ -29,11 +29,13 @@
 #include "gksql.h"
 #endif
 #include "clirw.h"
+#include "capctrl.h"
 #include "Toolkit.h"
 
 using namespace std;
 
 extern const char *ProxySection;
+extern const char *RoutedSec;
 
 namespace {
 
@@ -277,18 +279,27 @@ void Toolkit::RouteTable::ClearTable()
 // can't pass a reference of Address, or STL complains...
 PIPSocket::Address Toolkit::RouteTable::GetLocalAddress(const Address & addr) const
 {
+  // If a dynamic external IP retrieve external IP from DNS entries
+	if (DynExtIP && !addr.IsRFC1918()) {
+	  PIPSocket::Address extip;
+	  H323TransportAddress ex = H323TransportAddress(ExtIP);
+	  ex.GetIpAddress(extip);
+	  return extip;
+	}
+
 	RouteEntry *entry = find_if(rtable_begin, rtable_end,
 			bind2nd(mem_fun_ref(&RouteEntry::Compare), &addr));
 	return (entry != rtable_end) ? entry->GetDestination() : defAddr;
 }
 
-bool Toolkit::RouteTable::CreateTable()
+bool Toolkit::RouteTable::CreateRouteTable(const PString & extroute)
 {
 	InterfaceTable if_table;
 	if (!PIPSocket::GetInterfaceTable(if_table)) {
 		PTRACE(1, "Error: Can't get interface table");
 		return false;
 	}
+
 	PTRACE(4, "InterfaceTable:\n" << setfill('\n') << if_table << setfill(' '));
 	PIPSocket::RouteTable r_table;
 	if (!PIPSocket::GetRouteTable(r_table)) {
@@ -296,30 +307,79 @@ bool Toolkit::RouteTable::CreateTable()
 		return false;
 	}
 
-	int i = r_table.GetSize();
+	if (!extroute && AsBool(GkConfig()->GetString(ProxySection, "Enable", "0"))) {
+	  for (PINDEX i = 0; i < r_table.GetSize(); ++i) {
+		if (r_table[i].GetNetwork().IsRFC1918() && r_table[i].GetNetMask().AsString() != "255.255.255.255") {
+		  PString intAddr = r_table[i].GetNetwork().AsString() + "/" + r_table[i].GetNetMask().AsString();
+		  m_internalnetworks.resize( m_internalnetworks.size() + 1);
+		  m_internalnetworks[m_internalnetworks.size() - 1] = NetworkAddress(intAddr);
+		  PTRACE(2, "Internal Network Detected " << m_internalnetworks.back().AsString()); 
+		} 
+	  }
+	}
+
+	int i = (!extroute) ? r_table.GetSize()+1 : r_table.GetSize();
+
 	rtable_end = rtable_begin = static_cast<RouteEntry *>(::malloc(i * sizeof(RouteEntry)));
 	for (PINDEX r = 0; r < i ; ++r) {
-		PIPSocket::RouteEntry & r_entry = r_table[r];
-		if (r_entry.GetNetMask() != INADDR_ANY)
-			// placement operator
+		if (!extroute && (r==r_table.GetSize()))
+			::new (rtable_end++) RouteEntry(extroute);
+		else {
+		  PIPSocket::RouteEntry & r_entry = r_table[r];
+		  if (r_entry.GetNetMask() != INADDR_ANY) 
 			::new (rtable_end++) RouteEntry(r_entry, if_table);
+		}
 	}
+
 	return true;
 }
 
 bool Toolkit::VirtualRouteTable::CreateTable()
 {
 	PString nets = GkConfig()->GetString("NetworkInterfaces", "");
-	if (nets.IsEmpty())
-		return false;
-	PStringArray networks(nets.Tokenise(" ,;\t", FALSE));
-	int i = networks.GetSize();
-	if (i > 0) {
-		rtable_end = rtable_begin = static_cast<RouteEntry *>(::malloc(i * sizeof(RouteEntry)));
-		for (PINDEX r = 0; r < i ; ++r)
-			::new (rtable_end++) RouteEntry(networks[r]);
+	if (!nets) {		
+	   PStringArray networks(nets.Tokenise(" ,;\t", FALSE));
+	   int i = networks.GetSize();
+	   if (i > 0) {
+          rtable_end = rtable_begin = static_cast<RouteEntry *>(::malloc(i * sizeof(RouteEntry)));
+		  for (PINDEX r = 0; r < i ; ++r) 
+             ::new (rtable_end++) RouteEntry(networks[r]);
+       }
+	   return true;
 	}
-	return true;
+
+	// If we have an external IP setting then load the detected Route Table and add a route for the external IP
+	// If dynamic IP then only store the PString value and resolve the DNS when required.
+	PString extip = GkConfig()->GetString("ExternalIP", "");
+	DynExtIP = AsBool(GkConfig()->GetString("ExternalIsDynamic", "0"));
+
+	PIPSocket::Address ext((DWORD)0);
+	H323TransportAddress ex = H323TransportAddress(extip);
+	ex.GetIpAddress(ext);
+	if (ext.IsValid() && !ext.IsRFC1918()) {
+	   ExtIP = extip;
+	   PString extroute = PString();
+   	   if (!DynExtIP) 
+		  extroute = ext.AsString() + "/0";
+
+	   CreateRouteTable(extroute);
+	   PTRACE(1,"External IP = " << ExtIP << " dynamic " << DynExtIP);
+	   return true;
+	} else
+        DynExtIP = false;
+
+	return false;
+}
+
+bool Toolkit::VirtualRouteTable::IsMasquerade(PIPSocket::Address & addr)
+{
+	if (!ExtIP) {
+	  H323TransportAddress ex = H323TransportAddress(ExtIP);
+	  ex.GetIpAddress(addr);
+	  return true;
+	}
+
+	return false; 
 }
 
 // class Toolkit::ProxyCriterion
@@ -333,8 +393,6 @@ Toolkit::ProxyCriterion::~ProxyCriterion()
 
 void Toolkit::ProxyCriterion::LoadConfig(PConfig *config)
 {
-	m_networks.clear();
-
 	m_enable = AsBool(config->GetString(ProxySection, "Enable", "0"));
 	if (!m_enable) {
 		PTRACE(2, "GK\tH.323 Proxy disabled");
@@ -343,12 +401,23 @@ void Toolkit::ProxyCriterion::LoadConfig(PConfig *config)
 
 	PTRACE(2, "GK\tH.323 Proxy enabled");
 
+	m_networks.clear();
+
 	PStringArray networks(config->GetString(ProxySection, "InternalNetwork", "").Tokenise(" ,;\t", FALSE));
-	for (PINDEX i = 0; i < networks.GetSize(); ++i) {
+
+	// if no networks specified then use the detected values
+	if (networks.GetSize() == 0) {
+	  m_networks = Toolkit::Instance()->GetInternalNetworks();
+	  for (unsigned j = 0; j < m_networks.size(); ++j) 
+		PTRACE(2, "GK\tInternal Network " << j << " = " << m_networks[j].AsString());
+	  return;
+	} 
+
+	  for (PINDEX i = 0; i < networks.GetSize(); ++i) {
 		m_networks.resize(m_networks.size() + 1);
 		m_networks[m_networks.size() - 1] = NetworkAddress(networks[i]);
-		PTRACE(2, "GK\tInternal Network " << i << " = " << m_networks.back().AsString());
-	}
+		PTRACE(2, "GK\tINI Internal Network " << i << " = " << m_networks.back().AsString());
+	  }
 }
 
 bool Toolkit::ProxyCriterion::Required(const Address & ip1, const Address & ip2) const
@@ -368,21 +437,29 @@ bool Toolkit::ProxyCriterion::IsInternal(const Address & ip) const
 // class Toolkit::RewriteTool
 
 static const char *RewriteSection = "RasSrv::RewriteE164";
+static const char *AliasRewriteSection = "RasSrv::RewriteAlias";
 
-Toolkit::RewriteData::RewriteData(PConfig *config, const PString & section)
+void Toolkit::RewriteData::AddSection(PConfig *config, const PString & section)
 {
-	m_RewriteKey = NULL;
 	PStringToString cfgs(config->GetAllKeyValues(section));
-	m_size = cfgs.GetSize();
-	if (m_size > 0) {
-		map<PString, PString, pstr_prefix_lesser> rules;
-		for (PINDEX i = 0; i < m_size; ++i) {
+	PINDEX n_size = cfgs.GetSize();
+	if (n_size > 0) {
+		std::map<PString, PString, pstr_prefix_lesser> rules;
+		for (PINDEX i = 0; i < n_size; ++i) {
 			PString key = cfgs.GetKeyAt(i);
-			if (!key && (isdigit(key[0]) || key[0]=='!' || key[0]=='.' || key[0]=='%' || key[0]=='*' || key[0]=='#'))
+			PCaselessString first = PCaselessString(key[0]);				
+			if (!key && (isdigit(key[0]) || (first.FindOneOf("!.%*#ABCDEFGHIGKLMNOPQRSTUVWXYZ") != P_MAX_INDEX)))			
 				rules[key] = cfgs.GetDataAt(i);
 		}
 		// now the rules are ascendantly sorted by the keys
-		if ((m_size = rules.size()) > 0) {
+		if ((n_size = rules.size()) > 0) {
+			// Add any existing rules to be resorted
+			if (m_size > 0) {
+             for (PINDEX j = 0; j < m_size; ++j) {
+				 rules[Key(j)] = Value(j);
+			 }
+			}
+			m_size = m_size + n_size;
 			// replace array constructor with explicit memory allocation
 			// and in-place new operators - workaround for VC compiler
 //			m_RewriteKey = new PString[m_size * 2];
@@ -400,6 +477,12 @@ Toolkit::RewriteData::RewriteData(PConfig *config, const PString & section)
 		}
 	}
 }
+Toolkit::RewriteData::RewriteData(PConfig *config, const PString & section)
+{
+	m_RewriteKey = NULL;
+	m_size= 0;
+    AddSection(config, section);
+}
 
 Toolkit::RewriteData::~RewriteData()
 {
@@ -416,14 +499,23 @@ void Toolkit::RewriteTool::LoadConfig(
 {
 	m_RewriteFastmatch = config->GetString(RewriteSection, "Fastmatch", "");
 	m_TrailingChar = config->GetString("RasSrv::ARQFeatures", "RemoveTrailingChar", " ")[0];
+	m_defaultDomain = config->GetString("Gatekeeper::Main", "DefaultDomain", "");
 	delete m_Rewrite;
 	m_Rewrite = new RewriteData(config, RewriteSection);
+	m_Rewrite->AddSection(config,AliasRewriteSection);
 }
 
 bool Toolkit::RewriteTool::RewritePString(PString & s) const
 {
 	bool changed = false;
 
+	// If URL remove the domain if default domain
+	 PINDEX at = s.Find('@');
+	 if (at != P_MAX_INDEX) {
+	   if (s.Mid(at+1) == m_defaultDomain)
+		   s = s.Left(at);
+	 }
+
 	// remove trailing character
 	if (s.GetLength() > 1 && s[s.GetLength() - 1] == m_TrailingChar) {
 		s = s.Left(s.GetLength() - 1);
@@ -436,6 +528,10 @@ bool Toolkit::RewriteTool::RewritePString(PString & s) const
 	PString t;
 	for (PINDEX i = 0; i < m_Rewrite->Size(); ++i) {
 		const char *prefix = m_Rewrite->Key(i);
+		if (prefix == s){
+			s = m_Rewrite->Value(i);
+            return true;
+		}
 		const int len = MatchPrefix(s, prefix);
 		// try a prefix match through all keys
 		if (len > 0 || (len == 0 && prefix[0] == '!')) {
@@ -554,7 +650,7 @@ void Toolkit::GWRewriteTool::LoadConfig(PConfig *config) {
 	PString key, cfg_value;
 	PStringArray lines, tokenised_line;
 	GWRewriteEntry *gw_entry;
-	map<PString,PString> in_strings, out_strings;
+	std::map<PString,PString> in_strings, out_strings;
 	vector<std::pair<PString,PString> > sorted_in_strings, sorted_out_strings;
 	std::map<PString,PString>::reverse_iterator strings_iterator;
 	pair<PString,PString> rule;
@@ -1084,10 +1180,40 @@ PConfig* Toolkit::ReloadConfig()
 
 	delete m_cliRewrite;
 	m_cliRewrite = new CLIRewrite;
+
+	CapacityControl::Instance()->LoadConfig();
+
+	LoadCauseMap(m_Config);
 	
 	return m_Config;
 }
 
+void Toolkit::LoadCauseMap(
+	PConfig *cfg
+	)
+{
+	memset(m_causeMap, 0, 16);
+	
+	if (! Toolkit::AsBool(cfg->GetString(RoutedSec, "ActivateFailover", "0")))
+		return;
+
+	PStringArray causes(cfg->GetString(RoutedSec, "FailoverCauses", "1-15,21-127").Tokenise(", \t", FALSE));
+
+	for (PINDEX i = 0; i < causes.GetSize(); ++i)
+		if (causes[i].Find('-') == P_MAX_INDEX) {
+			unsigned c = causes[i].AsUnsigned() & 0x7f;
+			m_causeMap[c >> 3] |= (1UL << (c & 7));
+		} else {
+			PStringArray causeRange(causes[i].Tokenise("- ", FALSE));
+			if (causeRange.GetSize() == 2) {
+				unsigned cmin = causeRange[0].AsUnsigned() & 0x7f;
+				unsigned cmax = causeRange[1].AsUnsigned() & 0x7f;
+				for (; cmin <= cmax; ++cmin)
+					m_causeMap[cmin >> 3] |= (1UL << (cmin & 7));
+			}
+		}
+}
+
 BOOL Toolkit::MatchRegex(const PString &str, const PString &regexStr)
 {
 	PINDEX pos=0;
@@ -1107,7 +1233,8 @@ BOOL Toolkit::MatchRegex(const PString &str, const PString &regexStr)
 
 bool Toolkit::RewriteE164(H225_AliasAddress &alias)
 {
-	if (alias.GetTag() != H225_AliasAddress::e_dialedDigits)
+	if ((alias.GetTag() != H225_AliasAddress::e_dialedDigits) &&
+         (alias.GetTag() != H225_AliasAddress::e_h323_ID))
 		return FALSE;
 
 	PString E164 = ::AsString(alias, FALSE);
@@ -1159,6 +1286,15 @@ bool Toolkit::GWRewriteE164(PString gw, bool direction, H225_ArrayOf_AliasAddres
 	return changed;
 }
 
+bool Toolkit::isBehindNAT(PIPSocket::Address & externalIP) {
+
+   return (m_VirtualRouteTable.IsMasquerade(externalIP));
+}
+
+std::vector<NetworkAddress> Toolkit::GetInternalNetworks() {
+
+    return GkConfig()->HasKey("ExternalIP") ? m_VirtualRouteTable.GetInternalNetworks() : m_RouteTable.GetInternalNetworks(); 
+}
 
 PString Toolkit::GetGKHome(vector<PIPSocket::Address> & GKHome) const
 {
@@ -1192,9 +1328,12 @@ void Toolkit::SetGKHome(const PStringArray & home)
 					if (m_GKHome[n] == it[i].GetAddress())
 						break;
 				if (i == is) {
-					begin = m_GKHome.begin();
-					copy(begin + n + 1, begin + size, begin + n);
-					--size, --n;
+					PTRACE(1, "GK\tAddress " << m_GKHome[n] << " not found"
+						" in the PWLib interface table"
+						);
+					//begin = m_GKHome.begin();
+					//copy(begin + n + 1, begin + size, begin + n);
+					//--size, --n;
 				}
 			}
 		}
@@ -1328,7 +1467,7 @@ PString Toolkit::CypherDecode(const PString & key, const PString & crypto, int s
 PString Toolkit::GenerateAcctSessionId()
 {
 	PWaitAndSignal lock( m_acctSessionMutex );
-	return psprintf("%08x%08x",m_acctSessionBase,++m_acctSessionCounter);
+	return psprintf(PString("%08x%08x"),m_acctSessionBase,++m_acctSessionCounter);
 }
 
 bool Toolkit::AsTimeInterval(
@@ -1522,3 +1661,10 @@ void Toolkit::RewriteCLI(
 {
 	m_cliRewrite->OutRewrite(msg, authData, addr);
 }
+
+void Toolkit::SetRerouteCauses(
+	unsigned char *causeMap
+	)
+{
+	memcpy(causeMap, m_causeMap, 128/8);
+}
