@@ -40,9 +40,10 @@
 #include "SoftPBX.h"
 #include "gktimer.h"
 #include "gk.h"
+#include "pwlib_compat.h"
+#include "capctrl.h"
 
 using std::vector;
-using std::map;
 
 /*
  * many things here should be members of Gatkeeper. 
@@ -65,7 +66,7 @@ namespace { // keep the global objects private
 
 
 PMutex ShutdownMutex;
-PMutex ReloadMutex;
+PTimedMutex ReloadMutex;
 
 #ifndef _WIN32
 PString pidfile("/var/run/gnugk.pid");
@@ -80,6 +81,7 @@ void ShutdownHandler()
 	PTRACE(3, "GK\tDeleting global reference tables");
 
 	Job::StopAll();
+	delete CapacityControl::Instance();
 	delete CallTable::Instance();
 	delete RegistrationTable::Instance();
 	delete RasServer::Instance();
@@ -408,11 +410,12 @@ BOOL Gatekeeper::InitConfig(const PArgList &args)
 
 	InstanceOf<Toolkit>()->SetConfig(fp, section);
 
-	if( (GkConfig()->GetInteger("Fourtytwo") ) != 42) { 
+	if( (GkConfig()->GetInteger("Fortytwo")  != 42) &&
+		(GkConfig()->GetInteger("Fourtytwo") != 42)) { 
 		cerr << "WARNING: No config file found!\n"
 			 << "- Does the config file exist? The default (~/.pwlib_config/Gatekeeper.ini or gatekeeper.ini in current directory) or the one given with -c?\n"
 			 << "- Did you specify they the right 'Main' section with -s?\n" 
-			 << "- Is the line 'Fourtytwo=42' present in this 'Main' section?"<<endl;
+			 << "- Is the line 'Fortytwo=42' present in this 'Main' section?"<<endl;
 	}
 	
 	return CheckSectionName(GkConfig());
@@ -515,7 +518,13 @@ void Gatekeeper::Main()
 		cerr << "Fatal: Cannot find any interface to run GnuGK!\n";
 		ExitGK();
 	}
-	cout << "Listen on " << home << "\n\n";
+	cout << "Listen on " << home << "\n";
+
+	PIPSocket::Address addr;
+	if (Toolkit::Instance()->isBehindNAT(addr))
+		cout << "Public IP: " << addr.AsString() << "\n\n";
+	else
+	    cout << "\n";
 
 	// Copyright notice
 	cout <<
@@ -619,7 +628,7 @@ void Gatekeeper::GetRotateInterval(
 		if (strspn(s, "0123456") == (size_t)s.GetLength()) {
 			m_rotateDay = s.AsInteger();
 		} else {
-			map<PCaselessString, int> dayNames;
+			std::map<PCaselessString, int> dayNames;
 			dayNames["sun"] = 0; dayNames["sunday"] = 0;
 			dayNames["mon"] = 1; dayNames["monday"] = 1;
 			dayNames["tue"] = 2; dayNames["tuesday"] = 2;
@@ -814,6 +823,7 @@ bool Gatekeeper::SetLogFilename(
 	PTrace::SetStream(&cerr);
 	
 	delete m_logFile;
+	m_logFile = NULL;
 	
 	m_logFilename = filename;
 	m_logFile = new PTextFile(m_logFilename, PFile::WriteOnly, PFile::Create);
