@@ -30,14 +30,15 @@
 #include "RasTbl.h"
 #include "RasPDU.h"
 #include "sigmsg.h"
+#include "Routing.h"
 #include "gkauth.h"
+#include "pwlib_compat.h"
 
 namespace {
 const char* const GkAuthSectionName = "Gatekeeper::Auth";
 const char OID_CAT[] = "1.2.840.113548.10.1.2.1";
 }
 
-using std::map;
 using std::stable_sort;
 using std::for_each;
 using std::find_if;
@@ -50,7 +51,7 @@ ARQAuthData::ARQAuthData(
 	const callptr& call
 	) : m_rejectReason(-1), m_callDurationLimit(-1), 
 	m_requestingEP(ep),	m_call(call), m_billingMode(-1),
-	m_routeToAlias(NULL), m_routeToIP(NULL), m_proxyMode(CallRec::ProxyDetect)
+	m_routeToAlias(NULL), m_proxyMode(CallRec::ProxyDetect)
 {
 }
 
@@ -59,13 +60,11 @@ ARQAuthData::ARQAuthData(
 	) : m_rejectReason(obj.m_rejectReason), 
 	m_callDurationLimit(obj.m_callDurationLimit), 
 	m_requestingEP(obj.m_requestingEP), m_call(obj.m_call), 
-	m_billingMode(obj.m_billingMode), m_routeToAlias(NULL), m_routeToIP(NULL),
-	m_proxyMode(obj.m_proxyMode)
+	m_billingMode(obj.m_billingMode), m_routeToAlias(NULL),
+	m_destinationRoutes(obj.m_destinationRoutes), m_proxyMode(obj.m_proxyMode)
 {
 	if (obj.m_routeToAlias)
 		m_routeToAlias = new H225_AliasAddress(*obj.m_routeToAlias);
-	if (obj.m_routeToIP)
-		m_routeToIP = new H225_TransportAddress(*obj.m_routeToIP);
 }
 
 ARQAuthData& ARQAuthData::operator=(const ARQAuthData& obj)
@@ -81,18 +80,15 @@ ARQAuthData& ARQAuthData::operator=(const ARQAuthData& obj)
 	if (obj.m_routeToAlias)
 		m_routeToAlias = new H225_AliasAddress(*obj.m_routeToAlias);
 
-	delete m_routeToIP;
-	m_routeToIP = NULL;
-	if (obj.m_routeToIP)
-		m_routeToIP = new H225_TransportAddress(*obj.m_routeToIP);
-		
+	m_destinationRoutes = obj.m_destinationRoutes;
+			
 	return *this;
 }
 
 ARQAuthData::~ARQAuthData()
 {
 	delete m_routeToAlias;
-	delete m_routeToIP;
+	m_routeToAlias = NULL;
 }
 
 void ARQAuthData::SetRouteToAlias(H225_AliasAddress* alias)
@@ -112,24 +108,6 @@ void ARQAuthData::SetRouteToAlias(const PString& alias, int tag)
 	H323SetAliasAddress(alias, *m_routeToAlias, tag);
 }
 
-void ARQAuthData::SetRouteToIP(H225_TransportAddress* addr)
-{
-	delete m_routeToIP;
-	m_routeToIP = addr;
-}
-
-void ARQAuthData::SetRouteToIP(const H225_TransportAddress& addr)
-{
-	SetRouteToIP(new H225_TransportAddress(addr));
-}
-
-void ARQAuthData::SetRouteToIP(const PIPSocket::Address& addr, WORD port)
-{
-	SetRouteToIP(new H225_TransportAddress(
-		SocketToH225TransportAddr(addr, port ? port : GK_DEF_ENDPOINT_SIGNAL_PORT)
-		));
-}
-
 SetupAuthData::SetupAuthData(
 	/// call associated with the message (if any)
 	const callptr& call,
@@ -137,7 +115,7 @@ SetupAuthData::SetupAuthData(
 	bool fromRegistered
 	) : m_rejectReason(-1), m_rejectCause(-1), m_callDurationLimit(-1),
 	m_call(call), m_fromRegistered(fromRegistered), 
-	m_routeToAlias(NULL), m_routeToIP(NULL), m_proxyMode(CallRec::ProxyDetect)
+	m_routeToAlias(NULL), m_proxyMode(CallRec::ProxyDetect)
 {
 }
 
@@ -146,12 +124,11 @@ SetupAuthData::SetupAuthData(
 	) : m_rejectReason(obj.m_rejectReason), m_rejectCause(obj.m_rejectCause), 
 	m_callDurationLimit(obj.m_callDurationLimit), m_call(obj.m_call), 
 	m_fromRegistered(obj.m_fromRegistered),
-	m_routeToAlias(NULL), m_routeToIP(NULL), m_proxyMode(obj.m_proxyMode)
+	m_routeToAlias(NULL), m_destinationRoutes(obj.m_destinationRoutes),
+	m_proxyMode(obj.m_proxyMode)
 {
 	if (obj.m_routeToAlias)
 		m_routeToAlias = new H225_AliasAddress(*obj.m_routeToAlias);
-	if (obj.m_routeToIP)
-		m_routeToIP = new H225_TransportAddress(*obj.m_routeToIP);
 }
 
 SetupAuthData& SetupAuthData::operator=(const SetupAuthData& obj)
@@ -168,10 +145,7 @@ SetupAuthData& SetupAuthData::operator=(const SetupAuthData& obj)
 	if (obj.m_routeToAlias)
 		m_routeToAlias = new H225_AliasAddress(*obj.m_routeToAlias);
 
-	delete m_routeToIP;
-	m_routeToIP = NULL;
-	if (obj.m_routeToIP)
-		m_routeToIP = new H225_TransportAddress(*obj.m_routeToIP);
+	m_destinationRoutes = obj.m_destinationRoutes;
 		
 	return *this;
 }
@@ -179,7 +153,7 @@ SetupAuthData& SetupAuthData::operator=(const SetupAuthData& obj)
 SetupAuthData::~SetupAuthData()
 {
 	delete m_routeToAlias;
-	delete m_routeToIP;
+	m_routeToAlias = NULL;
 }
 
 void SetupAuthData::SetRouteToAlias(H225_AliasAddress* alias)
@@ -199,23 +173,6 @@ void SetupAuthData::SetRouteToAlias(const PString& alias, int tag)
 	H323SetAliasAddress(alias, *m_routeToAlias, tag);
 }
 
-void SetupAuthData::SetRouteToIP(H225_TransportAddress* addr)
-{
-	delete m_routeToIP;
-	m_routeToIP = addr;
-}
-
-void SetupAuthData::SetRouteToIP(const H225_TransportAddress& addr)
-{
-	SetRouteToIP(new H225_TransportAddress(addr));
-}
-
-void SetupAuthData::SetRouteToIP(const PIPSocket::Address& addr, WORD port)
-{
-	SetRouteToIP(new H225_TransportAddress(
-		SocketToH225TransportAddr(addr, port ? port : GK_DEF_ENDPOINT_SIGNAL_PORT)
-		));
-}
 
 // class GkAuthenticator
 GkAuthenticator::GkAuthenticator(
@@ -226,7 +183,7 @@ GkAuthenticator::GkAuthenticator(
 	: NamedObject(name), m_defaultStatus(e_fail), m_controlFlag(e_Required),
 	m_enabledRasChecks(~0U), m_supportedRasChecks(supportedRasChecks), 
 	m_enabledMiscChecks(~0U), m_supportedMiscChecks(supportedMiscChecks),
-	m_config(GkConfig()), m_h235Authenticators(NULL)
+	m_config(GkConfig())
 {
 	const PStringArray control(
 		m_config->GetString(GkAuthSectionName, name, "").Tokenise(";,")
@@ -251,7 +208,7 @@ GkAuthenticator::GkAuthenticator(
 			<< GetName() << '\''
 			);
 
-	map<PString, unsigned> rasmap;
+	std::map<PString, unsigned> rasmap;
 	rasmap["GRQ"] = RasInfo<H225_GatekeeperRequest>::flag,
 	rasmap["RRQ"] = RasInfo<H225_RegistrationRequest>::flag,
 	rasmap["URQ"] = RasInfo<H225_UnregistrationRequest>::flag,
@@ -261,7 +218,7 @@ GkAuthenticator::GkAuthenticator(
 	rasmap["LRQ"] = RasInfo<H225_LocationRequest>::flag,
 	rasmap["IRQ"] = RasInfo<H225_InfoRequest>::flag;
 		
-	map<PString, unsigned> miscmap;
+	std::map<PString, unsigned> miscmap;
 	miscmap["SETUP"] = e_Setup;
 	miscmap["SETUPUNREG"] = e_SetupUnreg;
 	
@@ -328,6 +285,8 @@ GkAuthenticator::GkAuthenticator(
 		<< rasFlagsStr << ", OTHER: " << miscFlagsStr
 		);
 #endif	
+
+	m_h235Authenticators = NULL;
 }
 
 GkAuthenticator::~GkAuthenticator()
@@ -1110,9 +1069,17 @@ SimplePasswordAuth::SimplePasswordAuth(
 	m_encryptionKey = GetConfig()->GetInteger(name, "KeyFilled", 0);
 	m_checkID = Toolkit::AsBool(GetConfig()->GetString(name, "CheckID", "0"));
 	m_cache = new CacheManager(GetConfig()->GetInteger(name, "PasswordTimeout", -1));
-	
+
+#ifdef OpenH323Factory
+    PFactory<H235Authenticator>::KeyList_T keyList = PFactory<H235Authenticator>::GetKeyList();
+    PFactory<H235Authenticator>::KeyList_T::const_iterator r;
+    for (r = keyList.begin(); r != keyList.end(); ++r) {
+       H235Authenticator * Auth = PFactory<H235Authenticator>::CreateInstance(*r);
+       AppendH235Authenticator(Auth);
+	}
+#else
 	H235Authenticator* authenticator;
-	
+
 	authenticator = new H235AuthSimpleMD5;
 	authenticator->SetLocalId("dummy");
 	authenticator->SetRemoteId("dummy");
@@ -1123,15 +1090,17 @@ SimplePasswordAuth::SimplePasswordAuth(
 	authenticator->SetRemoteId("dummy");
 	authenticator->SetPassword("dummy");
 	AppendH235Authenticator(authenticator);
-/*
-#if P_SSL
-	authenticator = new H235AuthProcedure1;
-	authenticator->SetLocalId("dummy");
-	authenticator->SetRemoteId("dummy");
-	authenticator->SetPassword("dummy");
-	AppendH235Authenticator(authenticator);
+
+//#if P_SSL
+//	authenticator = new H235AuthProcedure1;
+//	authenticator->SetLocalId("dummy");
+//	authenticator->SetRemoteId("dummy");
+//	authenticator->SetPassword("dummy");
+//	AppendH235Authenticator(authenticator);
+//#endif
+
 #endif
-*/
+
 }
 
 SimplePasswordAuth::~SimplePasswordAuth()
@@ -1430,6 +1399,56 @@ int SimplePasswordAuth::CheckCryptoTokens(
 	return e_next;
 }
 
+bool SimplePasswordAuth::ResolveUserName(
+		const H225_ArrayOf_ClearToken& tokens,
+	    const H225_ArrayOf_CryptoH323Token& cryptotokens,
+		PString & username
+		)
+{
+	for (PINDEX i = 0; i < cryptotokens.GetSize(); i++) {
+	  // MD5
+		if (cryptotokens[i].GetTag() == H225_CryptoH323Token::e_cryptoEPPwdHash) {
+			H225_CryptoH323Token_cryptoEPPwdHash& pwdhash = cryptotokens[i];
+			username = AsString(pwdhash.m_alias, false);
+			return true;
+		} else if (cryptotokens[i].GetTag() == H225_CryptoH323Token::e_nestedcryptoToken) {
+			H235_ClearToken clearToken;
+			bool found = false;
+			const H235_CryptoToken& nestedCryptoToken = cryptotokens[i];
+
+		    // H235.1
+			if (nestedCryptoToken.GetTag() == H235_CryptoToken::e_cryptoHashedToken) {
+			  const H235_CryptoToken_cryptoHashedToken& cryptoHashedToken = nestedCryptoToken;
+			  clearToken = cryptoHashedToken.m_hashedVals;
+			  found = true;
+			}
+			// H235.2
+			if (nestedCryptoToken.GetTag() == H235_CryptoToken::e_cryptoSignedToken) {			
+	          const H235_CryptoToken_cryptoSignedToken & cryptoSignedToken = nestedCryptoToken;;
+              H235_SIGNED<H235_EncodedGeneralToken> m_Signed = cryptoSignedToken.m_token;
+			  m_Signed.m_toBeSigned.DecodeSubType(clearToken);
+			  found = true;
+			}
+		
+			if (found && (clearToken.HasOptionalField(H235_ClearToken::e_sendersID))) {
+               username = clearToken.m_sendersID; 
+			   return true;
+			}
+		} 
+	}
+
+	for (PINDEX j = 0; j < cryptotokens.GetSize(); j++) {
+		H235_ClearToken& token = tokens[j];
+		// CAT
+			if (token.HasOptionalField(H235_ClearToken::e_generalID)) {
+                username = token.m_generalID;
+				return true;
+			}
+	}
+
+	return false;
+}
+
 // class AliasAuth
 AliasAuth::AliasAuth(
 	const char* name,
@@ -1618,7 +1637,7 @@ class AuthObj;
 class PrefixAuth : public GkAuthenticator 
 {
 public:
-	typedef map< PString, AuthRule *, greater<PString> > Rules;
+	typedef std::map< PString, AuthRule *, greater<PString> > Rules;
 
 	enum SupportedRasChecks {
 		PrefixAuthRasChecks = RasInfo<H225_AdmissionRequest>::flag
@@ -1922,6 +1941,7 @@ int AuthRule::Check(
 inline void delete_rule(PrefixAuth::Rules::value_type r)
 {
 	delete r.second;
+	r.second = NULL;
 }
 
 IPv4AuthRule::IPv4AuthRule(
@@ -2038,6 +2058,7 @@ PrefixAuth::PrefixAuth(
 			rls[k-1]->SetNext(rls[k]);
 		m_prefrules[key] = rls[0];
 		delete [] rls;
+		rls = NULL;
 	}
 
 	if (m_prefrules.size() == 0)
