@@ -3,7 +3,7 @@
  *
  * SQL authentication/authorization modules for GNU Gatekeeper
  *
- * $Id: sqlauth.cxx,v 1.8 2005/04/24 16:39:45 zvision Exp $
+ * $Id: sqlauth.cxx,v 1.10 2006/04/14 13:56:19 willamowius Exp $
  *
  * Copyright (c) 2004, Michal Zygmuntowicz
  *
@@ -28,6 +28,7 @@
 #include "stl_supp.h"
 #include "RasTbl.h"
 #include "RasPDU.h"
+#include "Routing.h"
 #include "Toolkit.h"
 #include "RasSrv.h"
 #include "gksql.h"
@@ -36,7 +37,7 @@
 #include "Neighbor.h"
 #include "gkauth.h"
 
-using std::map;
+using Routing::Route;
 
 /// Generic SQL authenticator for H.235 enabled endpoints
 class SQLPasswordAuth : public SimplePasswordAuth
@@ -207,7 +208,7 @@ bool RunQuery(
 	const PString &traceStr,
 	GkSQLConnection *conn,
 	const PString &query,
-	const map<PString, PString>& params,
+	const std::map<PString, PString>& params,
 	GkSQLResult::ResultRow& resultRow,
 	long timeout
 	)
@@ -328,7 +329,7 @@ bool SQLPasswordAuth::GetPassword(
 	)
 {
 	GkSQLResult::ResultRow result;
-	map<PString, PString> params;
+	std::map<PString, PString> params;
 	params["1"] = alias;
 	params["u"] = alias;
 	params["2"] = Toolkit::GKName();
@@ -400,7 +401,7 @@ bool SQLAliasAuth::GetAuthConditionString(
 	)
 {
 	GkSQLResult::ResultRow result;
-	map<PString, PString> params;
+	std::map<PString, PString> params;
 	params["1"] = alias;
 	params["u"] = alias;
 	params["2"] = Toolkit::GKName();
@@ -500,7 +501,7 @@ int SQLAuth::Check(
 	)
 {
 	H225_RegistrationRequest &rrq = rrqPdu;
-	map<PString, PString> params;
+	std::map<PString, PString> params;
 	
 	// get the username for User-Name attribute		
 	params["u"] = GetUsername(rrqPdu);
@@ -599,7 +600,7 @@ int SQLAuth::Check(
 	)
 {
 	const H225_AdmissionRequest &arq = arqPdu;
-	map<PString, PString> params;
+	std::map<PString, PString> params;
 	
 	PIPSocket::Address addr = (arqPdu.operator->())->m_peerAddr;
 
@@ -690,15 +691,22 @@ int SQLAuth::Check(
 	if (iter != result.end()) {
 		const PString &s = iter->first;
 		if (!s) {
-			PIPSocket::Address addr;
-			WORD port = 0;
+			PStringArray tokens(s.Tokenise("; \t", FALSE));
+			for (PINDEX i = 0; i < tokens.GetSize(); ++i) {
+				PIPSocket::Address addr;
+				WORD port = 0;
 					
-			if (GetTransportAddress(s, GK_DEF_ENDPOINT_SIGNAL_PORT, addr, port)
-					&& addr.IsValid() && port != 0) {
-				authData.SetRouteToIP(addr, port);
-				PTRACE(5, traceStr << " - call redirected to the address "
-					<< addr << ':' << port
-					);
+				if (GetTransportAddress(tokens[i], GK_DEF_ENDPOINT_SIGNAL_PORT, addr, port)
+						&& addr.IsValid() && port != 0) {
+					Route route("SQL", addr, port);
+					route.m_destEndpoint = RegistrationTable::Instance()->FindBySignalAdr(
+						SocketToH225TransportAddr(addr, port)
+						);
+					authData.m_destinationRoutes.push_back(route);
+					PTRACE(5, traceStr << " - call redirected to the address " <<
+						route.AsString()
+						);
+				}
 			}
 		}
 	}
@@ -724,7 +732,7 @@ int SQLAuth::Check(
 	)
 {
 	H225_LocationRequest &lrq = lrqPdu;
-	map<PString, PString> params;
+	std::map<PString, PString> params;
 	
 	PIPSocket::Address addr = (lrqPdu.operator->())->m_peerAddr;
 
@@ -796,7 +804,7 @@ int SQLAuth::Check(
 	SetupAuthData& authData
 	)
 {
-	map<PString, PString> params;
+	std::map<PString, PString> params;
 	
 	PIPSocket::Address addr;
 	setup.GetPeerAddr(addr);
@@ -866,15 +874,22 @@ int SQLAuth::Check(
 	if (iter != result.end()) {
 		const PString &s = iter->first;
 		if (!s) {
-			PIPSocket::Address addr;
-			WORD port = 0;
+			PStringArray tokens(s.Tokenise("; \t", FALSE));
+			for (PINDEX i = 0; i < tokens.GetSize(); ++i) {
+				PIPSocket::Address addr;
+				WORD port = 0;
 					
-			if (GetTransportAddress(s, GK_DEF_ENDPOINT_SIGNAL_PORT, addr, port)
-					&& addr.IsValid() && port != 0) {
-				authData.SetRouteToIP(addr, port);
-				PTRACE(5, traceStr << " - call redirected to the address "
-					<< addr << ':' << port
-					);
+				if (GetTransportAddress(tokens[i], GK_DEF_ENDPOINT_SIGNAL_PORT, addr, port)
+						&& addr.IsValid() && port != 0) {
+					Route route("SQL", addr, port);
+					route.m_destEndpoint = RegistrationTable::Instance()->FindBySignalAdr(
+						SocketToH225TransportAddr(addr, port)
+						);
+					authData.m_destinationRoutes.push_back(route);
+					PTRACE(5, traceStr << " - call redirected to the address " <<
+						route.AsString()
+						);
+				}
 			}
 		}
 	}
