@@ -32,6 +32,11 @@ using std::copy;
 using std::back_inserter;
 using std::find;
 
+namespace {
+/// time to recheck state of closed sockets owned by a proxy handler
+const long SOCKETSREADER_IDLE_TIMEOUT = 1000;
+}
+
 #ifdef LARGE_FDSET
 
 bool YaSelectList::Select(SelectType t, const PTimeInterval & timeout)
@@ -53,8 +58,8 @@ bool YaSelectList::Select(SelectType t, const PTimeInterval & timeout)
 	struct timeval tval;
 	tval.tv_sec  = msec / 1000;
 	tval.tv_usec = (msec - tval.tv_sec * 1000) * 1000;
-	bool r = ::select(maxfd + 1, readfds, writefds, 0, &tval) > 0;
-	if (r) {
+	int r = ::select(maxfd + 1, readfds, writefds, 0, &tval);
+	if (r > 0) {
 #if 1
 		std::vector<YaSocket*>::iterator last = remove_if(
 			fds.begin(), fds.end(),
@@ -93,7 +98,7 @@ bool YaSelectList::Select(SelectType t, const PTimeInterval & timeout)
 				hasfd = true;
 				break;
 			} else
-				j++;
+				++j;
 		}
 		// reorder remaining sockets, so non-selected sockets 
 		// are moved to the end of the vector
@@ -108,8 +113,9 @@ bool YaSelectList::Select(SelectType t, const PTimeInterval & timeout)
 		} else
 			fds.clear();
 #endif
-	}
-	return r;
+	} else if (r < 0)
+		PTRACE(3, GetName() << "\tSelect " << (t == Read ? "read" : "write") << " error - errno: " << errno);
+	return r > 0;
 }
 
 
@@ -151,7 +157,7 @@ bool YaSocket::Read(void *buf, int sz)
 bool YaSocket::ReadBlock(void *buf, int len)
 {
 	// lazy implementation, but it is enough for us...
-	return Read(buf, len) && (lastReadCount == len);
+	return Read(buf, len) && lastReadCount == len;
 }
 
 bool YaSocket::CanRead(
@@ -227,6 +233,19 @@ bool YaSocket::SetOption(int option, const void *value, int size, int level)
 	return ConvertOSError(::setsockopt(os_handle, level, option, (char *)value, size));
 }
 
+bool YaSocket::GetOption(int option, int & value, int level)
+{
+	socklen_t valSize = sizeof(value);
+	return ConvertOSError(::getsockopt(os_handle, level, option, (char *)&value, &valSize));
+}
+
+bool YaSocket::GetOption(int option, void * valuePtr, PINDEX valueSize, int level)
+{
+	return ConvertOSError(::getsockopt(os_handle, level, option,
+		(char *)valuePtr, (socklen_t *)&valueSize)
+		);
+}
+
 PString YaSocket::GetErrorText(PSocket::ErrorGroup group) const
 {
 	return PSocket::GetErrorText(GetErrorCode(group));
@@ -234,7 +253,7 @@ PString YaSocket::GetErrorText(PSocket::ErrorGroup group) const
 
 bool YaSocket::ConvertOSError(int libReturnValue, PSocket::ErrorGroup group)
 {
-	if ((libReturnValue < 0) && (errno == EAGAIN)) {
+	if (libReturnValue < 0 && errno == EAGAIN) {
 		lastErrorCode[group] = PSocket::Timeout;
 		lastErrorNumber[group] = errno;
 		return false;
@@ -246,8 +265,10 @@ bool YaSocket::SetNonBlockingMode()
 {
 	if (!IsOpen())
 		return false;
+	// is call to F_SETFD with F_CLOEXEC really neccessary?
 	int cmd = 1;
-	if (::ioctl(os_handle, FIONBIO, &cmd) == 0 && ::fcntl(os_handle, F_SETFD, 1) == 0)
+	if (ConvertOSError(::ioctl(os_handle, FIONBIO, &cmd))
+			&& ConvertOSError(::fcntl(os_handle, F_SETFD, 1)))
 		return true;
 	Close();
 	return false;
@@ -301,7 +322,7 @@ bool YaTCPSocket::SetLinger()
 
 bool YaTCPSocket::Listen(unsigned qs, WORD pt, PSocket::Reusability reuse)
 {
-	return Listen(INADDR_ANY, qs, pt);
+	return Listen(INADDR_ANY, qs, pt, reuse);
 }
 
 bool YaTCPSocket::Listen(const Address & addr, unsigned qs, WORD pt, PSocket::Reusability reuse)
@@ -310,7 +331,8 @@ bool YaTCPSocket::Listen(const Address & addr, unsigned qs, WORD pt, PSocket::Re
 	if (!ConvertOSError(os_handle))
 		return false;
 
-	SetOption(SO_REUSEADDR, reuse == PSocket::CanReuseAddress ? 1 : 0);
+	if (!SetOption(SO_REUSEADDR, reuse == PSocket::CanReuseAddress ? 1 : 0))
+		return false;
 	
 //	SetNonBlockingMode();
 	if (Bind(addr, pt) && ConvertOSError(::listen(os_handle, qs)))
@@ -433,17 +455,19 @@ YaUDPSocket::YaUDPSocket()
 
 bool YaUDPSocket::Listen(unsigned, WORD pt, PSocket::Reusability reuse)
 {
-	return Listen(INADDR_ANY, 0, pt, 0, reuse);
+	return Listen(INADDR_ANY, 0, pt, reuse);
 }
 
-bool YaUDPSocket::Listen(const Address & addr, unsigned, WORD pt, int, PSocket::Reusability reuse)
+bool YaUDPSocket::Listen(const Address & addr, unsigned, WORD pt, PSocket::Reusability reuse)
 {
 	os_handle = ::socket(PF_INET, SOCK_DGRAM, 0);
 	if (!ConvertOSError(os_handle))
 		return false;
 
-	SetOption(SO_REUSEADDR, reuse == PSocket::CanReuseAddress ? 1 : 0);
-	SetNonBlockingMode();
+	if (!SetNonBlockingMode())
+		return false;
+	if (!SetOption(SO_REUSEADDR, reuse == PSocket::CanReuseAddress ? 1 : 0))
+		return false;
 	return Bind(addr, pt);
 }
 
@@ -505,8 +529,11 @@ bool SocketSelectList::Select(SelectType t, const PTimeInterval & timeout)
 	else
 		wlist = this, rlist = &dumb;
 	const PSocket::Errors r = PSocket::Select(*rlist, *wlist, timeout);
-	PTRACE_IF(3, r != PSocket::NoError, "ProxyH\tSelect error: " << r);
-	return !IsEmpty();
+	if (r != PSocket::NoError) {
+		PTRACE(3, GetName() << "\tSelect " << (t == Read ? "read" : "write") << " error: " << r);
+		return false;
+	} else
+		return !IsEmpty();
 }
 
 PSocket *SocketSelectList::operator[](int i) const
@@ -533,6 +560,9 @@ USocket::~USocket()
 		queue.clear();
 		qsize = 0;
 	}
+	PIPSocket::Address addr(0);
+	WORD port = 0;
+	self->GetLocalAddress(addr, port);
 	PTRACE(3, type << "\tDelete socket " << Name());
 }
 
@@ -571,8 +601,24 @@ bool USocket::WriteData(const BYTE *buf, int len)
 {
 	if (IsSocketOpen()) {
 		if (qsize == 0 && !writeMutex.WillBlock()) {
+			const int MAX_SOCKET_CHUNK = 10240;	// send in 10K chunks
 			PWaitAndSignal lock(writeMutex);
-			return InternalWriteData(buf, len);
+			int remaining = len;
+			int sendnow = 0;
+			bool writeResult = true;
+			while ((remaining > 0) && writeResult) {
+				if (remaining > MAX_SOCKET_CHUNK)
+					sendnow = MAX_SOCKET_CHUNK;
+				else
+					sendnow = remaining;
+				writeResult = InternalWriteData(buf, sendnow);
+				remaining -= sendnow;
+				if (remaining > 0) {
+					PThread::Sleep(250);
+					buf += sendnow;
+				}
+			}
+			return writeResult;
 		}
 		if (qsize > 100) { // to be justitied
 			PTRACE(2, type << '\t' << Name() << " is dead and closed");
@@ -649,7 +695,7 @@ void USocket::ClearQueue()
 // class SocketsReader
 SocketsReader::SocketsReader(int t) : m_timeout(t), m_socksize(0), m_rmsize(0)
 {
-	SetName("SocketsReader");
+	SetName("SockRdr");
 }
 
 SocketsReader::~SocketsReader()
@@ -671,11 +717,15 @@ void SocketsReader::Stop()
 void SocketsReader::AddSocket(IPSocket *socket)
 {
 	m_listmutex.StartWrite();
-	m_sockets.push_back(socket);
-	++m_socksize;
+	iterator iter = find(m_sockets.begin(), m_sockets.end(), socket);
+	if (iter == m_sockets.end()) {
+		m_sockets.push_back(socket);
+		++m_socksize;
+	} else
+		PTRACE(1, GetName() << "\tTrying to add an already existing socket to the handler");
 	m_listmutex.EndWrite();
 	Signal();
-	PTRACE(5, GetName() << " total sockets " << m_socksize);
+	PTRACE(5, GetName() << "\tTotal sockets: " << m_socksize);
 }
 
 bool SocketsReader::BuildSelectList(SocketSelectList & slist)
@@ -705,7 +755,7 @@ bool SocketsReader::SelectSockets(SocketSelectList & slist)
 	}
 	ConfigReloadMutex.StartRead();
 #if PTRACING
-	PString msg(PString::Printf, " %u sockets selected from %u, total %u/%u", slist.GetSize(), ss, m_socksize, m_rmsize);
+	PString msg(PString::Printf, "\t%u sockets selected from %u, total %u/%u", slist.GetSize(), ss, m_socksize, m_rmsize);
 	PTRACE(5, GetName() << msg);
 #endif
 	return true;
@@ -749,7 +799,7 @@ void SocketsReader::RemoveSocket(IPSocket *s)
 void SocketsReader::Exec()
 {
 	ReadLock cfglock(ConfigReloadMutex);
-	SocketSelectList slist;
+	SocketSelectList slist(GetName());
 
 	if (BuildSelectList(slist)) {
 		if (SelectSockets(slist)) {
@@ -765,8 +815,8 @@ void SocketsReader::Exec()
 	} else {
 		CleanUp();
 		ConfigReloadMutex.EndRead();
-		PTRACE(3, GetName() << " waiting...");
-		Wait();
+		PTRACE(6, GetName() << " waiting...");
+		Wait(SOCKETSREADER_IDLE_TIMEOUT);
 		ConfigReloadMutex.StartRead();
 	}
 }
@@ -781,26 +831,22 @@ TCPListenSocket::TCPListenSocket(int timeout)
 
 TCPListenSocket::~TCPListenSocket()
 {
-	PTRACE(3, "TCP\t" << "Delete listener " << GetName());
+	PTRACE(3, "TCP\tDelete listener " << GetName());
 }
 
 bool TCPListenSocket::IsTimeout(const PTime *now) const
 {
-	if ( readTimeout < PMaxTimeInterval )
-	{
+	if (readTimeout < PMaxTimeInterval)
 		return IsOpen() ? ((readTimeout > 0) ? ((*now - start) > readTimeout) : false) : true;
-	}
 	else
-	{
-		return ! IsOpen();
-	}
+		return !IsOpen();
 }
 
 
 // class TCPServer
 TCPServer::TCPServer()
 {
-	SetName("TCPServer");
+	SetName("TCPSrv");
 	Execute();
 }
 
@@ -808,32 +854,41 @@ bool TCPServer::CloseListener(TCPListenSocket *socket)
 {
 	ReadLock lock(m_listmutex);
 	iterator iter = find(m_sockets.begin(), m_sockets.end(), socket);
-	return (iter != m_sockets.end()) ? ((*iter)->Close() ? true : false) : false;
+	if (iter != m_sockets.end()) {
+		PTRACE(6, GetName() << "\tListener " << (*iter)->GetName() << " closed");
+		(*iter)->Close();
+		return true;
+	} else
+		return false;
 }
 
 void TCPServer::ReadSocket(IPSocket *socket)
 {
-	PTRACE(4, "TCP\t" << "Accept request on " << socket->GetName());
+	PTRACE(4, GetName() << "\tAccept request on " << socket->GetName());
 	TCPListenSocket *listener = dynamic_cast<TCPListenSocket *>(socket);
 	ServerSocket *acceptor = listener->CreateAcceptor();
-	if (acceptor->Accept(*listener))
+	if (acceptor->Accept(*listener)) {
+		PTRACE(6, GetName() << "\tAccepted new connection on " << socket->GetName() << " from " << acceptor->GetName());
 		CreateJob(acceptor, &ServerSocket::Dispatch, "Acceptor");
-	else
+	} else {
+		PTRACE(4, GetName() << "\tAccept failed on " << socket->GetName());
 		delete acceptor;
+	}
 }
 
 void TCPServer::CleanUp()
 {
 	PTime now;
 	WriteLock lock(m_listmutex);
-	iterator iter = m_sockets.begin(), eiter = m_sockets.end();
-	while (iter != eiter) {
+	iterator iter = m_sockets.begin();
+	while (iter != m_sockets.end()) {
 		iterator i = iter++;
 		TCPListenSocket *listener = dynamic_cast<TCPListenSocket *>(*i);
 		if (listener && listener->IsTimeout(&now)) {
 			m_sockets.erase(i);
 			--m_socksize;
 			delete listener;
+			iter = m_sockets.begin();
 		}
 	}
 }
