@@ -15,11 +15,12 @@
 //////////////////////////////////////////////////////////////////
 
 #ifndef YASOCKET_H
-#define YASOCKET_H "@(#) $Id: yasocket.h,v 1.13 2005/04/24 16:39:46 zvision Exp $"
+#define YASOCKET_H "@(#) $Id: yasocket.h,v 1.21 2006/06/08 08:28:49 willamowius Exp $"
 
 #include <list>
 #include <vector>
 #include "job.h"
+#include "pwlib_compat.h"
 
 //#define LARGE_FDSET 32768
 #ifdef LARGE_FDSET
@@ -52,6 +53,8 @@ public:
 	void GetLocalAddress(Address &, WORD &) const;
 	bool SetOption(int, int, int = SOL_SOCKET);
 	bool SetOption(int, const void *, int, int = SOL_SOCKET);
+	bool GetOption(int, int &, int = SOL_SOCKET);
+	bool GetOption(int, void *, PINDEX, int);
 
 	PSocket::Errors GetErrorCode(PSocket::ErrorGroup group) const { return lastErrorCode[group]; }
 	int GetErrorNumber(PSocket::ErrorGroup group) const { return lastErrorNumber[group]; }
@@ -111,7 +114,7 @@ public:
 	YaUDPSocket();
 
 	bool Listen(unsigned, WORD, PSocket::Reusability reuse = PSocket::AddressIsExclusive);
-	bool Listen(const Address &, unsigned, WORD, int, PSocket::Reusability reuse = PSocket::AddressIsExclusive);
+	bool Listen(const Address &, unsigned, WORD, PSocket::Reusability reuse = PSocket::AddressIsExclusive);
 	void GetLastReceiveAddress(Address &, WORD &) const;
 	void SetSendAddress(const Address &, WORD);
 	/// Get the address to use for connectionless Write().
@@ -141,12 +144,27 @@ public:
 		/// estimated number of sockets to be put in this select list
 		size_t reserve = 512
 	) : maxfd(0) { fds.reserve(reserve); }
+
+	/// build a select list for more than one socket
+	YaSelectList(
+		/// name for this select list
+		const PString &name,
+		/// estimated number of sockets to be put in this select list
+		size_t reserve = 512
+	) : maxfd(0), m_name(name) { fds.reserve(reserve); }
 	
 	/// build a select list for signle socket only
 	YaSelectList(
 		YaSocket* singleSocket /// socket to be put on the list
 		) : fds(1, singleSocket), maxfd(singleSocket->GetHandle()) {}
 
+	/// build a select list for signle socket only
+	YaSelectList(
+		/// name for this select list
+		const PString &name,
+		YaSocket* singleSocket /// socket to be put on the list
+		) : fds(1, singleSocket), maxfd(singleSocket->GetHandle()), m_name(name) {}
+
 	void Append(
 		YaSocket* s /// the socket to be appended
 		)
@@ -181,9 +199,12 @@ public:
 		};
 	};
 
+	PString GetName() const { return m_name; }
+	
 private:
 	std::vector<YaSocket *> fds;
 	int maxfd;
+	PString m_name;
 };
 
 typedef YaSelectList SocketSelectList;
@@ -201,8 +222,15 @@ public:
 	};
 	SocketSelectList(size_t) {};
 	SocketSelectList(PIPSocket *s = 0) { if (s && s->IsOpen()) Append(s); }
+	SocketSelectList(const PString &name, size_t) : m_name(name) {};
+	SocketSelectList(const PString &name, PIPSocket *s = NULL) : m_name(name) { if (s && s->IsOpen()) Append(s); }
 	bool Select(SelectType, const PTimeInterval &);
 	PSocket *operator[](int i) const;
+
+	PString GetName() const { return m_name; }	
+
+private:
+	PString m_name;
 };
 
 typedef PIPSocket IPSocket;
@@ -314,7 +342,7 @@ private:
 	int qsize;
 
 	bool blocked;
-	PMutex writeMutex, queueMutex;
+	PTimedMutex writeMutex, queueMutex;
 	const char *type;
 };
 
@@ -385,7 +413,7 @@ class TCPListenSocket : public TCPSocket {
 	PCLASSINFO ( TCPListenSocket, TCPSocket )
 #endif
 public:
-	TCPListenSocket::TCPListenSocket(int seconds = 0);
+	TCPListenSocket(int seconds = 0);
 	~TCPListenSocket();
 
 	bool IsTimeout(const PTime *) const;
